/*--------------------------------
 * 
 * Jerome.Gouzy@inra.fr
 * Created: 20140120
 * Copyright INRA, All rights reserved
 * 
 * $Id: generalbioinfo.c 517 2016-08-17 14:31:58Z gouzy $
 * 
 ---------------------------------*/
#include "generalbioinfo.h"

void MallocMatrixOfShort(short ***p_mat,int len)
{
	int j;
	short **local_p_mat=(short**)malloc(sizeof(short*)*len);
	assert(local_p_mat!=NULL);
	for(j=0;j<len;j++)
	{
		local_p_mat[j]=(short *)malloc(sizeof(short)*len);
		assert(local_p_mat[j]!=NULL);
	}
	*p_mat=local_p_mat;
}

void FreeMatrixOfShort(short **p_mat,int len)
{
	int j;
	for(j=0;j<len;j++)
		free(p_mat[j]);
	free(p_mat);
}

void MallocMatrixOfFloat(float ***p_mat,int x,int y)
{
	int j;
	float **local_p_mat=(float**)malloc(sizeof(float*)*x);
	assert(local_p_mat!=NULL);
	for(j=0;j<x;j++)
	{
		local_p_mat[j]=(float *)malloc(sizeof(float)*y);
		assert(local_p_mat[j]!=NULL);
	}
	*p_mat=local_p_mat;
}

void FreeMatrixOfFloat(float **p_mat,int x)
{
	int j;
	for(j=0;j<x;j++)
		if ( p_mat[j] != NULL )
			free(p_mat[j]);
	free(p_mat);
}

void ParseAALNLine(char *hit,int query_col,hsp_t *hsp)
{
//  HWI-M00185:3:000000000-A1BH4:1:1101:13549:1830      39 151             1830       1 113     d   1     0       339 <
	char code;
	if (query_col==1)
	{
		sscanf(hit,"%s %d %d %s %d %d %c %f %f %f",hsp->query,&hsp->qstart,&hsp->qend,hsp->target,&hsp->tstart,&hsp->tend,&code,&hsp->pci,&hsp->mm,&hsp->score);
	}
	else
	{
		sscanf(hit,"%s %d %d %s %d %d %c %f %f %f",hsp->target,&hsp->tstart,&hsp->tend,hsp->query,&hsp->qstart,&hsp->qend,&code,&hsp->pci,&hsp->mm,&hsp->score);
	}
	hsp->len=hsp->qend-hsp->qstart+1;
	hsp->evalue=0;
	hsp->pci*=100;
	hsp->gap=abs(hsp->len-(hsp->qend-hsp->qstart+1));
	hsp->strand = '+'; //# ne gere que le +/+
}


char *RevCompSeq(char *str)
{
	int  len     = strlen(str);
	char *revseq = (char *)malloc(sizeof(char)*(len+1));
	int i,j;
	for(i=len-1,j=0;i>=0;i--,j++)
	{
		switch(str[i])
		{
			case 'A':
			case 'a':
				revseq[j]='T';
				break;
			case 'T':
			case 't':
				revseq[j]='A';
				break;
			case 'C':
			case 'c':
				revseq[j]='G';
				break;
			case 'G':
			case 'g':
				revseq[j]='C';
				break;
			default:
				revseq[j]=str[i];
				break;
		}
	}
	revseq[len]='\0';
	return revseq;
}

void AssertKmerEncoding(int klen,kmercode_t code)
{
//	fprintf(stderr,"%lf > %lf\n",(double)sizeof(code),(double)klen/(double)4);
	// char = 8 bits = 1 octet ==> 4 nt / char/octet
	assert((double)sizeof(code) >= (double)klen/(double)4);
}



void PrintMat(short **p_mat,char *s1,char *s2)
{
	int i,j;
	fprintf(stderr,"    ");
	for(j=0;j<strlen(s2)+1;j++)
	{
		fprintf(stderr,"%3d%c",j,s2[j-1]);
	}
	fprintf(stderr,"\n");
	for(i=0;i<strlen(s1)+1;i++)
	{
		fprintf(stderr,"%3d%c",i,s1[i-1]);
		for(j=0;j<strlen(s2)+1;j++)
		{
			fprintf(stderr," %3d",p_mat[i][j]);
		}
		fprintf(stderr,"\n");
	}
}

int PairwiseMatchUngapped(char *query,char *target,hsp_t *hsp,short **p_mat)
{
	int delta_code=0;
	//fprintf(stderr,"query:%s#\n#target:%s#\n",query,target);
	int maxq=strlen(query)+1;
	int maxt=strlen(target)+1;
#ifndef NDEBUG
	assert(maxt > 0);
	assert(maxq > 0);
	assert(query);
	assert(target);
#endif
	short i,j;
	//unsigned short p_mat[100][100]={0};
	int alloc_p_mat=0;
	if ( p_mat==NULL )
	{
		alloc_p_mat=1;
		p_mat=(short**)malloc(sizeof(short*)*maxq);
		for(i=0;i<maxq;i++)
			p_mat[i]=(short *)malloc(sizeof(short)*maxt);
	}
	for(i=0;i<maxq;i++)
		for(j=0;j<maxt;j++)
			p_mat[i][j]=0;
		
//	fprintf(stderr,"%s, l=%d ; %s, l=%d\n",query,maxq-1,target,maxt-1);
	short max_score=0;
	hsp->qend=0;
	hsp->tend=0;
	for(i=1;i<maxq;i++)
		for(j=1;j<maxt;j++)
		{
			delta_code = query[i-1]-target[j-1];
			p_mat[i][j] = p_mat[i-1][j-1] + (( delta_code == 0 || abs(delta_code) == 32  ) ? 1 : -1); // AA, aa et 32=aT
			if ( p_mat[i][j] < -1 )
				p_mat[i][j]=-1;
			if ( p_mat[i][j] > max_score )
			{
				max_score = p_mat[i][j];
				hsp->qend=i;
				hsp->tend=j;
			}
		}
	
	//PrintMat(p_mat,query,target);
	i=hsp->qend;
	j=hsp->tend;
	if ( i <= 0 || j<=0  )
	{
		if ( alloc_p_mat )
		{
			for(i=0;i<maxq;i++)
				free(p_mat[i]);
			free(p_mat);
		}
		InitHsp(hsp);
		return(0);
	}
	//fprintf(stderr,"%d,%d,%d,%d,max_score=%d\n",i,j,maxq,maxt,max_score);
	assert(i>0);
	assert(j>0);
	delta_code = query[i-1]-target[j-1];
	hsp->mm= (( delta_code == 0 || abs(delta_code) == 32  ) ? 0 : 1); // AA, aa et 32=aT
	while(i>0 && j>0)
	{
		i--;
		j--;
		delta_code = query[i]-target[j];
		hsp->mm+= (( delta_code == 0 || abs(delta_code) == 32  ) ? 0 : 1); // AA, aa et 32=aT
		if ( p_mat[i][j] == 0 )
		{
			hsp->qstart=i;
			hsp->tstart=j;
			break;
		}
	}
	hsp->strand='+';
	hsp->gap=0;
	hsp->qstart++;
	hsp->tstart++;
	hsp->len=hsp->qend-hsp->qstart+1;
	hsp->pci=(hsp->len-hsp->mm)*100/hsp->len;
	assert(hsp->len > 0);
//	assert(hsp->pci > 0);
	assert(hsp->qstart <= hsp->qend);
	assert(hsp->tstart <= hsp->tend);
	assert(hsp->mm >= 0);
	assert(hsp->qstart > 0 );
	assert(hsp->tstart > 0 );
	assert(hsp->qend < maxq);
	assert(hsp->tend < maxt);
	//fprintf(stderr,"%d-%d  %d-%d, mm=%f pci=%f %d\n",hsp->qstart,hsp->qend,hsp->tstart,hsp->tend,hsp->mm,hsp->pci,hsp->len);
	if ( alloc_p_mat )
	{
		for(i=0;i<maxq;i++)
			free(p_mat[i]);
		free(p_mat);
	}
	return(1);
}


int PairwiseMatchGapped(char *query,char *target,hsp_t *hsp,short gap_score,short **p_mat)
{
	assert(gap_score);
	if( gap_score > 0 )
		gap_score *= -1;
	int delta_code=0;
	//fprintf(stderr,"query:%s#\n#target:%s#\n",query,target);
	int maxq=strlen(query)+1;
	int maxt=strlen(target)+1;
#ifndef NDEBUG
	assert(maxt > 0);
	assert(maxq > 0);
	assert(query);
	assert(target);
#endif
	short i,j;
	int alloc_p_mat=0;
	if ( p_mat == NULL )
	{	
		alloc_p_mat=1;
		p_mat=(short**)malloc(sizeof(short*)*maxq);
		for(i=0;i<maxq;i++)
			p_mat[i]=(short *)malloc(sizeof(short)*maxt);
	}
	for(i=0;i<maxq;i++)
		for(j=0;j<maxt;j++)
			p_mat[i][j]=0;
		
		
		/*
	for(i=1;i<maxq;i++)
		p_mat[i][0] = gap_score * i;
	for(j=1;j<maxt;j++)
		p_mat[0][j] = gap_score * j;
	*/
	//PrintMat(p_mat,query,target);
	
		
//	fprintf(stderr,"%s, l=%d ; %s, l=%d\n",query,maxq-1,target,maxt-1);
	short match=0;
	short gaps=0;
	short gapt=0;
	short max_score=0;
	short lastmatchi=0;
	short lastmatchj=0;
	short dropoff=0;
	hsp->qend=0;
	hsp->tend=0;
	for(i=1;i<maxq;i++)
		for(j=1;j<maxt;j++)
		{
			delta_code = query[i-1]-target[j-1];
			match = p_mat[i-1][j-1] + (( delta_code == 0 || abs(delta_code) == 32  ) ? 1 : -1);
			gaps  = p_mat[i][j-1]+gap_score;
			gapt  = p_mat[i-1][j]+gap_score;
			p_mat[i][j] = ( match > gaps ) ? match : gaps;
			if ( p_mat[i][j] < gapt )
				p_mat[i][j] = gapt;
			// rajoute le 20140716 pour ne pas descendre trop bas
			if ( p_mat[i][j] < -1 )
				p_mat[i][j] = -1;
			if ( p_mat[i][j] > max_score )
			{
				max_score = p_mat[i][j];
				hsp->qend=i;
				hsp->tend=j;
			}
		}
	
	//PrintMat(p_mat,query,target);
	i=hsp->qend;
	j=hsp->tend;
	if ( i <= 0 || j<=0  )
	{
		if ( alloc_p_mat )
		{
			for(i=0;i<maxq;i++)
				free(p_mat[i]);
			free(p_mat);
		}
		InitHsp(hsp);
		return(0);
	}
	//fprintf(stderr,"%d,%d,%d,%d,max_score=%d\n",i,j,maxq,maxt,max_score);
#ifndef NDEBUG
	assert(i>0);
	assert(j>0);
#endif
	hsp->gap=0;
	hsp->mm=0;
	while(i>0 && j>0)
	{
		delta_code = query[i-1]-target[j-1];
		match = ( delta_code == 0 || abs(delta_code) == 32  ) ? 1 : -1;
		if (p_mat[i][j] == p_mat[i-1][j-1]+match )
		{
			i--;
			j--;
			hsp->mm+=( match > 0 ) ? 0 : 1;
			if ( match > 0 )
			{
				dropoff=5;
				lastmatchi=i;
				lastmatchj=j;
			}
			else
				dropoff--;
		}
		else if ( p_mat[i][j]-gap_score == p_mat[i][j-1] )
		{
			dropoff--;
			hsp->gap++;
			j--;
		}
		else if ( p_mat[i][j]-gap_score == p_mat[i-1][j] )
		{
			dropoff--;
			hsp->gap++;
			i--;
		}
		else
		{
			fprintf(stderr,"PairwiseMatchGapped PB %d,%d = %d\n",i,j,p_mat[i][j]);
			exit(0);
		}
		//hsp->mm+=( query[i] != target[j] ) ? 1 : 0;
		// ATTENTION: verifier qu'il ne faut pas sortir l'init de la boucle (cf PairwiseMatchGappedConcatTargets)
		// JER 20180214 fprintf(stderr,"WARNING: verifier qu'il ne faut pas sortir l'init de la boucle (cf PairwiseMatchGappedConcatTarget\n");
		if ( p_mat[i][j] <= 0 || dropoff < 0 )
		{
			hsp->qstart=lastmatchi;
			hsp->tstart=lastmatchj;
			break;
		}
	}
	hsp->strand='+';
	hsp->qstart++;
	hsp->tstart++;
	hsp->len=hsp->qend-hsp->qstart+1;
	hsp->pci=(hsp->len-hsp->mm-hsp->gap)*100/hsp->len;
#ifndef NDEBUG
	assert(hsp->len > 0);
//	assert(hsp->pci > 0);
	assert(hsp->qstart <= hsp->qend);
	assert(hsp->tstart <= hsp->tend);
	assert(hsp->mm >= 0);
	assert(hsp->qstart > 0 );
	assert(hsp->tstart > 0 );
	assert(hsp->qend < maxq);
	assert(hsp->tend < maxt);
#endif
	//fprintf(stderr,"%d-%d  %d-%d, mm=%f pci=%f %d\n",hsp->qstart,hsp->qend,hsp->tstart,hsp->tend,hsp->mm,hsp->pci,hsp->len);
	if ( alloc_p_mat )
	{
		for(i=0;i<maxq;i++)
			free(p_mat[i]);
		free(p_mat);
	}
	return(1);
}

//plus bon/modifie mais non valide depuis que je fais de l'entrelace
void PrintMatConcatTargets(int *p_mat_block,int maxq,int maxonet,int n_targets)
{
	int i,j,t;
	int maxlent_ntgt=maxonet*n_targets;
	fprintf(stderr,"maxq=%d, maxonet=%d, n_targets=%d\n",maxq,maxonet,n_targets);
	for(t=0;t<n_targets;t++)
	{
		//int offset=maxq*maxonet*t;
		//fprintf(stderr,"T%d (%d)\t",t,offset);
		fprintf(stderr,"T%d\t",t);
		for(i=0;i<maxq;i++)
		{
			fprintf(stderr,"\t%d",i);
		}
		fprintf(stderr,"\n");
		for(j=0;j<maxonet;j++)
		{
			fprintf(stderr,"%d\t",j);
			for(i=0;i<maxq;i++)
			{
				//int icurr = offset+i*maxonet+j;
				int icurr = i*maxlent_ntgt+j*n_targets+t;
				fprintf(stderr,"\t%d",p_mat_block[icurr]);
			}
			fprintf(stderr,"\n");
		}
		fprintf(stderr,"\n");
	}
}

// Test de vectorisation:
// j'envois un tableau de sequences de meme taille
// target est en fait une concatenation de X sequences de meme taille ==> tentative de vectorisation
// je change la facon d'allouer pour garantir un seul bloc et faciliter l'init

int PairwiseMatchGappedConcatTargetsForAssembly(char * query,char targets[],int tgtlen,hsp_t *hsp[],short gap_score,int * p_mat_block, int n_targets,int ref_dropoff,float minpci,int minovl,int maxovht)
{
	//fprintf(stderr,"NEW: %zu\n%s#\n",strlen(targets),targets);
	assert(gap_score);
	if( gap_score > 0 )
		gap_score *= -1;
	int maxq=strlen(query)+1;
	int i,j,t;
	int maxonet=tgtlen+1;
	int maxt=maxonet*n_targets;
	for (t=0;t<n_targets;t++)
	{
		hsp[t]->qend=0;
		hsp[t]->tend=0;
	}
	
	int alloc_p_mat=0;
	
#if defined(__MIC__)
	__assume_aligned(p_mat_block, 32);
	__assume_aligned(targets, 32);
#endif
	
	if ( p_mat_block == NULL )
	{	
		alloc_p_mat=1;
		p_mat_block=AlignedMallocOnPageSize(sizeof(int)*maxq*maxt);
	}
	
	memset(p_mat_block,0,sizeof(int)*(maxq*maxt));
	
	//PrintMatConcatTargets(p_mat_block,maxq,maxonet,n_targets);
	
	
//int *mat = (int *)malloc(rows * cols * sizeof(int)); => rows = t, cols = q
//int offset = i * cols + j; for row-major ordering 
//and int offset = i + rows * j; for column-major ordering.
// not mat[offset] corresponds to m(i, j)
	//PrintMat(p_mat,query,target);
	
	int a_max_scores[n_targets];
//	int a_offset[n_targets];
	int a_delta_code[n_targets];
	int a_match[n_targets];
	int a_gaps[n_targets];
	int a_gapt[n_targets];
	int a_val[n_targets];
//	int a_icurr[n_targets];
	int a_qend[n_targets];
	int a_tend[n_targets];
	int a_iprev_mem[n_targets];
	int a_jprev_mem[n_targets];
	for(t=0;t<n_targets;t++)
	{
		a_max_scores[t]=0;
		a_jprev_mem[t]=0;
		a_iprev_mem[t]=0;
		a_qend[t]=0;
		a_tend[t]=0;
	}
//	char target[n_targets];
	
	// entrelacement p_mat_block pour que les donnees des differentes targets soient proches
	// 11112222333344445555
	
	// idem pour les targets qui sont entrelacees avant d'arriver
	
	int maxlent_ntgt = maxonet*n_targets;
	for(i=1;i<maxq;i++)
	{
		int icurr = i*maxlent_ntgt;
		int iprev = (i-1)*maxlent_ntgt;
		for(t=0;t<n_targets;t++)
		{
			a_jprev_mem[t]=0;
			a_iprev_mem[t]=0;
		}
		for(j=1;j<maxonet;j++)
		{
			int jprev = (j-1)*n_targets;
			int jcurr = j*n_targets;
			// verification faite que la vectorisation se fait bien avec ces blocks
			
			// je pourrais essayer de garder (j,i-1) que je vais reutiliser
			for(t=0;t<n_targets;t++)
			{
#pragma loop_count avg (4)
				a_delta_code[t]= query[i-1]-targets[(j-1)*n_targets+t];
			}
				
#if defined(__MIC__)
//	#pragma vector aligned
#endif
			for(t=0;t<n_targets;t++)
			{
#pragma loop_count avg (4)
				//a_gapt[t]= p_mat_block[iprev+jcurr+t]+gap_score;
				a_gapt[t]= a_iprev_mem[t]+gap_score;
				
				a_iprev_mem[t]=p_mat_block[iprev+jprev+t];
				
				a_match[t]= a_iprev_mem[t] + (( a_delta_code[t] == 0 || abs(a_delta_code[t]) == 32  ) ? 1 : -1);
				
				//a_gaps[t]= p_mat_block[icurr+jprev+t]+gap_score;
				a_gaps[t]= a_jprev_mem[t]+gap_score;
				
				// ATTENTION: effet de bord: pas de difference entre un hit de 1 et de 2 si l'on commence par un mm (score=1 dans les deux cas)
				a_val[t] = ( a_match[t] > a_gaps[t] ) ? a_match[t] : a_gaps[t];

			}

			for(t=0;t<n_targets;t++)
			{
#pragma loop_count avg (4)
				if ( a_val[t] < a_gapt[t] )
					a_val[t] = a_gapt[t];
				if ( a_val[t] < -1 )
					a_val[t] = -1;
				if ( a_val[t] > a_max_scores[t] )
				{
					a_max_scores[t] = a_val[t];
					a_qend[t]=i;
					a_tend[t]=j;
				}
			}
			
#if defined(__MIC__)
//	#pragma vector aligned
#endif
			for(t=0;t<n_targets;t++)
			{
#pragma loop_count avg (4)
				/*
				p_mat_block[a_icurr[t]]=a_val[t];
				*/
				p_mat_block[icurr+jcurr+t]=a_val[t];
				a_jprev_mem[t]=a_val[t];
			}
		}
	}
	
	for(t=0;t<n_targets;t++)
	{
#pragma loop_count avg (4)
		hsp[t]->qend=a_qend[t];
		hsp[t]->tend=a_tend[t];
	}


#ifndef NDEBUG
	if (maxonet < 50 )
		PrintMatConcatTargets(p_mat_block,maxq,maxonet,n_targets);
#endif
	
	// je desactive les candidats qui ne pourront pas atteindre les seuils
	// et pour lesquels il n'est pas necessaire de faire la remontee
	for(t=0;t<n_targets;t++)
	{
		if ( hsp[t]->tend+maxovht < maxonet-1 || a_max_scores[t] < (minovl*minpci/100) )
			hsp[t]->tend=0;
	}
			
	for(t=0;t<n_targets;t++)
	{
		i=hsp[t]->qend;
		j=hsp[t]->tend;
		if ( i <= 0 || j<=0 )
		{
			InitHsp(hsp[t]);
			continue;
		}
		/*
		int offset=maxq*maxonet*t;
		*/
		int lastmatchi=0;
		int lastmatchj=0;
		int dropoff = 0;
#ifndef NDEBUG
		assert(i>0);
		assert(j>0);
#endif
		hsp[t]->gap=0;
		hsp[t]->mm=0;
		while(i>0 && j>0)
		{
			int delta_code = query[i-1]-targets[(j-1)*n_targets+t];
			int match = ( delta_code == 0 || abs(delta_code) == 32  ) ? 1 : -1;
			/*
			int icurr = offset+i*maxonet+j;
			*/
			int icurr =i*maxlent_ntgt+j*n_targets+t;
			//fprintf(stderr,"%d => %d\n",icurr,p_mat_block[icurr]);
				/*
			if ( p_mat_block[icurr] == p_mat_block[offset+(i-1)*maxonet+j-1]+match )
				*/
			if ( p_mat_block[icurr] == p_mat_block[(i-1)*maxlent_ntgt+(j-1)*n_targets+t]+match )
			{
				i--;
				j--;
				hsp[t]->mm+=( match > 0 ) ? 0 : 1;
				if ( match > 0 )
				{
					dropoff=ref_dropoff;
					lastmatchi=i;
					lastmatchj=j;
				}
				else
					dropoff--;
			}
			/*
			else if ( p_mat_block[icurr]-gap_score == p_mat_block[offset+i*maxonet+j-1] )
				*/
			else if ( p_mat_block[icurr]-gap_score == p_mat_block[i*maxlent_ntgt+(j-1)*n_targets+t] )
			{
				dropoff--;
				hsp[t]->gap++;
				j--;
			}
			/*
			else if ( p_mat_block[icurr]-gap_score == p_mat_block[offset+(i-1)*maxonet+j] )
				*/
			else if ( p_mat_block[icurr]-gap_score == p_mat_block[(i-1)*maxlent_ntgt+j*n_targets+t] )
			{
				dropoff--;
				hsp[t]->gap++;
				i--;
			}
			else
			{
				fprintf(stderr,"PairwiseMatchGapped PB %d,%d = %d\n",i,j,p_mat_block[icurr]);
				exit(0);
			}
			//hsp->mm+=( query[i] != target[j] ) ? 1 : 0;
			if ( p_mat_block[icurr] <= 0 || dropoff < 0 )
			{
				break;
			}
		}
		hsp[t]->qstart=lastmatchi;
		hsp[t]->tstart=lastmatchj;
	}
	
	// je n'essaye pas de vectoriser vu que je filtre avant de faire le traceback il n'y en aura pas n_targets
	for(t=0;t<n_targets;t++)
	{
		hsp[t]->strand='+';
		hsp[t]->qstart++;
		hsp[t]->tstart++;
		hsp[t]->len=hsp[t]->qend-hsp[t]->qstart+1;
		hsp[t]->pci=(hsp[t]->len-hsp[t]->mm-hsp[t]->gap)*100/hsp[t]->len;
	}
	if ( alloc_p_mat )
#if defined(__MIC__)
		_mm_free(p_mat_block);
#else
		free(p_mat_block);
#endif
	return(1);
}

void InitHsp(hsp_t *hsp)
{
	hsp->mm=99999.;
	hsp->gap=SHRT_MAX;
	hsp->evalue=1;
	hsp->score=0;
	hsp->pci=0;
	hsp->qend=0;
	hsp->tstart=0;
	hsp->tend=0;
	hsp->len=0;
	hsp->strand='+';
	hsp->pci=0;
}

void PrintHsp(hsp_t *hsp)
{
	fprintf(stderr,"mm=%.2f, gap=%d, pci=%.2f, qstart=%d, qend=%d, tstart=%d, tend=%d len=%d strand=%c\n",
			hsp->mm, hsp->gap, hsp->pci, hsp->qstart, hsp->qend, hsp->tstart, hsp->tend, hsp->len, hsp->strand);
	//fprintf(stderr,"mm=%f, gap=%d, evalue=%.2lf, score=%f, pci=%f, qstart=%d, qend=%d, tstart=%d, tend=%d len=%d strand=%c\n",
	//		hsp->mm, hsp->gap, hsp->evalue, hsp->score, hsp->pci, hsp->qstart, hsp->qend, hsp->tstart, hsp->tend, hsp->len, hsp->strand);
}
