/*--------------------------------
 * 
 * Jerome.Gouzy@inra.fr
 * Created: 20140120
 * Copyright INRA, All rights reserved
 * Mettre une licence permissive ici et sur les autres utilitaires
 * 
 * $Id: general.c 540 2017-09-27 09:52:37Z gouzy $
 * 
 ---------------------------------*/
#include "general.h"

void Fatal(char *msg)
{
	fprintf(stderr,"%s\n",msg);
	exit(1);
}

void Date(char *tag)
{
	char *time_str = GetTime();
	fprintf(stderr,"DATE %s: %s\n",tag,time_str);
	fflush(stderr);
	free(time_str);
}

int GetExtension(char *file)
{
	char *ext=strrchr(file,'.');
	if ( ext == NULL )
		return COMPRESSED_NO;
	else
	{
		if ( strcmp(ext,".gz") == 0 || strcmp(ext,".GZ") == 0)
			return (COMPRESSED_GZIP);
		else if ( strcmp(ext,".bzip2") == 0 || strcmp(ext,".bz2") == 0 || strcmp(ext,".BZIP2") == 0 || strcmp(ext,".BZ2") == 0)
			return (COMPRESSED_BZIP2);
		else if ( strcmp(ext,".xz") == 0 || strcmp(ext,".XZ") == 0)
			return (COMPRESSED_XZ);
		else
			return COMPRESSED_NO;
	}
}

void CloseStatusBar(void)
{
	fprintf(stderr,"\r|");
	int i;
	for(i=1;i<=10;i++)
		fprintf(stderr,"====");
	fprintf(stderr,"| 100%c",'%');
	fprintf(stderr,"\n");
	fflush(stderr);
}

void CloseSMPStatusBar(int nb_cpus,arg_threads_t *argt)
{
	int i;
	for(i=0;i<nb_cpus;i++)
		sprintf(argt->statusbar+i*4," 100");
	fprintf(stderr,"\r%%   %s",argt->statusbar);
	fprintf(stderr,"\n");
	fflush(stderr);
	free(argt->statusbar);
}

void InitSMPStatusBar(int nb_cpus,arg_threads_t *argt)
{
	int i;
	argt->statusbar=(char *)malloc(nb_cpus*5+10);
	for(i=0;i<(nb_cpus*4);i++)
		*(argt->statusbar+i)=' ',
	*(argt->statusbar+nb_cpus*4)='\0';
	fprintf(stderr,"CPU ");
	for(i=0;i<(nb_cpus);i++)
		fprintf(stderr,"%4d",i+1);
	fprintf(stderr,"\n");
	fprintf(stderr,"%%   %s",argt->statusbar);
	fflush(stderr);
	// je le fais dans le InitArrayOfThreadArgs if ( pthread_rwlock_init(argt->ptr_lock,NULL) != 0) Fatal("can't create rwlock");
}

unsigned long long SMPStatusBar(short cpu_id,unsigned long long done, unsigned long long max,arg_threads_t *argt)
{
	if ( max == 0 )
	{
		done=1;
		max=1;
	}
	unsigned long long decime=max/10;
	if ( decime == 0 )
		decime = 1;
	char txt[10]={0};
	int pc = ceil(done*100/max);
	if ( pc % 10 != 0 )
		pc++;
	if ( pc == 100 )
		sprintf(txt," %03d",pc);
	else
		sprintf(txt,"  %02d",pc);
	if ( pc < 0 )
	{
		fprintf(stderr,"ATTENTION : %d<<\n",pc);
		fflush(stderr);
	}
	int i;
	if (pthread_rwlock_wrlock(argt->ptr_lock) != 0) Fatal("can't get wrlock");
	for(i=0;i<4;i++)
		*(argt->statusbar+cpu_id*4+i)=txt[i];
	fprintf(stderr,"\r%%   %s",argt->statusbar);
	fflush(stderr);
	pthread_rwlock_unlock(argt->ptr_lock);
	return(decime);
}

unsigned long long StatusBar(unsigned long long done, unsigned long long max)
{
	int i;
	if ( max == 0 )
	{
		done=1;
		max=1;
	}
	fprintf(stderr,"\r|");
	unsigned long long decime=max/10;
	if ( decime == 0 )
		decime = 1;
	for(i=1;i<=10;i++)
	{
		if ( done >= decime*i )
			fprintf(stderr,"====");
		else
			fprintf(stderr,"    ");
	}
	fprintf(stderr,"| %llu%c",(unsigned long long)(done*100/max),'%');
	fflush(stderr);
	return(decime);
}


void CloseStream(char *file,FILE *f_in)
{
	/*
	fprintf(stderr,"CloseStreamIn:%s:\n",file);
	fflush(stderr);
	*/
	int ext_code = GetExtension(file);
	if ( ext_code == COMPRESSED_NO )
		fclose(f_in);
	else
		pclose(f_in);
}

FILE *GetStreamIn(char *file)
{
	/*
	fprintf(stderr,"GetStreamIn:%s:\n",file);
	fflush(stderr);
	*/
	FILE *f_in=NULL;
	char cmd[MAXFILENAME]="";
	int ext_code = GetExtension(file);
	if ( ext_code == COMPRESSED_NO )
	{
		f_in=(FILE *)fopen(file,"r");
	}
	else
	{
		switch(ext_code)
		{
			case(COMPRESSED_GZIP):
				sprintf(cmd,"gzip -cd %s",file);
				break;
			case(COMPRESSED_BZIP2):
				sprintf(cmd,"bzip2 -cd %s",file);
				break;
			case(COMPRESSED_XZ):
				sprintf(cmd,"xz -cd %s",file);
				break;
		}
		f_in=popen(cmd,"r"); // on ne recupere pas le code d'erreur du gzip
	}
	if ( f_in == NULL )
	{
	    fprintf(stderr,"GetStreamIn/FATAL error: unable to read file >%s< [with cmd >%s<]\n",file,cmd);
	    exit(EXIT_FAILURE);
	}
	return f_in;
}


FILE *GetStreamOut(char *file)
{
	//fprintf(stderr,"GetStreamOut:%s:\n",file);
	//fflush(stderr);
	FILE *f_out=NULL;
	char cmd[MAXFILENAME];
	int ext_code = GetExtension(file);
	if ( ext_code == COMPRESSED_NO )
	{
		f_out=(FILE *)fopen(file,"w");
	}
	else
	{
		switch(ext_code)
		{
			case(COMPRESSED_GZIP):
				sprintf(cmd,"gzip -c > %s",file);
				break;
			case(COMPRESSED_BZIP2):
				sprintf(cmd,"bzip2 -c > %s",file);
				break;
			case(COMPRESSED_XZ):
				sprintf(cmd,"xz -c > %s",file);
				break;
		}
		f_out=popen(cmd,"w");
	}
	if( f_out == NULL )
	{
	    fprintf(stderr,"GetStreamOut/FATAL error: unable to read file >%s<\n",file);
	    exit(EXIT_FAILURE);
	}
	return f_out;
}

void GetExePath(char *prg)
{
	fprintf(stderr,"GetExePath not implemented yet >%s<\n",prg);
}


int GetListOfFiles(char *pattern,char *lof)
{
char cmd[MAXFILENAME];
char line[MAXLINE];
int cpt=0;

	sprintf(cmd,"ls %s",pattern);
	FILE *f_ls=(FILE *)popen(cmd,"r");
	while(fgets(line,MAXLINE-1,f_ls) != NULL) 
	{
		lof=(char *)malloc(sizeof(char)*(strlen(line)+1));
		strcpy(lof,line);
		lof++;
		cpt++;
	}
	pclose(f_ls);
	return cpt;
}

int Split(char *str,int c,char *a_split[])
{
	char *str_start=str;
	int pstart=0;
	int i=0;
	int len=strlen(str);
	int len_substr=0;
	int cpt=0;
	while(i<len)
	{
		if( str[i] == c )
		{
			len_substr=i-pstart;
			if ( len_substr > 0 )
			{
				if ( a_split[cpt] != NULL )
				{
					free(a_split[cpt]);
					a_split[cpt]=NULL;
				}
				a_split[cpt]=(char *)malloc(len_substr+1);
				assert(a_split[cpt]!=NULL);
				strncpy(a_split[cpt],str_start,len_substr);
				a_split[cpt][len_substr]='\0';
				cpt++;
			}
			pstart=i+1;
			str_start=(str+i+1);
		}
		i++;
	}
	//i--;
	len_substr=i-pstart;
	if ( len_substr > 0 )
	{
		if ( a_split[cpt] != NULL )
		{
			free(a_split[cpt]);
			a_split[cpt]=NULL;
		}
		a_split[cpt]=(char *)malloc(len_substr+1); //
		strncpy(a_split[cpt],str_start,len_substr);
		a_split[cpt][len_substr]='\0';
	}
	return(cpt);
}

int StrEndWith(char *str,char *substr)
{
	if ( substr == NULL )
		return(FALSE);
	char *start = str;
	char *substr_search = NULL;
	while ((substr_search=strstr(start,substr)) != NULL )
	{
		if ( strlen(substr_search) == strlen(substr) )
			return(TRUE);
		start = substr_search+1;
		if ( *start == '\0' )
			return(FALSE);
	}
	return(FALSE);
}

char *StrRStr(char *str,char *substr)
{
	if ( substr == NULL )
		return(NULL);
	char *start = str;
	char *substr_search = NULL;
	int found = FALSE;
	while ( *start != 0 && (substr_search=strstr(start,substr)) != NULL )
	{
		found = TRUE;
		start = substr_search+1;
	}
	if ( found )
		return(start-1);
	else
		return(NULL);
}


int StrStartWith(char *str,char *substr)
{
	if ( substr == NULL )
		return(FALSE);
	
	char *substr_search = strstr(str,substr);
	
	if ( substr_search != NULL && substr_search == str )
		return(TRUE);
	return(FALSE);
}

char *GetTime(void)
{
	time_t now = time(NULL);
	struct tm *t = localtime(&now);
	char buf[1000]={0};
	sprintf(buf,"%04d%02d%02d %02d:%02d:%02d",t->tm_year+1900, t->tm_mon+1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec);
	char *time_str=(char *)malloc(sizeof(char)*(strlen(buf)+1));
	strcpy(time_str,buf);
	return(time_str);
}

void InitArrayOfThreadArgs(arg_threads_t *a_argt,arg_threads_t *argt,int nb_cpus)
{
	int i;
	
	argt->ptr_lock_self_array=(pthread_rwlock_t *)malloc(sizeof(pthread_rwlock_t));
	if ( pthread_rwlock_init(argt->ptr_lock_self_array,NULL) != 0) Fatal("can't create rwlock");
	
	argt->ptr_lock=(pthread_rwlock_t *)malloc(sizeof(pthread_rwlock_t));
	if ( pthread_rwlock_init(argt->ptr_lock,NULL) != 0) Fatal("can't create rwlock");
	
	
	for(i=0;i<nb_cpus;i++)
	{
		(a_argt+i)->imin=0;
		(a_argt+i)->imax=0;
		(a_argt+i)->ncpus = nb_cpus;
		(a_argt+i)->thread_id=-1;
		(a_argt+i)->running=-1;
		(a_argt+i)->self_array=a_argt;
		(a_argt+i)->ptr_lock_self_array=argt->ptr_lock_self_array;
		// les deux ci dessous sont initialises/utilises par InitSMPStatusBar qui fait une premiere initialisation de la structure de donnees (BOF!)
		(a_argt+i)->statusbar=argt->statusbar;
		(a_argt+i)->ptr_lock=argt->ptr_lock;  
	}
}

int Multithreading(void *(*OneThread)(void*),arg_threads_t *argt,unsigned long long maxjobs,int nb_cpus)
{
/* recuperation des parametres dans fonction appele
void *MyThread(void *argt)
{
	arg_threads_t *thr_data;
	thr_data = (arg_threads_t *)argt;
	fprintf(stderr,"imin=%ld, imax=%ld, thread_id=%d\n",thr_data->imin, thr_data->imax,thr_data->thread_id);
	fprintf(stderr,"%d TXT %s\n",thr_data->thread_id,(char *)thr_data->complex_arg[0]);
}
*/
	if ( nb_cpus < 1 )
		nb_cpus = 1;
	
	if ( maxjobs < 1 )
		return(FALSE);
	
	nb_cpus         = ( maxjobs < nb_cpus ) ? maxjobs : nb_cpus;
	unsigned long long step   = ceil(maxjobs/nb_cpus);
	if ( maxjobs % nb_cpus != 0 )
		step++;
	
	//	fprintf(stderr,"%d, %ld, %ld\n",nb_cpus,step,maxjobs);
	unsigned long long i;
	int t;
	int rc;
	void *status;
	
	pthread_t      a_threads[MAX_CPUS];
	pthread_attr_t attr;
	pthread_attr_init(&attr);
//	pthread_attr_setschedpolicy(&attr,SCHED_BATCH); je desactive car ca penalise trop en cas de concurrence > #core)
	pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_JOINABLE);
	
	arg_threads_t *a_argt=(arg_threads_t *)malloc(sizeof(arg_threads_t)*nb_cpus);
	InitArrayOfThreadArgs(a_argt,argt,nb_cpus);
	
	int complex_arg;
	
	for (i=0,t=0;i<maxjobs;i+=step,t++)
	{
		a_argt[t].imin=i;
		a_argt[t].imax=( maxjobs-1 < i+step-1 ) ? maxjobs-1 : i+step-1;
		a_argt[t].thread_id=t;
		a_argt[t].job_id=t;
		a_argt[t].running=1;
		// recuperation des arguments specifiques
		for(complex_arg=0;complex_arg<PTHREAD_MAX_COMPLEX_ARGS;complex_arg++)
			a_argt[t].complex_arg[complex_arg]=argt->complex_arg[complex_arg];
		
		if ( nb_cpus > 1 )
		{
			rc = pthread_create(&a_threads[t],&attr,OneThread,(void *)&a_argt[t]);
			if ( rc )
			{
				fprintf(stderr,"FATAL error in Multithreading, return code from pthread_create() is %d\n",rc);
				exit(MULTITHREADING_ERROR);
			}
		}
        else
			OneThread((void *)&a_argt[t]);
	}
		
	pthread_attr_destroy(&attr);
	for (i=0,t=0;i<maxjobs;i+=step,t++)
	{
		if ( nb_cpus <= 1 ) // bizarre, a sortir
			continue;
		rc = pthread_join(a_threads[t],&status);
		if ( rc )
		{
			fprintf(stderr,"FATAL error in Multithreading, return code from pthread_join() is %d\n",rc);
			exit(MULTITHREADING_ERROR);
		}
	}
	free(argt->ptr_lock);
	free(argt->ptr_lock_self_array);
	free(a_argt);
	//pthread_exit(NULL);
	return(TRUE);
}


// pas de slice, c'est le thread qui determinera ce qu'il fait: ex  if ( (i % nb_cpus ) == thr_data->thread_id )
int MultithreadingNoSlice(void *(*OneThread)(void*),arg_threads_t *argt,unsigned long long maxjobs,int nb_cpus)
{
	if ( nb_cpus < 1 )
		nb_cpus = 1;
	
	if ( maxjobs < 1 )
		return(FALSE);
	
	nb_cpus         = ( maxjobs < nb_cpus ) ? maxjobs : nb_cpus;
	
	int t;
	int rc;
	void *status;
	
	pthread_t      a_threads[MAX_CPUS];
	pthread_attr_t attr;
	
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_JOINABLE);
	
	arg_threads_t *a_argt=(arg_threads_t *)malloc(sizeof(arg_threads_t)*nb_cpus);
	InitArrayOfThreadArgs(a_argt,argt,nb_cpus);
	
	int complex_arg;
	
	for (t=0;t<nb_cpus;t++)
	{
		a_argt[t].imin=0;
		a_argt[t].imax=maxjobs-1;
		a_argt[t].thread_id=t;
		a_argt[t].job_id=t;
		a_argt[t].running=1;
		// recuperation des arguments specifiques
		for(complex_arg=0;complex_arg<PTHREAD_MAX_COMPLEX_ARGS;complex_arg++)
			a_argt[t].complex_arg[complex_arg]=argt->complex_arg[complex_arg];
		
		if ( nb_cpus > 1 )
		{
			rc = pthread_create(&a_threads[t],&attr,OneThread,(void *)&a_argt[t]);
			if ( rc )
			{
				fprintf(stderr,"FATAL error in Multithreading, return code from pthread_create() is %d\n",rc);
				exit(MULTITHREADING_ERROR);
			}
		}
        else
			OneThread((void *)&a_argt[t]);
	}
		
	pthread_attr_destroy(&attr);
	for (t=0;t<nb_cpus;t++)
	{
		if ( nb_cpus <= 1 ) // bizarre, a sortir
			continue;
		rc = pthread_join(a_threads[t],&status);
		if ( rc )
		{
			fprintf(stderr,"FATAL error in Multithreading, return code from pthread_join() is %d\n",rc);
			exit(MULTITHREADING_ERROR);
		}
	}
	free(argt->ptr_lock);
	free(argt->ptr_lock_self_array);
	free(a_argt);
	//pthread_exit(NULL);
	return(TRUE);
}

// pas de slice, c'est le thread qui determinera ce qu'il fait: ex  if ( (i % nb_cpus ) == thr_data->thread_id )
// je reutilise la structure a_argt si le param n'est pas null

void FreeArrayOfThreadArgsIfNeeded(arg_threads_t *argt,arg_threads_t *a_argt)
{
	free(argt->ptr_lock);
	free(argt->ptr_lock_self_array);
	free(a_argt);
}

int MultithreadingNoSliceReuse(void *(*OneThread)(void*),arg_threads_t *argt,unsigned long long maxjobs,int nb_cpus,arg_threads_t *a_argt)
{
	if ( nb_cpus < 1 )
		nb_cpus = 1;
	
	if ( maxjobs < 1 )
		return(FALSE);
	
	nb_cpus         = ( maxjobs < nb_cpus ) ? maxjobs : nb_cpus;
	
	int t;
	int rc;
	void *status;
	
	pthread_t      a_threads[MAX_CPUS];
	pthread_attr_t attr;
	
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_JOINABLE);
	
	if ( a_argt == NULL )
	{
		a_argt=(arg_threads_t *)malloc(sizeof(arg_threads_t)*nb_cpus);
		InitArrayOfThreadArgs(a_argt,argt,nb_cpus);
	}
	
	int complex_arg;
	
	for (t=0;t<nb_cpus;t++)
	{
		a_argt[t].imin=0;
		a_argt[t].imax=maxjobs-1;
		a_argt[t].thread_id=t;
		a_argt[t].job_id=t;
		a_argt[t].running=1;
		// recuperation des arguments specifiques
		for(complex_arg=0;complex_arg<PTHREAD_MAX_COMPLEX_ARGS;complex_arg++)
			a_argt[t].complex_arg[complex_arg]=argt->complex_arg[complex_arg];
		
		if ( nb_cpus > 1 )
		{
			rc = pthread_create(&a_threads[t],&attr,OneThread,(void *)&a_argt[t]);
			if ( rc )
			{
				fprintf(stderr,"FATAL error in Multithreading, return code from pthread_create() is %d\n",rc);
				exit(MULTITHREADING_ERROR);
			}
		}
        else
			OneThread((void *)&a_argt[t]);
	}
		
	pthread_attr_destroy(&attr);
	for (t=0;t<nb_cpus;t++)
	{
		if ( nb_cpus <= 1 ) // bizarre, a sortir
			continue;
		rc = pthread_join(a_threads[t],&status);
		if ( rc )
		{
			fprintf(stderr,"FATAL error in Multithreading, return code from pthread_join() is %d\n",rc);
			exit(MULTITHREADING_ERROR);
		}
	}
	/*
		free(argt->ptr_lock);
		free(argt->ptr_lock_self_array);
		free(a_argt);
	*/
	//pthread_exit(NULL);
	return(TRUE);
}

//
//
// ca rend incoherent le SMPStatus
//
int MultithreadingShareSliceWithHelper(void *(*OneThread)(void*),arg_threads_t *argt,unsigned long long imin, unsigned long long imax,
									   int currthread_id,int helperthread_id,int job_id)
{
	int nb_cpus = 2;
	arg_threads_t *a_argt = argt->self_array;
	if ( imin < 0 || ! (imin<=imax) )
	{
		fprintf(stderr,"Warning:  MultithreadingShareSliceWithHelper error: imin=%llu imax=%llu\n",imin,imax);
		SetThreadIdAsRunning(a_argt,helperthread_id);
		return(FALSE);
	}
//	fprintf(stderr,"MultithreadingShareSliceWithHelper: imin=%llu imax=%llu curr=%d, helper=%d, job_id=%d\n",imin,imax,currthread_id,helperthread_id,job_id);
	
	unsigned long long maxjobs = imax-imin+1;
	
	if ( maxjobs < nb_cpus )
	{
//		fprintf(stderr,"Warning:  Max jobs < 2 (%llu) imin=%llu imax=%llu\n",maxjobs,imin,imax);
		SetThreadIdAsRunning(a_argt,helperthread_id);
		return(FALSE);
	}
	
	unsigned long long step   = ceil(maxjobs/nb_cpus);
	if ( maxjobs % nb_cpus != 0 )
		step++;
	
	int t;
	int rc;
	void *status;
	
	pthread_t      a_threads[MAX_CPUS];
	pthread_attr_t attr;
	
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_JOINABLE);
	
//  cree par le niveau au dessus, ici je suis dans le helper
//	arg_threads_t *a_argt=(arg_threads_t *)malloc(sizeof(arg_threads_t)*nb_cpus);
//	InitArrayOfThreadArgs(a_argt,argt,nb_cpus);
	
	int complex_arg;
	
	unsigned long long i;
	
	// 0 ca sera le thread courant qui aura une tranche 2 fois moins grande
	// 1 ca sera le helper
	int first=FALSE;
	int ct=0;
	for (i=imin;i<=imax;i+=step,ct++)
	{
		first = ( i == imin ) ? TRUE : FALSE;
		t = (first) ? currthread_id:helperthread_id;	
		a_argt[t].imin=(first) ? imin : (imin+step);
		a_argt[t].imax=(first) ? (imin+step-1) : imax;
		a_argt[t].job_id=job_id;          // si initialise on peut partager une tranche d'un multithreading "superieur" entre sous tranches
//		fprintf(stderr,"MultithreadingShareSliceWithHelper : thread(%d) job_id(%d) imin=%llu imax=%llu\n",t,job_id,a_argt[t].imin,a_argt[t].imax);
// c'est fait dans la fonction qui donne l'helper
//		if ( ! first ) // l'actuel est running par definition
//			SetThreadIdAsRunning(a_argt,t);
		
		// recuperation des arguments specifiques
		for(complex_arg=0;complex_arg<PTHREAD_MAX_COMPLEX_ARGS;complex_arg++)
			a_argt[t].complex_arg[complex_arg]=argt->complex_arg[complex_arg];
		
		rc = pthread_create(&a_threads[ct],&attr,OneThread,(void *)&a_argt[t]);
		if ( rc )
		{
			fprintf(stderr,"FATAL error in Multithreading, return code from pthread_create() is %d\n",rc);
			exit(MULTITHREADING_ERROR);
		}
	}
		
	pthread_attr_destroy(&attr);
	for (t=0;t<nb_cpus;t++)
	{
		rc = pthread_join(a_threads[t],&status);
		if ( rc )
		{
			fprintf(stderr,"FATAL error in Multithreading, return code from pthread_join(), thread(%d) is %d\n",t,rc);
			exit(MULTITHREADING_ERROR);
		}
	}
// il ne faut pas faire les free lorsque l'on splitte le job
//	free(argt->ptr_lock);
//	free(argt->ptr_lock_self_array);
//	free(a_argt);
	//pthread_exit(NULL);
	return(TRUE);
}

// if ( GetFinishedThreadId(argt->self,2) >= 0 )

int GetFinishedThreadId(arg_threads_t *argt)
{
	int i;
	int found=-1;
	pthread_rwlock_t *lock=argt->ptr_lock_self_array; // je prends le premier car il est partage
	if (pthread_rwlock_wrlock(lock) != 0) Fatal("can't get wrlock");
	for(i=0;i<argt->ncpus&&found==-1;i++)
	{
		if ( (argt+i)->running == 0 )
		{
			found=i;
			(argt+i)->running++;
//			fprintf(stderr,"GetFinishedThreadId: found=%d\n",found);
		}
	}
	pthread_rwlock_unlock(lock);
	return found;
}

void SetThreadIdAsFinished(arg_threads_t *argt,int thread_id)
{
	if (pthread_rwlock_wrlock(argt->ptr_lock_self_array) != 0) Fatal("can't get wrlock");
//	fprintf(stderr,"SetThreadIdAsFinished: %d, running=%d\n",thread_id,(argt+thread_id)->running);
	if( (argt+thread_id)->running > 0 )
		(argt+thread_id)->running--;
	pthread_rwlock_unlock(argt->ptr_lock_self_array);
}

void SetThreadIdAsRunning(arg_threads_t *argt,int thread_id)
{
	if (pthread_rwlock_wrlock(argt->ptr_lock_self_array) != 0) Fatal("can't get wrlock");
	(argt+thread_id)->running++;
	pthread_rwlock_unlock(argt->ptr_lock_self_array);
}

char *GetListOfIP()
{
	struct ifaddrs *ifaddr,*ifa;
	char host[NI_MAXHOST]="";
	int family, n;
	char *list=NULL;
	char tmpstr[1000]="";
	if (getifaddrs(&ifaddr) == -1) 
	{
		perror("getifaddrs");
		exit(EXIT_FAILURE);
	}
	n=0;
	for(ifa=ifaddr;ifa!=NULL;ifa=ifa->ifa_next)
	{
		if( ifa->ifa_addr == NULL)
			continue;
		family = ifa->ifa_addr->sa_family;
		getnameinfo(ifa->ifa_addr, (family == AF_INET) ? sizeof(struct sockaddr_in) : sizeof(struct sockaddr_in6), host, NI_MAXHOST, NULL, 0, NI_NUMERICHOST);
		if ( strlen(host) == 0 )
			continue;
		n++;
		strcat(tmpstr,host);
		strcat(tmpstr," ");
	}
	list=(char *)malloc(strlen(tmpstr)+1);
	strcpy(list,tmpstr);
	freeifaddrs(ifaddr);
	return(list);
}

void CheckIPAuth(char lstok[])
{
	char *myip=GetListOfIP();
	
	char *a_myips[100];
	char *a_authips[100];
	int i;
	for(i=0;i<100;i++)
	{
		a_myips[i]=NULL;
		a_authips[i]=NULL;
	}
	Split(myip,' ',a_myips);
	Split(lstok,',',a_authips);
	/*
	for(i=0;i<10;i++)
		fprintf(stderr,"MY IP=%s\n",a_myips[i]);
	fprintf(stderr,"======\n");
	for(i=0;i<10;i++)
		fprintf(stderr,"AUTH IP=%s\n",a_authips[i]);
	*/
	
	int ok = FALSE;
	int j;
	for(i=0;i<100&&a_myips[i]!=NULL && ! ok;i++)
		for(j=0;j<100 && a_authips[j] != NULL && ! ok;j++)
		{
			if ( strcmp(a_authips[j],a_myips[i]) == 0 )
				ok=TRUE;
//			fprintf(stderr,"%s#%s#%d\n",a_authips[j],a_myips[i],ok);
		}
	
	FreeSplit(a_authips,100);
	FreeSplit(a_myips,100);
	if ( ok )
	{
//		fprintf(stderr,"%s OK\n",myip);
	}
	else
	{
		fprintf(stderr,"ERROR: unable to run the software\n");
		exit(EXIT_FAILURE);
	}
	free(myip);
}

// Specifique Linux

void InitMaxMemUsage(unsigned long long *max_mem)
{
	pid_t pid = getpid();
	char path[MAXFILENAME]="";
	char line[MAXLINE]="";
	sprintf(path,"/proc/%d/status",(int)pid);
	unsigned long long tmp_max_mem=0;
	FILE *f_status=fopen(path,"r");
	if ( f_status == NULL )
	{
		fprintf(stderr,"Warning: unable to get memory usage for pid: %d",(int)pid);
	}
	else
	{
		int done=FALSE;
		while(fgets(line,MAXLINE,f_status) != NULL && ! done)
		{
			// VmPeak: 112801144 kB
			if ( strstr(line,"VmPeak") )
			{
				sscanf(line,"%*s %llu",&tmp_max_mem);
//				fprintf(stderr,"%s\ntmp_max_mem %llu\n",line,tmp_max_mem);
				if ( tmp_max_mem > *max_mem )
					*max_mem = tmp_max_mem;
				break;
			}
		}
		fclose(f_status);
	}
}

void FreeSplit(char **a_strs,int nb)
{
	int r=0;
	for(r=0;r<nb;r++)
		if(a_strs[r]!=NULL)
		{
			free(a_strs[r]);
			a_strs[r]=NULL;
		}
}


void SetStackSize(int kbvalue)
{
	struct rlimit limit;
	int min_stack_size = kbvalue*1024;
	getrlimit (RLIMIT_STACK, &limit);
	if ( limit.rlim_cur < min_stack_size )
	{
		limit.rlim_cur=min_stack_size;
		if ( setrlimit(RLIMIT_STACK, &limit) != 0 )
		{
			fprintf(stderr,"Error, unable to set the stacksize limit to %d, do it and re run\n",min_stack_size);
			exit(EXIT_FAILURE);
		}
		getrlimit (RLIMIT_STACK, &limit);
	}
}

void *AlignedMallocOnPageSize(size_t size)
{
	int pg_size=sysconf ( _SC_PAGESIZE );
	void *ptr=NULL;
//	int error = posix_memalign(&ptr,pg_size,size);
	int error = posix_memalign(&ptr,64,size);
	if (error)
		Fatal("Unable to allocate via AlignedMallocOnPageSize\n");
	return(ptr);
}
