/*-------------------------------
 * 
 * Jerome.Gouzy@inra.fr
 * Created: 20140120
 * Copyright INRA, All rights reserved
 * 
 * $Id: read_booster.c 520 2016-12-02 12:35:31Z gouzy $
 * 
 --------------------------------*/

#include "read_booster.h"

UT_icd seqid_t_icd = {sizeof(seqid_t), NULL, NULL, NULL };

/* Rajout d'une nouvelle sequence si la profondeur est ok */

void InitBoostr(read_booster_t *read_booster,paramparser_t *o_param)
{
	char *outdir=NULL;
	read_booster->o_param = o_param;
	read_booster->current_nt_size=0;
	read_booster->nextid=1;
	read_booster->h_deja=NULL;
	read_booster->h_seqcontigs=NULL;
	read_booster->h_depth_of_contigpairs=NULL;
	read_booster->a_kmercode=NULL;
	read_booster->a_kmercode_selection=NULL;
	read_booster->klen        = (short) atoi(ParamParserGet(&o_param,"klen"));
	read_booster->nb_cpus     = (short) atoi(ParamParserGet(&o_param,"nb_threads"));
	read_booster->minkdepth   = (short) atoi(ParamParserGet(&o_param,"minkdepth")); // problematic kmers
	outdir = ParamParserGet(&o_param,"outdir");
	read_booster->outdir=(char *)malloc((strlen(outdir)+1));
	assert(read_booster->outdir != NULL);
	strcpy(read_booster->outdir,outdir);
	read_booster->paired_files       = ( ParamParserIsDefined(&o_param,"interleaved") )          ? FALSE: TRUE;
	read_booster->single_end         = FALSE;
	read_booster->debug              = ( ParamParserIsDefined(&o_param,"DEBUG") ) ? TRUE : FALSE;
	if ( ! ParamParserIsDefined(&o_param,"minimal_read_booster") )
	{
		sscanf(ParamParserGet(&o_param,"depth"),"%hu",&read_booster->depth);
	//	read_booster->depth       = (short) atoi(ParamParserGet(&o_param,"depth"));
		read_booster->minovl      = (short) atoi(ParamParserGet(&o_param,"minovl"));
		read_booster->maxmm       = (short) atoi(ParamParserGet(&o_param,"maxmm"));
		read_booster->minpci      = (float) atof(ParamParserGet(&o_param,"minpci"));
		read_booster->maxovh      = (short) atoi(ParamParserGet(&o_param,"maxovh"));
		read_booster->maxpaircov  = (short) atoi(ParamParserGet(&o_param,"maxpaircov"));
		read_booster->expected_nt_size= (unsigned int) (atof(ParamParserGet(&o_param,"size"))*1000000);
		read_booster->insert_size = (short) atoi(ParamParserGet(&o_param,"insert_size"));
		read_booster->compress    = (short) atoi(ParamParserGet(&o_param,"compress"));
	}
	if ( ParamParserIsDefined(&o_param,"minkmers") ) // 20161130: je rajoute le minkmers et maxkmers en param mais je veux garder la compat dans les prog qui ne le gere pas
	{
		read_booster->minkmers                  = (short) atoi(ParamParserGet(&o_param,"minkmers"));
		read_booster->maxkmers                  = (short) atoi(ParamParserGet(&o_param,"maxkmers"));
		read_booster->minkmers_for_lr_extension = (short) atoi(ParamParserGet(&o_param,"minkmers_for_lr_extension"));
	}
	else
	{
		if ( read_booster->klen < 14 )
		{
			read_booster->minkmers=5;
			read_booster->maxkmers=7;
			read_booster->minkmers_for_lr_extension=3;
		}
		else
		{
			read_booster->minkmers=4;
			read_booster->maxkmers=6;
			read_booster->minkmers_for_lr_extension=2;
	//	read_booster->maxkmers             = read_booster->max_readlen_used/read_booster->klen;
		}
	}
	read_booster->offsetk = ( read_booster->klen <= 16 )  ? 0 : read_booster->klen-15;
	read_booster->min_readlen_foruse   = (read_booster->klen <= 16) ? (read_booster->maxkmers * read_booster->klen) : (6*15+read_booster->offsetk);
	read_booster->map_high_freq_cutoff = 10000;
		
	if ( read_booster->klen > 29 )
	{
		fprintf(stderr,"FATAL ERROR: klen=%d must be <= 29\n",read_booster->klen);
		exit(EXIT_FAILURE);
	}
	read_booster->nrdb= ( ParamParserIsDefined(&o_param,"no-nrdb") ) ? FALSE: TRUE;
	if ( ParamParserIsDefined(&o_param,"minkmers") )
		read_booster->minkmers = (short) atoi(ParamParserGet(&o_param,"minkmers"));
	read_booster->lminkmers=read_booster->minkmers;
//	if ( ParamParserIsDefined(&o_param,"densifying_skeleton") )
	read_booster->lminkmers++; // desnification du skelete
	read_booster->stranded_rnaseq      = ( ParamParserIsDefined(&o_param,"stranded_rnaseq") )      ? TRUE : FALSE;
	read_booster->disable_optimization = ( ParamParserIsDefined(&o_param,"disable_optimization") ) ? TRUE : FALSE;
	if ( ! ParamParserIsDefined(&o_param,"minimal_read_booster") )
	{
		if (read_booster->debug)
			fprintf(stderr,"boost-r: klen=%d, minpci=%.2f, minovl=%d, maxovh=%d, maxmm=%d, minkdepth=%d, maxpaircov=%d, expected_size=%llu, minkmers=%d, maxkmers=%d, targeted depth=%d\n",read_booster->klen,read_booster->minpci,read_booster->minovl,read_booster->maxovh,read_booster->maxmm,read_booster->minkdepth, read_booster->maxpaircov, read_booster->expected_nt_size, read_booster->minkmers,read_booster->maxkmers,read_booster->depth);
		else
			fprintf(stderr,"boost-r: klen=%d, minpci=%.2f, minovl=%d, maxovh=%d, maxmm=%d, expected_size=%llu targeted depth=%d\n",read_booster->klen,read_booster->minpci,read_booster->minovl,read_booster->maxovh,read_booster->maxmm,read_booster->expected_nt_size,read_booster->depth);
	}
//	AssertKmerEncoding(read_booster->klen,code); pas bon si j'utilise klen > 16
	short maxk_for_alloc = Min(read_booster->klen,16);
	unsigned long long max = pow(4,maxk_for_alloc);
	struct sysinfo sysi;
	sysinfo(&sysi);
	if ( sysi.totalram < (sizeof(UT_array *)*max*3) )
	{
		fprintf(stderr,"Cararal memory error: your system cannot manage (RAM=%lu, required: ~ %llu * 3) such a klen value (klen=%d), use a lower value \n",(unsigned long)sysi.totalram,sizeof(UT_array *)*max*2,read_booster->klen);
		exit(0);
	}
	unsigned long long max_deja = pow(4,MAXK_FOR_DEJAVU);
	read_booster->a_deja=(UT_array **)malloc(sizeof(UT_array *)*max_deja);
	assert(read_booster->a_deja!=NULL);
	unsigned long long i;
	read_booster->a_kmer2contigs=NULL;
	if ( ! ParamParserIsDefined(&o_param,"minimal_read_booster") )
	{
		read_booster->a_kmer2contigs=(UT_array **)malloc(sizeof(UT_array *)*max);
		assert(read_booster->a_kmer2contigs!=NULL);
	//	fprintf(stderr,"max kmers=%llu\n",max);
		for(i=0;i<max;i++)
			read_booster->a_kmer2contigs[i]=NULL;
	}
	for(i=0;i<max_deja;i++)
		read_booster->a_deja[i]=NULL;
	// masque pour initialiser le code (il faut que j'enleve le nucleotide de gauche)
	// je veux conserver les klen-1 nucleotide de droite (==> je mets des 11 sur ce que je veux garder)
	read_booster->kmercode_mask=( read_booster->klen <= 16 ) ? pow(4,read_booster->klen-1)-1 : pow(4,15-1)-1;
//	fprintf(stderr,"INIT a_kmer2contigs done, mask=%u\n",read_booster->kmercode_mask);
	// lock pour Multithreading
	if ( pthread_rwlock_init(&read_booster->lock_process_done,NULL) != 0)            Fatal("can't create rwlock");
	if ( pthread_rwlock_init(&read_booster->lock_seqcontigs,NULL) != 0)              Fatal("can't create rwlock");
	if ( pthread_rwlock_init(&read_booster->lock_h_deja,NULL) != 0)                  Fatal("can't create rwlock");
	if ( pthread_rwlock_init(&read_booster->lock_h_depth_of_contigpairs,NULL) != 0)  Fatal("can't create rwlock");
	for(i=0;i<MAX_THREAD_LOCKS;i++)
	{
		if ( pthread_rwlock_init(&read_booster->a_lock_a_kmer2contigs[i],NULL) != 0) Fatal("can't create rwlock");
		if ( pthread_rwlock_init(&read_booster->a_lock_a_deja[i],NULL) != 0) Fatal("can't create rwlock");
	}
	for(i=0;i<MAX_CPUS;i++)
		if ( pthread_rwlock_init(&read_booster->a_lock_current_nt_size_per_threads[i],NULL) != 0) Fatal("can't create rwlock");
	read_booster->process_done=FALSE;
	for(i=0;i<MAX_CPUS;i++)
		read_booster->current_nt_size_per_threads[i]=0;
	for(i=0;i<MAX_CPUS;i++)
	{
		read_booster->nextid_per_threads_curr[i]=0;
		read_booster->nextid_per_threads_max[i]=0;
	}
	// defini le moment ou l'on verifie que la couverture n'est pas atteinte ou l'ID suivant
	// il faut eviter de le faire trop souvent => 1x / nb_cpus / 100
	// chaque thread va le faire 100x par x analyse
	if ( ! ParamParserIsDefined(&o_param,"minimal_read_booster") )
	{
		read_booster->step_thread = Max(SYNC_STEP,ceil((read_booster->expected_nt_size/read_booster->nb_cpus)/100));
		fprintf(stderr,"%d\n",read_booster->step_thread);
		read_booster->pairdepth=(char *)malloc(UINT_MAX);
		assert(read_booster->pairdepth!=NULL);
		for(i=0;i<UINT_MAX;i++) // il y a un cas particulier a UINT_MAX que je dois eviter, je n'arrive pas a allouer plus que UINT_MAX (cf size_t)
			read_booster->pairdepth[i]=0;
		assert(read_booster->pairdepth);
		read_booster->mask_pairdepth=pow(2,(sizeof(seqid_t)*8)/2)-1;
	}
	read_booster->maxdepth_attempts=MAXDEPTH_ATTEMPTS;
	read_booster->max_seqlen=0;
}

/* suppression des structures de donnees non essentielles */

void DeleteDepth(depth_t **h_depth_of_contigpairs,int debug)
{
//	fprintf(stderr,"DeleteDepth\n");
	depth_t *onedepth=NULL;
	depth_t *tmpdepth=NULL;
	int depth_cpt=0;
	int depth_count=0;
	int depth_max=0;
	HASH_ITER(hh,*h_depth_of_contigpairs,onedepth,tmpdepth)
	{
		HASH_DEL(*h_depth_of_contigpairs,onedepth);
		depth_cpt++;
		depth_count+=onedepth->count;
		if ( onedepth->count > depth_max )
		{
			depth_max = onedepth->count;
		}
		free(onedepth->id);
		free(onedepth);
	}
}

void DeleteSeqCount(seqcount_t **h_depth_of_contigpairs,int debug)
{
//	fprintf(stderr,"DeleteDepth\n");
	seqcount_t *onedepth=NULL;
	seqcount_t *tmpdepth=NULL;
	HASH_ITER(hh,*h_depth_of_contigpairs,onedepth,tmpdepth)
	{
		HASH_DEL(*h_depth_of_contigpairs,onedepth);
		free(onedepth->filename);
		free(onedepth);
	}
}

void DeleteDeja(dejavu_t **h_deja)
{
//	fprintf(stderr,"DeleteDeja\n");
	dejavu_t *onedeja=NULL;
	dejavu_t *tmpdeja=NULL;
	HASH_ITER(hh,*h_deja,onedeja,tmpdeja)
	{
		HASH_DEL(*h_deja,onedeja);
		free(onedeja->str);
		free(onedeja);
	}
}

// pas utilise mais cela devrait

void DeleteDejaCode(dejavucode_t **h_deja)
{
	dejavucode_t *onedeja=NULL;
	dejavucode_t *tmpdeja=NULL;
	HASH_ITER(hh,*h_deja,onedeja,tmpdeja)
	{
		HASH_DEL(*h_deja,onedeja);
		free(onedeja);
	}
}

void DeleteDejaArray(read_booster_t *read_booster)
{
	kmercode_t i;
	unsigned long long max_deja=pow(4,MAXK_FOR_DEJAVU);
	for(i=0;i<max_deja;i++)
	{
		if ( read_booster->a_deja[i] != NULL )
		{
			utarray_free(read_booster->a_deja[i]);
			read_booster->a_deja[i]=NULL;
		}
	}
}
		//DelSeqLinks(read_booster->h_seqcontigs,read_booster->nb_max_seq);

void CleanUpBoostr(read_booster_t *read_booster,int mode)
{
	size_t seq_size;
	if ( read_booster->debug )
	{	
		if ( read_booster->h_depth_of_contigpairs != NULL )
		{
			seq_size = HASH_OVERHEAD(hh,read_booster->h_depth_of_contigpairs);
			fprintf(stderr,"OVERHEAD h_depth_of_contigpairs: %lu\n",seq_size);
		}
	    if ( read_booster->h_deja != NULL )
		{
			seq_size = HASH_OVERHEAD(hh,read_booster->h_deja);
			fprintf(stderr,"OVERHEAD h_deja: %lu\n",seq_size);
		}
	}
			
	if ( mode == CONTIGER_CLEAN_SEQOBS )
	{
		DeleteSequenceSet(&read_booster->h_seqcontigs,read_booster->nb_max_seq,TRUE);
	}
	if ( mode == CONTIGER_CLEAN_ALL || mode == CONTIGER_CLEAN_TMP )
	{
		if ( read_booster->pairdepth != NULL )
		{
			free(read_booster->pairdepth);
			read_booster->pairdepth=NULL;
		}
		DeleteDepth(&read_booster->h_depth_of_contigpairs,read_booster->debug);
		DeleteDeja(&read_booster->h_deja);
		DeleteDejaArray(read_booster);

	}
	
	if ( mode == CONTIGER_CLEAN_ALL )
	{
		DeleteSequenceSet(&read_booster->h_seqcontigs,read_booster->nb_max_seq,FALSE);
		if ( read_booster->a_kmer2contigs != NULL )
		{
			ResetKmers(read_booster->a_kmer2contigs,read_booster->klen);
			free(read_booster->a_kmer2contigs);
			read_booster->a_kmer2contigs=NULL;
		}
		free(read_booster->a_kmercode);
		read_booster->a_kmercode=NULL;
		free(read_booster->outdir);
		free(read_booster->a_deja);
		read_booster->a_deja=NULL;
#ifdef REVMAP
		if ( read_booster->a_kmercode_selection != NULL )
		{
			free(read_booster->a_kmercode_selection);
			read_booster->a_kmercode_selection=NULL;
		}
#endif
		
	}
}


/* 
 * =============================================
 * =============================================
 * 
 */

int  AddIfNotCodeDejaVu(dejavucode_t **h_deja,kmercode_t kmercode,short klen,int revcomp)
{
	dejavucode_t *deja=NULL;
	if ( revcomp )
	{
		kmercode_t rckmercode;
		RevCompCode(kmercode,klen,&rckmercode);
		kmercode=rckmercode;
	}
	HASH_FIND(hh, *h_deja, &kmercode, sizeof(kmercode_t), deja);
		
	if ( deja )
		return FALSE;
			
	deja=(dejavucode_t *)malloc(sizeof(dejavucode_t));
	assert(deja!=NULL);
	deja->code=kmercode;
	HASH_ADD(hh,*h_deja,code,sizeof(kmercode_t),deja);
	return TRUE;
}

/*
#ifndef BOOST_R_H

int  AddIfNotDejaVu(dejavu_t **h_deja,char *str,int do_digest,int revcomp,int noadd,pthread_rwlock_t *lock)
{
	fprintf(stderr,"je passe\n");
}

#else
*/

int  AddIfNotDejaVu(dejavu_t **h_deja,char *str,int do_digest,int revcomp,int noadd,pthread_rwlock_t *lock)
{
	dejavu_t *deja=NULL;
	char *what=NULL;
	char digest[33]={0};
	if (pthread_rwlock_rdlock(lock) != 0) Fatal("can't get rdlock");
	if ( do_digest )
	{
		DigestMD5(str,digest);
	    HASH_FIND_STR(*h_deja, digest, deja);
		if ( deja )
		{
		//	fprintf(stderr,"0 F %s %s %s\n",str,digest,deja->str);
			pthread_rwlock_unlock(lock);
			return FALSE;
		}
		if ( revcomp )
		{
			char *revcompseq= RevCompSeq(str);
			assert(revcompseq);
			DigestMD5(revcompseq,digest);
			free(revcompseq);
		    HASH_FIND_STR(*h_deja, digest, deja);
		}
		what = digest;
	}
	else
	{
	    HASH_FIND_STR(*h_deja, str, deja);
		what = str;
	}
	pthread_rwlock_unlock(lock);
	if ( deja )
		return FALSE;
	
	if ( noadd )
		return TRUE;
	
	deja=(dejavu_t *)malloc(sizeof(dejavu_t));
	assert(deja!=NULL);
	deja->str=(char *)malloc((strlen(what)+1));
	assert(deja->str!=NULL);
    strcpy(deja->str,what);  
	if (pthread_rwlock_wrlock(lock) != 0) Fatal("can't get wrlock");
    HASH_ADD_KEYPTR(hh,*h_deja,deja->str,strlen(deja->str),deja);
	pthread_rwlock_unlock(lock);
	return TRUE;
}

/*
#endif
*/


seqid_t AddIfDepthOk(read_booster_t *read_booster,char *read1,char *read2,seqid_t ctg1ns,seqid_t ctg2ns,int max_depth,int kmerise,short *used,int thr_id)
{
	seqid_t ctgid=0;
	if ( UpdatePairDepth(read_booster,ctg1ns,ctg2ns) )
	{
		ctgid = AddNewContigSeqDejaArray(read_booster,read1,read2,kmerise,0,thr_id); /* pas kamerize donc on ne pourra pas mapper */
		if ( ctgid )
			*used=TRUE;
	}
	return ctgid;
}


int  AddPairOfReads(read_booster_t *read_booster,char *read1,char *read2,int mode,short max_depth,short *used,int thr_id)
{
	if ( strlen(read1)/read_booster->klen < read_booster->minkmers )
		return FALSE;
	
	if ( strlen(read2)/read_booster->klen < read_booster->minkmers )
		return FALSE;
	
	// c'est con de le faire a chaque lecture
	short local_minkmers = ( mode == ONLY_UNMAPPED ) ? read_booster->lminkmers : read_booster->minkmers;
	short mapping_score = ( mode == ONLY_UNMAPPED_PAIRS || mode == INCREASE_DEPTH_PAIRS ) ? ALL_SCORE : BEST_SCORE;
	
	hits_t *rh_hits1=NULL;
	Map(read_booster,read1,&rh_hits1,local_minkmers,mapping_score,0,FALSE,FALSE);
	
	/* inutile de mapper le second si cela foire deja sur le premier */
	if(  mode == ONLY_UNMAPPED )
	{
		if ( rh_hits1 != NULL )
		{	
			FreeHits(&rh_hits1);
			return FALSE;
		}
	}
	else if  ( mode == INCREASE_DEPTH && rh_hits1 == NULL )
	{
		FreeHits(&rh_hits1);
		return FALSE;
	}
	
	seqid_t ctgid;
	seqid_t ctg1ns;
	seqid_t ctg2ns;
//	char ctgpair[200];
	hits_t *curr_hits1=NULL;
	hits_t *curr_hits2=NULL;
	hits_t *tmp_hits1=NULL;
	hits_t *tmp_hits2=NULL;
	hits_t *rh_hits2 = NULL;
	
	Map(read_booster,read2,&rh_hits2,local_minkmers,mapping_score,0,FALSE,FALSE);
	
	if ( mode == ONLY_UNMAPPED )
	{
		if ( rh_hits1 == NULL && rh_hits2 == NULL )
		{
			AddNewContigSeqDejaArray(read_booster,read1,read2,TRUE,0,thr_id);
			*used=TRUE;
		}
	}
	else if ( mode == INCREASE_DEPTH )
	{
		if ( rh_hits1  && rh_hits2 )
		{
			/* print STDERR "read1 et read2 mappes\n"; */
			ctgid = 0;
			// je rajoute le tri mais je ne traite plus que les read_booster->maxdepth_attempts meilleurs
			HASH_SORT(rh_hits1,HitSortByCount);
			HASH_SORT(rh_hits2,HitSortByCount);
			int maxi=0;
			short maxdepth_attempts=read_booster->maxdepth_attempts;
			HASH_ITER(hh,rh_hits1,curr_hits1,tmp_hits1)
			{
				int maxj=0;
				HASH_ITER(hh,rh_hits2,curr_hits2,tmp_hits2)
				{
					if ( curr_hits1->id == curr_hits2->id )
						continue;
					
					// je pourrais mettre un nombre max de controle de profondeur atteinte
					// mais cela pourrait gener lorsqu'il peut y avoir collision
					// je tente
					
					if ( ctgid ) // sequence deja rajoute, je mets a jour l'info de couverture de la paire
						UpdatePairDepth(read_booster,curr_hits1->id,curr_hits2->id);
					else
					{
						ctgid = AddIfDepthOk(read_booster,read1,read2,curr_hits1->id,curr_hits2->id,max_depth,FALSE,used,thr_id);
						if ( ctgid == 0 && maxj++ == maxdepth_attempts )
							break;
					}
				}
				if ( ctgid == 0 && maxi++ == maxdepth_attempts )
					break;
			}
		}
	}
	else if ( mode == ONLY_UNMAPPED_PAIRS )
	{
		// je rajoute si je n'arrive pas a m'ancrer sur une paire existante
		HASH_ITER(hh,rh_hits1,curr_hits1,tmp_hits1)
		{
			ctg1ns = curr_hits1->id;
			HASH_ITER(hh,rh_hits2,curr_hits2,tmp_hits2)
			{
				ctg2ns = curr_hits2->id;
				if ( ctg1ns == ctg2ns)
				{
					FreeHits(&rh_hits1);
					FreeHits(&rh_hits2);
					return(0);
				}
			}
		}
		AddNewContigSeqDejaArray(read_booster,read1,read2,TRUE,0,thr_id);
		*used=TRUE;
	}
	FreeHits(&rh_hits1);
	FreeHits(&rh_hits2);
	return(FALSE);
}

int UpdatePairDepth(read_booster_t *read_booster,seqid_t ctg1ns, seqid_t ctg2ns)
{
	if ( ctg1ns > ctg2ns )
	{
		seqid_t tmp;
		tmp = ctg2ns;
		ctg2ns = ctg1ns;
		ctg1ns = tmp;
	}
	seqid_t ctg1ns_mask = ctg1ns & read_booster->mask_pairdepth;
	seqid_t ctg2ns_mask = ctg2ns & read_booster->mask_pairdepth;
//	fprintf(stderr,"1: %u, %u  2: %u, %u\n",ctg1ns,ctg1ns_mask,ctg2ns,ctg2ns_mask);
	int offset = sizeof(seqid_t)*8/2;
	seqid_t paircode=ctg1ns_mask;
	paircode<<=offset;
//	fprintf(stderr,"1: %u, %u  2: %u, %u code=%u, offset=%i\n",ctg1ns,ctg1ns_mask,ctg2ns,ctg2ns_mask,paircode,offset);
	paircode+=ctg2ns_mask;
	if ( paircode == UINT_MAX ) // cas particulier de collision (je n'ai pas pu allouer plus de UINT_MAX slots a cause de size_t=uint, tant pis)
		paircode--;
	int retcode=FALSE;
// j'utilise a_lock_a_kmer2contigs car il n'est pas utilise ailleur en meme temps
	pthread_rwlock_t *lock=&read_booster->a_lock_a_kmer2contigs[paircode%MAX_THREAD_LOCKS];  // mettre dans le if ?
	if (pthread_rwlock_wrlock(lock) != 0) Fatal("can't get wrlock"); // idem ?
	if ( read_booster->pairdepth[paircode] < read_booster->maxpaircov )
	{
//		fprintf(stderr,"OK paircode=%u, count=%d max=%d\n",paircode,read_booster->pairdepth[paircode],read_booster->maxpaircov);
		read_booster->pairdepth[paircode]++;
		retcode=TRUE;
	}
	pthread_rwlock_unlock(lock);
	return(retcode);
}

int UpdatePairDepthHash(read_booster_t *read_booster,char *ctgpair)
{
depth_t *curr_depth;

	if (pthread_rwlock_wrlock(&read_booster->lock_h_depth_of_contigpairs) != 0) Fatal("can't get wrlock");
	HASH_FIND_STR(read_booster->h_depth_of_contigpairs,ctgpair,curr_depth);
	if ( ! curr_depth )
	{
		curr_depth=(depth_t *) malloc(sizeof(depth_t));
		assert(curr_depth!=NULL);
		curr_depth->count=1;
		int lenid=strlen(ctgpair);
		curr_depth->id=(char *)malloc((lenid+1));
		assert(curr_depth->id!=NULL);
		strcpy(curr_depth->id,ctgpair);
		HASH_ADD_KEYPTR(hh,read_booster->h_depth_of_contigpairs,curr_depth->id,lenid,curr_depth);
	}
	// je pense que je peux deverouiller avant car je ne change plus la structure du hash
	pthread_rwlock_unlock(&read_booster->lock_h_depth_of_contigpairs);
	if ( curr_depth->count < read_booster->maxpaircov ) //  la profondeur de liens entre paires
	{
		curr_depth->count++;
		return(TRUE);
	}
	return(FALSE);
}

int IsSequenceDejaVu(read_booster_t *read_booster,char *seq,int what,kmercode_t *kmercode) 
{
	int len = strlen(seq);
	seqid_t *p_seqid;
	int i = len-MAXK_FOR_DEJAVU;
	char str[50]={0};
	*kmercode=0;
	if( what == LR ) // je cree le code sur le debut et la fin pour avoir plus de diversite
	{
		strncpy(str,seq,5);
		strcat(str,seq+i+5);
		if ( strlen(str) != MAXK_FOR_DEJAVU )
		{
			fprintf(stderr,"\nWarning: STR=%s# seq=%s#\n",str,seq);
			*kmercode=0;
			return(TRUE);
		}
	}
	else
		strcpy(str,seq+i);
	if ( ! KmerEncode(str,MAXK_FOR_DEJAVU,kmercode) ) // en principe ne doit pas arriver
		return FALSE;
	
	if ( read_booster->a_deja[*kmercode] == NULL )
		return FALSE;
	
//	fprintf(stderr,"\n%s=%u;%d############\n",str,*kmercode,*kmercode%MAX_THREAD_LOCKS);
	pthread_rwlock_t *kmercodelock=&read_booster->a_lock_a_deja[*kmercode%MAX_THREAD_LOCKS];
	if ( pthread_rwlock_rdlock(kmercodelock) != 0) Fatal("can't get rdlock");
	// est ce qu'une sequence qui a le meme code est identique ?
	for(p_seqid=(seqid_t*)utarray_front(read_booster->a_deja[*kmercode]); p_seqid!=NULL; p_seqid=(seqid_t*)utarray_next(read_booster->a_deja[*kmercode],p_seqid)) 
	{	
		char *currseq = GetSeqHub(read_booster->h_seqcontigs,*p_seqid,what);
		if ( currseq == NULL )
		{
			fprintf(stderr,"\nIsSequenceDejaVu unexpected error (currseq vide): seqid=%u kmercode=%u\n",*p_seqid,*kmercode);
		}
		else
		{
			if ( len == strlen(currseq) && strcmp(seq,currseq) == 0 ) // je fais strlen d'abord car ca utilise sse et c'est rapide
			{
//					fprintf(stderr,"DEJA: seqid=%u, %u, %s\n",*p_seqid,kmercode,seq);
				pthread_rwlock_unlock(kmercodelock);
				return TRUE;
			}
		}
	}
	pthread_rwlock_unlock(kmercodelock);
	return FALSE;
}

// je change la facon de controler la presence de la sequence
// je calcule un code avec la fin de la sequence (extremite de la read2 ou il y a potentiellement le plus d'erreur
// puis je vais faire une comparaison de sequence avec toutes les sequences qui finissent avec ce meme code
seqid_t AddNewContigSeqDejaArray(read_booster_t *read_booster, char *read1, char *read2,int kmerize,seqid_t id,int thr_id)
{
	char *seq=MallocSeqPair(read1,read2);
	
	// je pourrais utiliser IsSequenceDejaVu mais bon je ne le fais pas
	int len = strlen(seq);
	kmercode_t kmercode;
	seqid_t *p_seqid;
	int i = len-MAXK_FOR_DEJAVU;
	if ( ! KmerEncode(seq+i,MAXK_FOR_DEJAVU,&kmercode) ) // en principe ne doit pas arriver
	{
		free(seq);
		return 0;
	}
	pthread_rwlock_t *kmercodelock=&read_booster->a_lock_a_deja[kmercode%MAX_THREAD_LOCKS];
	if ( read_booster->a_deja[kmercode] != NULL )
	{
		if ( pthread_rwlock_rdlock(kmercodelock) != 0) Fatal("can't get rdlock");
		// est ce qu'une sequence qui a le meme code est identique ?
		for(p_seqid=(seqid_t*)utarray_front(read_booster->a_deja[kmercode]); p_seqid!=NULL; p_seqid=(seqid_t*)utarray_next(read_booster->a_deja[kmercode],p_seqid)) 
		{	
			char *currseq = GetSeq(read_booster->h_seqcontigs,*p_seqid);
			if ( currseq == NULL )
			{
				fprintf(stderr,"\nAddNewContigSeqDejaArray unexpected error (currseq vide): seqid=%u kmercode=%u\n",*p_seqid,kmercode);
			}
			else
			{
				if ( len == strlen(currseq) && strcmp(seq,currseq) == 0 ) // je fais strlen d'abord car ca utilise sse et c'est rapide
				{
			//		fprintf(stderr,"DEJA: seqid=%u, %u, %s\n",*p_seqid,kmercode,seq);
					free(seq);
					pthread_rwlock_unlock(kmercodelock);
					return 0;
				}
			}
		}
		pthread_rwlock_unlock(kmercodelock);
	}
	seqid_t newid;
	if ( thr_id < 0 )
	{
		if (pthread_rwlock_wrlock(&read_booster->lock_seqcontigs) != 0) Fatal("can't get wrlock");
		newid = ( id ) ? id : GetNextContigSeqID(read_booster,1);
		pthread_rwlock_unlock(&read_booster->lock_seqcontigs);
		UpdateCurrentNtSize(read_booster,strlen(seq)-1);
	}
	else
	{
		// je ne locke plus j'ai une variable par thread
		newid = ( id ) ? id : GetNextContigSeqIDOneThread(read_booster,thr_id);
		UpdateCurrentNtSizeOneThread(read_booster,strlen(seq)-1,thr_id);
	}
	
	if ( pthread_rwlock_wrlock(kmercodelock) != 0) Fatal("can't get wrlock");
	if ( read_booster->a_deja[kmercode] == NULL )
		utarray_new(read_booster->a_deja[kmercode],&seqid_t_icd);
	utarray_push_back(read_booster->a_deja[kmercode],&newid);
	// 20140610 je le sort du lock pour limiter les blocages
	// 20140722 je l'y remets car ca arrive que j'en ai besoin et qu'il n'y soit pas
	SetSeq(read_booster->h_seqcontigs,newid,seq);
	pthread_rwlock_unlock(kmercodelock);
	
	if ( kmerize )
		KmerizeContigSeq(read_booster,newid,NULL,NULL,PAIRS);
	free(seq);
	
	return(newid);
}


int AddNewVeryLongRead(read_booster_t *read_booster,char *seq,seqid_t seqid)
{
	kmercode_t kmercode=0;
	if ( IsSequenceDejaVu(read_booster,seq,LR,&kmercode) )
		return(FALSE);
	
//	fprintf(stderr,"\nAddNew %u\n",kmercode);
	
	pthread_rwlock_t *kmercodelock=&read_booster->a_lock_a_deja[kmercode%MAX_THREAD_LOCKS];
	if ( pthread_rwlock_wrlock(kmercodelock) != 0) Fatal("can't get wrlock");
	if ( read_booster->a_deja[kmercode] == NULL )
		utarray_new(read_booster->a_deja[kmercode],&seqid_t_icd);
	utarray_push_back(read_booster->a_deja[kmercode],&seqid);
	// le SetSeqLR est fait dans la fonction appelante // SetSeqLR(read_booster->h_seqcontigs,seqid,seq);
	pthread_rwlock_unlock(kmercodelock);
	return(TRUE);
}

int  ReplaceProblematicPairs(read_booster_t *read_booster,char *read1,char *read2,short max_depth,short *used)
{
	char *seq=MallocSeqPair(read1,read2);
	
	kmercode_t kmercode=0;
	if ( IsSequenceDejaVu(read_booster,seq,PAIRS,&kmercode) )
	{
		free(seq);
		return(FALSE);
	}
	
	int seqlen = strlen(seq);
	short klen = read_booster->klen;
	int nb_kmers = seqlen/klen;
	if ( nb_kmers < (read_booster->minkmers*2) )
	{
		free(seq);
		return(FALSE);
	}
	
	/* avant de mapper je vais verifier que la nouvelle sequence ne contient pas de kmer rare */
	if ( ! IsSequenceConsistantWithKmers(read_booster,seq,read_booster->minkdepth) )
	{
		free(seq);
		return(FALSE);
	}
	
	hits_t *rh_hits=NULL;
	hits_t *onehit=NULL;
	hits_t *tmphit=NULL;
	Map(read_booster,seq,&rh_hits,read_booster->minkmers*2,BEST_SCORE,0,FALSE,FALSE); /* a voir si c'est bien BEST qu'il faut */
	//return if ( ! defined($rh_hits1) || scalar(keys(%$rh_hits1)) != 1 );
	//je capte pas la raison du 1 seul hit
	
	//if ( rh_hits == NULL || HASH_COUNT(rh_hits) != 1 )
	if ( rh_hits == NULL )
	{
		free(seq);
		return(FALSE);
	}
	
	seqid_t ctgns;
//	char ctgpair[25]; /* je ne garde pas les deux, ce n'est plus une paire */
	seqid_t ctgid=0;
	HASH_ITER(hh,rh_hits,onehit,tmphit)
	{
		ctgns = onehit->id;
//		ctgid  = AddIfDepthOk(read_booster,read1,read2,ctgpair,max_depth,FALSE,used,&stop_on_maxcoverage);
		ctgid  = AddIfDepthOk(read_booster,read1,read2,ctgns,ctgns,max_depth,FALSE,used,-1); // 20140308 je kmerize, 20140520, je ne kmerise plus
//		ctgid  = AddIfDepthOk(read_booster,read1,read2,ctgpair,max_depth,TRUE,used,&stop_on_maxcoverage); // tant que je debugge je fais comme avant donc je kmerise
		if ( ctgid ) // je pourrais mettre a jour la couverture pour tout le monde mais c'est du best et je prefere limiter les iterations
			break;
	}
	FreeHits(&rh_hits);
	free(seq);
	return(FALSE);
}

// ici la comparaison est faite par rapport a la couverture dans les read selectionnees IsKmerCompositionOK la fait dans la banque totale
// maintenant je recalcule la klib apres la selection, je dois pouvoir fusionner ca ca  permettrait d'eviter le len
// je faire presque commme  IsKmerCompositionOK mais dans IsKmerCompositionOK je gere le fait que les sequences sont bruitees et je rabote
int IsSequenceConsistantWithKmers(read_booster_t *read_booster,char *seq,short mindepth)
{
	int i;
	short klen = read_booster->klen;
	char *ptr_str=seq;
	int lmax = strlen(seq)-klen;
	kmercode_t kmercode;
	short init = FALSE;
//	short mindepth = read_booster->depth/3; // 20140515, avant read_booster->minkdepth
//	short mindepth = read_booster->minkdepth;
	for(i=0;i<=lmax;i++,ptr_str++)
	{	
		init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
		if ( ! init )
		{
			i+=(klen-1);
			ptr_str+=(klen-1);
			continue;
		}
// 20140526		if ( read_booster->a_kmer2contigs[kmercode] == NULL ||  utarray_len(read_booster->a_kmer2contigs[kmercode]) < mindepth )
		if ( read_booster->a_kmercode[kmercode] < mindepth )
			return(FALSE);
	}
	return(TRUE);
}

/*
 * 
 */

float GetQualScore(read_booster_t *read_booster,char *seq)
{
	if ( read_booster->a_kmercode == NULL )
		return(0);
	int i;
	short klen = read_booster->klen;
	char *ptr_str=seq;
	int lmax = strlen(seq)-klen;
	kmercode_t kmercode;
//	char kmerdec[50];
	//for(i=0;i<nb_kmers;i++,ptr_str+=klen)
	short init = FALSE;
	int sum=0;
	for(i=0;i<=lmax;i++,ptr_str++)
	{	
		init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
		if ( ! init )
		{
			i+=(klen-1);
			ptr_str+=(klen-1);
			continue;
		}
		sum += read_booster->a_kmercode[kmercode];
		//KmerDecode(kmercode,read_booster->klen,kmerdec);
		// fprintf(stderr,"%s %d %d %d\n",kmerdec,read_booster->a_kmercode[kmercode],sum,i);
	}
	float qualscore = (float)sum/(float)i;
	return(qualscore);
}


// le debut est un copier/coller de ReplaceProblematicPairs

int  RescuePairsWithFrequentsKmers(read_booster_t *read_booster,char *read1,char *read2,short max_depth,short *used)
{
	char *seq=MallocSeqPair(read1,read2);
	int seqlen = strlen(seq);
	kmercode_t kmercode=0;
	
	if ( IsSequenceDejaVu(read_booster,seq,PAIRS,&kmercode) )
	{
		free(seq);
		return(FALSE);
	}
	short klen = read_booster->klen;
	int nb_kmers = seqlen/klen;
	if ( nb_kmers < (read_booster->minkmers*2) )
	{
		free(seq);
		return(FALSE);
	}
	
	int i;
	char *ptr_str=seq;
	int lmax = seqlen-klen;
	kmercode=0;
	char kmer[100];
	seqid_t ctgid=0;
	//hits_t *rh_hits=NULL;
	short init = FALSE;
	//short minkmers = read_booster->minkmers-1;
	int count=0; // depuis que j'ai des klen > 16, il me faut 2 kmers rares pour declencher
	for(i=0;i<=lmax;i++,ptr_str++)
	{	
		init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
		if ( ! init )
		{
			i+=(klen-1);
			ptr_str+=(klen-1);
			continue;
		}
		
#ifndef REVMAP
		// je trouve une lecture avec un kmer oublie=presente dans le jeu de depart et absente ou pas suffisament presente 
		if ( read_booster->a_kmercode[kmercode] > read_booster->minkdepth && (read_booster->a_kmer2contigs[kmercode] == NULL || utarray_len(read_booster->a_kmer2contigs[kmercode]) < read_booster->minkdepth) ) 
#else
		if ( read_booster->a_kmercode[kmercode] > read_booster->minkdepth && read_booster->a_kmercode_selection[kmercode] < read_booster->minkdepth ) 
#endif
			count++;
		
		if ( count > 1 )
		{
			strncpy(kmer,ptr_str,klen);
			kmer[klen]='\0';
			if ( UpdatePairDepthHash(read_booster,kmer) ) // je ne veux pas faire des rescue que du meme kmer
				ctgid = AddNewContigSeqDejaArray(read_booster,read1,read2,TRUE,0,-1);
			else
				continue;
			
			if ( ctgid )
			{
				*used=1;
				ptr_str=seq;
				// je rend les kmer de la sequence que je garde de moins en moins problematiques
				short init = FALSE;
				for(i=0;i<=lmax;i++,ptr_str++)
				{	
					init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
					if ( ! init )
					{
						i+=(klen-1);
						ptr_str+=(klen-1);
						continue;
					}
//					read_booster->a_kmercode[kmercode] /= 2;
//20140611 je rajoute le lock, j'utilise le tableau de lock a_lock_a_deja
					if (pthread_rwlock_wrlock(&read_booster->a_lock_a_deja[kmercode%MAX_THREAD_LOCKS]) != 0) Fatal("can't get wrlock");
					read_booster->a_kmercode[kmercode] -= 2; // il y a un depth par kmer qui borne a maxpaircov
					// mais je ne dois pas descendre en desssous de minkdepth
					if ( read_booster->a_kmercode[kmercode] < read_booster->minkdepth )
						read_booster->a_kmercode[kmercode] = read_booster->minkdepth;
					pthread_rwlock_unlock(&read_booster->a_lock_a_deja[kmercode%MAX_THREAD_LOCKS]);
				}
			}
			// que la sequence soit dans la base ou que je vienne de la rajouter j'arrete
			free(seq);
			return(FALSE);
		}
	}
	free(seq);
	return(FALSE);
}


// la fonction renvoie le nombre de kmer utilisables
int Map(read_booster_t *read_booster,char *read,hits_t **h_hits,short minkmers,int type,seqid_t avoid, int no_high_freq,int for_assembly)
{
	UT_array **kmer2contigs = read_booster->a_kmer2contigs;
	
	hits_t *onehit    = NULL;
	hits_t *tmphit    = NULL;
	
	hits_t *h_firstlast = NULL;
	hits_t *firstlasthit = NULL;
	
 /* on recupere les contigs matchant avec la liste de kmer */

 	int hash_mode = FALSE;
 	kmercode_t kmercode,rckmercode;
	short klen = read_booster->klen;
	int lmax = strlen(read)-klen;
#ifndef REVMAP
	short offsetk = read_booster->offsetk;
#endif
	int 	i;
	char 	*ptr_str = read;
	seqid_t *p_seqid;
//	char kmerdec[100];
	char first_or_last_kmer=FALSE;
		
	UT_array*     a_fw=NULL;
	UT_array*     a_rv=NULL;
	
	if(  ! hash_mode )
	{
		utarray_new(a_fw,&seqid_t_icd);
		utarray_new(a_rv,&seqid_t_icd);
	}
	
	for_assembly = ( read_booster->maxovh != 0 ) ? FALSE : for_assembly; // je desactive l'optim si ton tolere les ovh
 	for_assembly = FALSE; 
	int usable_kmers=0;
	
	// depuis que j'ai des lectures de longueur variable je ne peux plus faire avec un pas de klen car si j'ai des N ca fout le bordel
	int init=FALSE;
	for(i=0;i<=lmax;i++,ptr_str++)
	{	
		// je fais tous les klen-offset, pas d'optim
#ifndef REVMAP
		init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,0);
#else
		init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
#endif
		if ( ! init ) 
			continue;
		
		first_or_last_kmer = ( i == 0 || i == lmax ) ? TRUE : FALSE;
		
	//	KmerDecode(kmercode,klen,kmerdec);
		if ( kmer2contigs[kmercode] != NULL )
		{
			if ( ! no_high_freq || read_booster->a_kmercode[kmercode] < read_booster->map_high_freq_cutoff )
			{
				usable_kmers++;
				// peut etre mettre un lock pour STEP1 qui Mappe en meme temps qu'il kmerize
				if( hash_mode )
				{
					for(p_seqid=(seqid_t*)utarray_front(kmer2contigs[kmercode]); p_seqid!=NULL; p_seqid=(seqid_t*)utarray_next(kmer2contigs[kmercode],p_seqid)) 
					{	
						if ( ! avoid || avoid != *p_seqid )
						{
							HitPlusPlus(h_hits,*p_seqid,'+');
							if ( for_assembly && first_or_last_kmer )
								HitPlusPlus(&h_firstlast,*p_seqid,'+');
						}
					}
				}
				else
					utarray_concat(a_fw,kmer2contigs[kmercode]);
			}
		} 
		if ( ! read_booster->stranded_rnaseq )
		{
			RevCompCode(kmercode,klen,&rckmercode);
			if ( kmer2contigs[rckmercode] != NULL )
			{
				if ( ! no_high_freq || read_booster->a_kmercode[rckmercode] < read_booster->map_high_freq_cutoff )
				{
					if ( hash_mode )
					{
						for(p_seqid=(seqid_t*)utarray_front(kmer2contigs[rckmercode]); p_seqid!=NULL; p_seqid=(seqid_t*)utarray_next(kmer2contigs[rckmercode],p_seqid)) 
						{
							if ( ! avoid || avoid != *p_seqid )
							{
								HitPlusPlus(h_hits,*p_seqid,'-');
								if ( for_assembly && first_or_last_kmer )
									HitPlusPlus(&h_firstlast,*p_seqid,'+');
							}
						}
					}
					else
						utarray_concat(a_rv,kmer2contigs[rckmercode]);
					usable_kmers++;
				}
			} 
		}
#ifndef REVMAP
		i+=(klen-1-offsetk);
		ptr_str+=(klen-1-offsetk);
#endif
	}
	short max=0;
	if (! hash_mode )
	{
		UpdateHits(a_fw,h_hits,'+',minkmers,&max,type,avoid);
		UpdateHits(a_rv,h_hits,'-',minkmers,&max,type,avoid);
	}
	
	HASH_SORT(*h_hits,HitSortByCount);
	int delete=FALSE;
	HASH_ITER(hh, *h_hits, onehit, tmphit) 
	{
		delete=FALSE;
		if ( onehit->count < minkmers || (avoid > 0 && onehit->id == avoid) )
			delete=TRUE;
		else if ( type==BEST_SCORE )
		{
			if ( onehit->count > max )
				max = onehit->count;
			else if (onehit->count < max )
				delete=TRUE;
		}
		if ( for_assembly && ! delete )
		{
			HASH_FIND(hh,h_firstlast,&onehit->id,sizeof(seqid_t),firstlasthit);
			if ( firstlasthit == NULL )
				delete=TRUE;
		}
		if ( delete )
		{
			HASH_DEL(*h_hits,onehit); 
			free(onehit);           
		}
	}
	if ( for_assembly )
		FreeHits(&h_firstlast);
	int count = HASH_COUNT(*h_hits);
	if ( count == 0 )
		h_hits = NULL; 
	return(usable_kmers);
}

int SortArrayBySeqid(const void *_a, const void *_b) 
{
	seqid_t a = *(seqid_t*)_a;
	seqid_t b = *(seqid_t*)_b;
	return a - b;
}

int UpdateHits(UT_array *a_lst,hits_t **rh_hits,char strand,short minkmers,short *max,int type,seqid_t avoid)
{
short count=0;
	// j'edite et utilise max a la volee mais ce n'est pas fini, c'est le 2e passage dans le HASH_ITER qui regle

	seqid_t *p_seqid;
	if ( a_lst != NULL && utarray_len(a_lst) >= minkmers )
	{
		utarray_sort(a_lst,SortArrayBySeqid);
		seqid_t previd = 0;
		for(p_seqid=(seqid_t*)utarray_front(a_lst); p_seqid!=NULL; p_seqid=(seqid_t*)utarray_next(a_lst,p_seqid))
		{
			if ( *p_seqid != previd )
			// ATTENTION, est ce qu'il ne faudrait pas mettre if ( previd && *p_seqid != previd )
			{
				if ( avoid == 0 || avoid != previd )
				{
					if ( count >= minkmers && (type == ALL_SCORE || count >= *max) )
						HitSetCount(rh_hits,previd,strand,count);
					if ( count > *max )
						*max = count;
				}
				count = 0;
			}
			previd = *p_seqid;
			count++;
		}
		if ( avoid == 0 || avoid != previd )
		{
			if ( count >= minkmers && (type == ALL_SCORE || count >= *max) )
				HitSetCount(rh_hits,previd,strand,count);
			if ( count > *max )
				*max = count;
		}
	}
	utarray_free(a_lst);
	return *max;
}


int KmerizeContigSeq(read_booster_t *read_booster,seqid_t id,UT_array **kmer2ctg_param,sequenceset_t **seqcontigs_param,int what)
{
	UT_array **kmer2contigs = ( kmer2ctg_param   != NULL ) ? kmer2ctg_param  : read_booster->a_kmer2contigs;
	sequenceset_t  **seqset = ( seqcontigs_param != NULL ) ? seqcontigs_param: read_booster->h_seqcontigs;
	
	char *seq;
// je ne l'utilise pas lorsque j'edite, pas besoin de locker	if (pthread_rwlock_rdlock(&read_booster->lock_seqcontigs) != 0) Fatal("can't get wrlock");
	seq = GetSeqHub(seqset,id,what);
//	pthread_rwlock_unlock(&read_booster->lock_seqcontigs);
	if (seq == NULL )
		return(0);
	
	dejavucode_t *h_kmers=NULL;
	dejavucode_t *deja_curr=NULL;
	dejavucode_t *deja_tmp=NULL;
	kmercode_t kmercode;
	
	short klen = read_booster->klen;
	int   lmax = strlen(seq)-klen;
	int   i;
	char *ptr_str=seq;
#ifdef REVMAP
	short offsetk = read_booster->offsetk;
#endif
	
	short init = FALSE;
	for(i=0;i<=lmax;i++,ptr_str++)
	{	
#ifdef REVMAP
		init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,0);
#else
		init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
#endif
		if ( ! init )
		{
			i+=(klen-1);
			ptr_str+=(klen-1);
			continue;
		}
	  	AddIfNotCodeDejaVu(&h_kmers,kmercode,klen,FALSE);
#ifdef REVMAP
		i+=(klen-1-offsetk);
		ptr_str+=(klen-1-offsetk);
#endif
	}
	HASH_ITER(hh,h_kmers,deja_curr,deja_tmp)
	{
		PushContigIntoKmerHash(read_booster,kmer2contigs,deja_curr->code,id);
		HASH_DEL(h_kmers,deja_curr);
		free(deja_curr);
	}
	return (lmax+klen);
} 


void ResetKmerize(read_booster_t *read_booster,int what)
{
	ResetKmers(read_booster->a_kmer2contigs,read_booster->klen);
	if ( what == PAIRS )
		read_booster->current_nt_size=0;
	seqid_t maxid = read_booster->nextid;
	unsigned long long current_nt_size=0;
	unsigned long long len=0;
	seqid_t num=0;
	
	unsigned long long a_len[MAX_THREAD_LOCKS];
	seqid_t a_num[MAX_THREAD_LOCKS];
	seqid_t a_maxseqlen[MAX_THREAD_LOCKS];
	int i;
	for(i=0;i<read_booster->nb_cpus;i++)
	{
		a_len[i]=0;
		a_num[i]=0;
		a_maxseqlen[i]=0;
	}
	
	arg_threads_t argt;
	argt.complex_arg[0]=read_booster;
	argt.complex_arg[1]=a_len;
	argt.complex_arg[2]=a_num;
	argt.complex_arg[3]=&what;
	argt.complex_arg[4]=a_maxseqlen;
	
	InitSMPStatusBar(read_booster->nb_cpus,&argt);
	Multithreading(ResetKmerizeSlice,&argt,maxid,Min(16,read_booster->nb_cpus));
	CloseSMPStatusBar(read_booster->nb_cpus,&argt);
	
	for(i=0;i<read_booster->nb_cpus;i++)
	{
		len+=a_len[i];
		num+=a_num[i];
		if ( a_maxseqlen[i] > read_booster->max_seqlen )
			read_booster->max_seqlen=a_maxseqlen[i];
	}
	current_nt_size=len;
		
	if ( read_booster->debug )
		fprintf(stderr,"Reset done # sequences=%u, size=%llu nt, coverage=%.2Lf\n",num,current_nt_size,(long double)current_nt_size/(long double)read_booster->expected_nt_size);
	if ( what == PAIRS )
		UpdateCurrentNtSize(read_booster,current_nt_size);
}

void *ResetKmerizeSlice(void *argt)
{
	arg_threads_t *thr_data=(arg_threads_t *)argt;
	unsigned long long imin=thr_data->imin;
	unsigned long long imax=thr_data->imax;
	//fprintf(stderr,"imin=%ld, imax=%ld, thread_id=%d\n",thr_data->imin, thr_data->imax,thr_data->thread_id);
	read_booster_t *read_booster=(read_booster_t *)thr_data->complex_arg[0];
	unsigned long long   *a_len=(unsigned long long*)thr_data->complex_arg[1];
	seqid_t *a_num=(seqid_t *)(thr_data->complex_arg[2]);
	int what=*(int *)thr_data->complex_arg[3];
	seqid_t *a_maxseqlen=(seqid_t *)(thr_data->complex_arg[4]);
	
	unsigned long long decime = (imax-imin)/10;
	if ( decime==0 )
		decime=1;
	UT_array **kmer2contigs = read_booster->a_kmer2contigs;
	sequenceset_t **seqset  = read_booster->h_seqcontigs;
	unsigned long long seqid;
	unsigned long long len;
	for(seqid=imin;seqid<=imax;seqid++)
	{
		if ( (seqid-imin) % decime == 0 )
			SMPStatusBar(thr_data->thread_id,(unsigned long long)(seqid-imin),(unsigned long long)(imax-imin+1),thr_data);
		if ( seqset[seqid] == NULL )
			continue;
		if ( ( what == PAIRS && seqset[seqid]->status == '0') || 
		     ( what == OBSOLETE && seqset[seqid]->status == '1' ) ||
		     ( what == LR && seqset[seqid]->lr != NULL ) )
		{
			len = KmerizeContigSeq(read_booster,seqid,kmer2contigs,seqset,what);
			if ( len )
			{
				*(a_len+thr_data->thread_id)+=(len-1);
				*(a_num+thr_data->thread_id)+=1;
				if ( len > *(a_maxseqlen+thr_data->thread_id) )
					*(a_maxseqlen+thr_data->thread_id)=len;
			}
		}
	}
	return(NULL);
}


void IdentifyErroneousSequencesSeqBySeq(read_booster_t *read_booster)
{
//	fprintf(stderr,"IdentifyErroneousSequencesSeqBySeq start\n");
	int i;
	unsigned long long a_nt_remove[MAX_THREAD_LOCKS];
	seqid_t a_not_ok[MAX_THREAD_LOCKS];
	for(i=0;i<read_booster->nb_cpus;i++)
	{
		a_nt_remove[i]=0;
		a_not_ok[i]=0;
	}
	
	arg_threads_t argt;
	argt.complex_arg[0]=read_booster;
	argt.complex_arg[1]=a_nt_remove;
	argt.complex_arg[2]=a_not_ok;
	
	InitSMPStatusBar(read_booster->nb_cpus,&argt);
	Multithreading(IdentifyErroneousSequencesSeqBySeqSlice,&argt,read_booster->nextid-1,read_booster->nb_cpus);
	CloseSMPStatusBar(read_booster->nb_cpus,&argt);
	
	unsigned long long nt_remove = 0;
	seqid_t not_ok = 0;
	for(i=0;i<read_booster->nb_cpus;i++)
	{
		nt_remove+=a_nt_remove[i];
		not_ok+=a_not_ok[i];
	}
	assert(nt_remove<read_booster->current_nt_size);
	read_booster->current_nt_size = read_booster->current_nt_size-nt_remove;
	if ( read_booster->debug )
		fprintf(stderr,"# pairs OK=%u; # pairs NOK=%u, #nt removed=%llu\n",GetSequenceSetCount(read_booster->h_seqcontigs,read_booster->nb_max_seq,PAIRS),not_ok,nt_remove);
//	fprintf(stderr,"IdentifyErroneousSequencesSeqBySeq done\n");
	fflush(stderr);
}

void *IdentifyErroneousSequencesSeqBySeqSlice(void *argt)
{
	arg_threads_t *thr_data=(arg_threads_t *)argt;
	unsigned long long imin=thr_data->imin;
	unsigned long long imax=thr_data->imax;
	read_booster_t *read_booster=(read_booster_t *)thr_data->complex_arg[0];
	unsigned long long   *a_nt_remove=(unsigned long long*)thr_data->complex_arg[1];
	seqid_t *a_not_ok=(seqid_t *)(thr_data->complex_arg[2]);
	
	unsigned long long decime = (imax-imin)/10;
	if ( decime==0 )
		decime=1;
	
	int klen=read_booster->klen;
	int i;
	seqid_t seqid;
	for(seqid=imin;seqid<=imax;seqid++)
	{
		if ( (seqid-imin) % decime == 0 )
			SMPStatusBar(thr_data->thread_id,(unsigned long long)(seqid-imin),(unsigned long long)(imax-imin+1),thr_data);
		
		char *seq=GetSeq(read_booster->h_seqcontigs,seqid);
		if ( seq == NULL )
			continue;
		
		char *ptr_str=seq;
		int lmax = strlen(seq)-klen;
		kmercode_t kmercode=0;
		short init = FALSE;
		int stop=FALSE;
		for(i=0;i<=lmax && !stop;i++,ptr_str++)
		{	
			init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
			if ( ! init )
			{
				i+=(klen-1);
				ptr_str+=(klen-1);
				continue;
			}
			stop=FALSE;
//			fprintf(stderr,"%u, %d, %u %d,%d\n",seqid, i, kmercode, read_booster->a_kmercode_selection[kmercode], read_booster->a_kmercode[kmercode]);
			if ( read_booster->a_kmercode_selection[kmercode]  < read_booster->minkdepth  && read_booster->a_kmercode[kmercode] < read_booster->depth )
			{
	//			fprintf(stderr,"%u, %d, %d\n",seqid,read_booster->a_kmercode_selection[kmercode],read_booster->a_kmercode[kmercode]);
				if ( SetSeqStatus(read_booster->h_seqcontigs,seqid,'1') )
				{
					// je ne vois pas pourquoi il faut KmerizeContigSeq sur une sequence que l'on declare obsolete, c'est une trace d'un vieux truc
					// nt_remove += KmerizeContigSeq(read_booster,seqid,NULL,read_booster->h_seqcontigs,PAIRS);
					*(a_nt_remove+thr_data->thread_id) += strlen(seq);
					*(a_not_ok+thr_data->thread_id)+=1;
					stop=TRUE;
				}
			}
		}
	}
	return(NULL);
}

// je ne locke pas la structure de donnee entiere, cela devrait eviter pas mal la concurrence
void IdentifyErroneousSequences(read_booster_t *read_booster)
{
	unsigned long long nt_remove = 0;
	
	short maxk_for_alloc = Min(read_booster->klen,16);
	unsigned long long max=pow(4,maxk_for_alloc);
	unsigned long long i;
	
	seqid_t not_ok = 0;
	for(i=0;i<max;i++)
	{
		if ( read_booster->a_kmer2contigs[i] == NULL )
			continue;
		
		int nb = utarray_len(read_booster->a_kmer2contigs[i]);
		if (nb < read_booster->minkdepth  && read_booster->a_kmercode[i] < read_booster->depth ) // 20140309 je rejoute que cela ne doit pas etre un kmer frequent dans les donnees brutes
		                                                                             // je prefere eviter de creer des trous
		{
			seqid_t *seqid;
			for(seqid=(seqid_t*)utarray_front(read_booster->a_kmer2contigs[i]); seqid!=NULL; seqid=(seqid_t*)utarray_next(read_booster->a_kmer2contigs[i],seqid)) 
			{
				if ( SetSeqStatus(read_booster->h_seqcontigs,*seqid,'1') )
				{
					// je ne pense plus que cela soit necessaire nt_remove += KmerizeContigSeq(read_booster,*seqid,NULL,read_booster->h_seqcontigs,PAIRS);
					char *seq=GetSeq(read_booster->h_seqcontigs,*seqid);
					nt_remove += strlen(seq);
					not_ok++;
				}
			}
			utarray_free(read_booster->a_kmer2contigs[i]);
			read_booster->a_kmer2contigs[i]=NULL;
		}
	}
	assert(nt_remove<read_booster->current_nt_size);
	read_booster->current_nt_size = read_booster->current_nt_size-nt_remove;
	if ( read_booster->debug )
		fprintf(stderr,"# pairs OK=%u; # pairs NOK=%u, #nt removed=%llu\n",GetSequenceSetCount(read_booster->h_seqcontigs,read_booster->nb_max_seq,PAIRS),not_ok,nt_remove);
}

void PushContigIntoKmerHash(read_booster_t *read_booster,UT_array **kmer2contigs,kmercode_t kmercode,seqid_t id)
{
	if (pthread_rwlock_wrlock(&read_booster->a_lock_a_kmer2contigs[kmercode%MAX_THREAD_LOCKS]) != 0) Fatal("can't get wrlock");
	if ( kmer2contigs[kmercode] == NULL )
		utarray_new(kmer2contigs[kmercode],&seqid_t_icd);
	utarray_push_back(kmer2contigs[kmercode],&id);
	pthread_rwlock_unlock(&read_booster->a_lock_a_kmer2contigs[kmercode%MAX_THREAD_LOCKS]);
}



/* 
 * =============================
 * =============================
 */

int TestMaxCoverage (read_booster_t *read_booster,short depth)
{
	int ret= ( (short)GetCurrentCoverage(read_booster) > depth ) ? TRUE : FALSE;
	return ret;
}


int HitSortByCount(hits_t *a,hits_t *b)
{
	return (b->count - a->count);
}


void UpdateCurrentNtSize(read_booster_t *read_booster,unsigned long long nt)
{
	read_booster->current_nt_size+=nt;
}

void UpdateCurrentNtSizeOneThread(read_booster_t *read_booster,unsigned long long nt,int thr_id)
{
	if (pthread_rwlock_wrlock(&read_booster->a_lock_current_nt_size_per_threads[thr_id]) != 0) Fatal("can't get wrlock");
	read_booster->current_nt_size_per_threads[thr_id]+=nt;
	pthread_rwlock_unlock(&read_booster->a_lock_current_nt_size_per_threads[thr_id]);
}

unsigned long long SetCurrentNtSize(read_booster_t *read_booster)
{
	int i;
	unsigned long long total = read_booster->current_nt_size;
	for(i=0;i<read_booster->nb_cpus;i++)
	{
		if (pthread_rwlock_rdlock(&read_booster->a_lock_current_nt_size_per_threads[i]) != 0) Fatal("can't get rdlock");
		total+=read_booster->current_nt_size_per_threads[i];
		pthread_rwlock_unlock(&read_booster->a_lock_current_nt_size_per_threads[i]);
	}
	read_booster->current_nt_size=total;
	return read_booster->current_nt_size;
}

seqid_t GetNextContigSeqID(read_booster_t *read_booster,int step)
{
	seqid_t previd = read_booster->nextid;
	read_booster->nextid+=step;
	assert(read_booster->nextid>previd);
	assert(read_booster->nextid<read_booster->nb_max_seq);
	return previd;
}

// attention ca presuppose que tous les threads avancent en meme temps
seqid_t GetNextContigSeqIDOneThread(read_booster_t *read_booster,int thr_id)
{
	seqid_t previd = read_booster->nextid_per_threads_curr[thr_id];
	if ( previd + 1 > read_booster->nextid_per_threads_max[thr_id] )
	{
		if (pthread_rwlock_wrlock(&read_booster->lock_seqcontigs) != 0) Fatal("can't get wrlock");
		time_t ltime;
		time(&ltime);
		read_booster->nextid_per_threads_curr[thr_id]=GetNextContigSeqID(read_booster,read_booster->step_thread);
		pthread_rwlock_unlock(&read_booster->lock_seqcontigs);
		read_booster->nextid_per_threads_max[thr_id]=read_booster->nextid_per_threads_curr[thr_id]+read_booster->step_thread-1;
//		fprintf(stderr,"\n#%d# (prev previd+1=%u) min=%u max=%u time=%tu\n",thr_id,previd+1,read_booster->nextid_per_threads_curr[thr_id],read_booster->nextid_per_threads_max[thr_id],ltime);
		previd = read_booster->nextid_per_threads_curr[thr_id];
	}
	read_booster->nextid_per_threads_curr[thr_id]++;
	assert(read_booster->nextid_per_threads_curr[thr_id]<read_booster->nb_max_seq);
	return previd;
}

/* 
 * =============================
 *  Inputs/Outputs
 * =============================
 */

int DumpFastaHash(read_booster_t *read_booster,char *filename,sequencesethash_t **seqcontigs,int pair_files,int nrdb)
{
	assert(filename);
	assert(strlen(filename)>0);
	assert(seqcontigs!=NULL);
//	sequencesethash_t **curr_seqcontigs = (seqcontigs==NULL) ? &read_booster->h_seqcontigs : seqcontigs;
	sequencesethash_t **curr_seqcontigs = seqcontigs;
	FILE *f_out    = GetStreamOut(filename);
	dejavu_t *rh_dejavu=NULL;
	
	sequencesethash_t *curr_seq=NULL;
	sequencesethash_t *tmp_seq=NULL;
	seqid_t nr=0;
	unsigned long long nucleotides=0;
	float qualscore=0;
	
	fprintf(stderr,"DumpFasta %s: # sequences=%d ",filename,HASH_COUNT(*curr_seqcontigs));
	HASH_ITER(hh, *curr_seqcontigs, curr_seq, tmp_seq)
	{
		if ( nrdb && ! AddIfNotDejaVu(&rh_dejavu,curr_seq->sequence,TRUE,TRUE,FALSE,&read_booster->lock_h_deja) )
		{
			nr++;
			continue;
		}
		qualscore=GetQualScore(read_booster,curr_seq->sequence);
		nucleotides += DumpOneSeq(f_out,curr_seq->id,curr_seq->sequence,qualscore,pair_files);
	}
	fclose(f_out);
	fprintf(stderr," coverage=%.2lf\n",(double)nucleotides/(double)read_booster->expected_nt_size);
	DeleteDeja(&rh_dejavu);
	if ( nr && read_booster->debug )
		fprintf(stderr,"%u redundant sequences removed\n",nr);
	return(TRUE);
}


int DumpFasta(read_booster_t *read_booster,char *filename,int pair_files,int nrdb,int what,seqid_t init)
{
	assert(filename);
	assert(strlen(filename)>0);
	FILE *f_out    = GetStreamOut(filename);
	dejavu_t *rh_dejavu=NULL;
	seqid_t nr=0;
	unsigned long long nucleotides=0;
	
	sequenceset_t **seqsrc=read_booster->h_seqcontigs;
	if ( read_booster->debug )
		fprintf(stderr,"Dump %s: # sequences=%d ",filename,GetSequenceSetCount(seqsrc,read_booster->nb_max_seq,what));
	seqid_t i;
	char *seq;
	for(i=init;i<read_booster->nextid;i++)
		if ( seqsrc[i] != NULL )
		{
			seq = GetSeqHub(seqsrc,i,what);
			if ( seq == NULL )
				continue;
			if ( what == PAIRS && seqsrc[i]->status != '0' )
				continue;
				
			if ( nrdb && ! AddIfNotDejaVu(&rh_dejavu,seq,TRUE,TRUE,FALSE,&read_booster->lock_h_deja) )
			{
				nr++;
				continue;
			}
			float qualscore = (what == PAIRS ) ? 0 : GetQualScore(read_booster,seq);
			nucleotides += DumpOneSeq(f_out,i,seq,qualscore,pair_files);
		}
	fclose(f_out);
	if ( read_booster->debug )
		fprintf(stderr," coverage=%.2lf\n",(double)nucleotides/(double)read_booster->expected_nt_size);
	DeleteDeja(&rh_dejavu);
	if ( nr && read_booster->debug )
		fprintf(stderr,"%u redundant sequences removed\n",nr);
	return(TRUE);
}

double GetMeanLengthLR(read_booster_t *read_booster)
{
	seqid_t i;
	double mean=0;
	seqid_t cpt=0;
	sequenceset_t **seqsrc=read_booster->h_seqcontigs;
	for(i=0;i<read_booster->nextid;i++)
		if ( seqsrc[i] != NULL )
		{
			char *seq = GetSeqHub(seqsrc,i,LR);
			if ( seq != NULL )
			{
				mean+=strlen(seq);
				cpt++;
			}
		}
	return (mean/cpt);
}



int CleanSequences(read_booster_t *read_booster,sequencesethash_t **seqsrc)
{
//	seqid_t i;
	int klen=read_booster->klen;
	assert(klen <= 16);
	int len;
	int j;
	char *seq,*ptr_str;
	kmercode_t kmercode;
	kmercode_t rckmercode;
	short curr,prev,next;
	short prevcurr=0;
	short rc=0;
	sequencesethash_t *curr_seq=NULL;
	sequencesethash_t *tmp_seq=NULL;
	HASH_ITER(hh, *seqsrc, curr_seq, tmp_seq)
	{
		seq=curr_seq->sequence;
		len=strlen(seq);
		int lmax = len-klen;
		ptr_str=seq+(len-klen);
		for(j=lmax;j>=klen;j--,ptr_str--)
		{	
			if ( ! KmerEncode(ptr_str,klen,&kmercode) )
				continue;
			curr= read_booster->a_kmercode[kmercode];
			if ( ! KmerEncode(ptr_str+klen,klen,&kmercode) )
				continue;
			next= read_booster->a_kmercode[kmercode];
			if ( ! KmerEncode(ptr_str-klen,klen,&kmercode) )
				continue;
			prev= read_booster->a_kmercode[kmercode];
			if ( j < lmax && curr < 5 && abs(prevcurr-curr) > 10 )
			{
				RevCompCode(kmercode,klen,&rckmercode);
				rc = read_booster->a_kmercode[kmercode];
				if ( (rc+curr) < 10 &&  // c'est faible sur les deux brins
					  abs(prevcurr-rc) > 10 && // il y a aussi un shift avec le compte du rc
					  abs(curr-prev) > 10 ) // decalage avec le precedent
					fprintf(stderr,"%d  %d,%d,%d prevcurr=%d, rc=%d\n%s\n%13s%s\n%26s%s\n",j,prev,curr,next,prevcurr,rc,ptr_str-klen,"",ptr_str,"",ptr_str+klen);
			}
			prevcurr = curr;
		}
	}
	return(TRUE);
}


int DumpOneSeq(FILE *f_out,seqid_t seqid,char *seq,float qualscore,int pair_files)
{
	char end1[MAXREADNT]={0};
	char end2[MAXREADNT]={0};
	char buf[MAXLINE]={0};
	int len=0;
	char *revseq;
	int i;
	unsigned long long nucleotides=0;
	if ( pair_files )
	{
		strcpy(buf,seq);
		len=strlen(buf);
		for(i=0;i<len;i++)
		{
			if ( buf[i] == 'N' )
				buf[i] = ' ';
		}
		sscanf(buf,"%s %s",end1,end2);
		fprintf(f_out,">%d/1\n%s\n",seqid,end1);
		nucleotides+=strlen(end1);
		revseq =  RevCompSeq(end2);
		nucleotides+=strlen(end2);
		fprintf(f_out,">%d/2\n%s\n",seqid,revseq);
		free(revseq);
	}
	else
	{
		i=0;
		size_t len=strlen(seq);
		if ( qualscore )
			fprintf(f_out,">%u_%zu_%.3f\n%s\n",seqid,len,qualscore,seq);
		else
			fprintf(f_out,">%u_%zu\n%s\n",seqid,len,seq);
		nucleotides+=strlen(seq);
	}
	return nucleotides;
}


void DumpQual(FILE *f_qual,char *seq,int qual)
{
	int len=strlen(seq);
	int i;
	for(i=0;i<len;i++)
	{
		fprintf(f_qual,"%3d",qual);
		if ( (i+1) % DEFAULT_NT_PER_LINE ==0)
			fprintf(f_qual,"\n");
	} 
	if ( len % DEFAULT_NT_PER_LINE != 0 )
	    fprintf(f_qual,"\n");
}

seqid_t Load(read_booster_t *read_booster,char *filename,int kmerize,int preserve_id,int paired,unsigned int *maxrdlen,int what)
{
	fprintf(stderr,"Loading started\n");
	if ( ! preserve_id )
		fprintf(stderr,"Warnings: original IDs are lost\n");
	seqid_t curr_id=0;
	seqid_t max_id=0;
	char line[MAXREADNT];
	seqid_t cpt = 0;
	FILE *f_in = GetStreamIn(filename);
	char read1[MAXREADNT];
	char read2[MAXREADNT];
	unsigned int read1len=0;
	max_id=0;
	while(fgets(line,MAXREADNT,f_in) != NULL)
	{
		switch(line[0])
		{
			case '>':
				if ( preserve_id )
				{
					curr_id=GetSeqId(line);
					if (max_id < curr_id )
						max_id = curr_id;
				}
				if ( fgets(read1,MAXREADNT,f_in) != NULL )
				{
					read1len=strlen(read1);
					read1[read1len-1]='\0';
					if ( paired )
					{
						if ( fgets(read2,MAXREADNT,f_in) != NULL && fgets(read2,MAXREADNT,f_in) != NULL )
							read2[strlen(read2)-1]='\0';
						else
							exit(EXIT_FAILURE);
					}
				}
				else
					exit(EXIT_FAILURE);
				break;
		}
		if ( maxrdlen != NULL )
		{
			if ( read1len > *maxrdlen ) // pas besoin de tester read2
				*maxrdlen=read1len;
		}
		switch(what)
		{
			case PAIRS:
				if ( paired )
				{
//					id =  AddNewContigSeqDejaArray(read_booster,read1,read2,kmerize,curr_id,-1);
					char *seq=MallocSeqPair(read1,read2);
					SetSeq(read_booster->h_seqcontigs,curr_id,seq);
					if ( kmerize )
						KmerizeContigSeq(read_booster,curr_id,NULL,NULL,PAIRS);
					free(seq);
				}
				else
				{
//					id =  AddNewContigSeqDejaArray(read_booster,read1,NULL,kmerize,curr_id,-1);
					SetSeq(read_booster->h_seqcontigs,curr_id,read1);
					if ( kmerize )
						KmerizeContigSeq(read_booster,curr_id,NULL,NULL,PAIRS);
				}
				break;
			case LR:
				if ( read_booster->h_seqcontigs[curr_id] == NULL )
				{
					fprintf(stderr,"Fatal error: for restarting properly, pairs must be loaded before lr\n");
					exit(EXIT_FAILURE);
				}
				SetSeqLR(read_booster->h_seqcontigs,curr_id,read1);
				break;
			default:
				fprintf(stderr,"Fatal error: unmanage data type >%d<\n",what);
				exit(EXIT_FAILURE);
		}
		//if ( id )
			cpt++;
	}
	CloseStream(filename,f_in);
	fprintf(stderr,"Loading done, maxid=%d, %llu nt\n",cpt,read_booster->current_nt_size);
	if ( preserve_id )
	{
		read_booster->nextid = max_id+1;
	}
	return(cpt);
}

seqid_t GetSeqId(char *line)
{
	int j=1;
	int jj=0;
	char buf_id[200];
	seqid_t curr_id;
	while(line[j]!='/'&&line[j]!=' '&&line[j]!='\0'&&line[j]!='\n'&&line[j]!='_')
	{
		buf_id[jj++]=line[j++];
	}
	buf_id[jj]='\0';
	sscanf(buf_id,"%u",&curr_id);
	return(curr_id);
}

/* 
 * =============================
 * utils for kmer data structure
 * =============================
 */

void DumpKmers(kmer2contigs_t *kmer2contigs,FILE *out)
{
	kmer2contigs_t *s=NULL;
	seqid_t *p;
		
	for(s=kmer2contigs; s != NULL; s=s->hh.next) 
	{
		fprintf(out,"KMER: %u\n", s->kmer);
		for(p=(seqid_t*)utarray_front(s->list_of_contigs); p!=NULL; p=(seqid_t*)utarray_next(s->list_of_contigs,p)) 
		{
			fprintf(out,"%u ",*p);
		}
		fprintf(out,"\n");
	}
}

void ResetKmers(UT_array **kmer2contigs_param,short klen)
{
	short maxk_for_alloc = Min(klen,16);
	unsigned long long max = pow(4,maxk_for_alloc);
	unsigned long long i;
	for(i=0;i<max;i++)
	{
		if ( kmer2contigs_param[i] != NULL )
		{
//			utarray_clear(kmer2contigs_param[i]);
			utarray_free(kmer2contigs_param[i]);
			kmer2contigs_param[i]=NULL;
		}
	}
}

void FreeHits(hits_t **h_hits)
{
	hits_t *onehit=NULL;
	hits_t *tmphit=NULL;
	HASH_ITER(hh, *h_hits, onehit, tmphit) 
	{
		HASH_DEL(*h_hits,onehit);
		free(onehit);
	}
}

void PrintHits(hits_t **h_hits)
{
	hits_t *onehit=NULL;
	hits_t *tmphit=NULL;
	fprintf(stderr,"=============\n");
	HASH_ITER(hh, *h_hits, onehit, tmphit) 
	{
		fprintf(stderr,"\t%u,count=%d,s=%c\n",onehit->id,onehit->count,onehit->strand);
	}
}

// utilitaire
// ca serait plus econome en ram de faire un premier passage pour compter, allouer au plus juste et repasser

seqid_t InitArrayOfSeqOk(read_booster_t *read_booster,seqid_t **a_tasks,seqid_t *maxid,int what)
{
	seqid_t max=0;
	sequenceset_t **seqsrc=read_booster->h_seqcontigs;
	seqid_t nb_seq_max = read_booster->nb_max_seq;
	*a_tasks=(seqid_t *)malloc(sizeof(seqid_t)*GetSequenceSetCount(seqsrc,nb_seq_max,what));
	assert(*a_tasks!=NULL);
	seqid_t *onetask=*a_tasks;
	seqid_t i;
	*maxid=0;
	for(i=0;i<=nb_seq_max;i++)
	{
		if ( seqsrc[i] == NULL )
			continue;
		
		if ( what == PAIRS || (what == LR && seqsrc[i]->lr != NULL ) || (what == VLR_CAND && seqsrc[i]->status == '5' ) )
		{
			*onetask=i;
			if ( i > *maxid )
				*maxid=i;
			onetask++;
			max++;
		}
	}
	max--;
	return(max);
}


void BuildLongReads(read_booster_t *read_booster,char *longreadsfile,unsigned long long *max_mem)
{
	int maxrdlen=read_booster->insert_size*4;
	if ( read_booster->debug )
		fprintf(stderr,"BuildLongReads: minpci=%f minovl=%d maxovh=%d maxmm=%d maxreadlen=%d, nb_threads=%d insert_size=%d\n",read_booster->minpci,read_booster->minovl,read_booster->maxovh,read_booster->maxmm,maxrdlen,read_booster->nb_cpus,read_booster->insert_size);
	seqid_t *a_tasks=NULL;
	seqid_t maxid=0;
	seqid_t max=InitArrayOfSeqOk(read_booster,&a_tasks,&maxid,PAIRS);
	
//	fprintf(stderr,"Disabling HF, est ce que je garde?\n");
	int no_high_freq = TRUE;
	sequencesethash_t *seqloci_nok=NULL;
	
	pthread_rwlock_t lock_seqloci_nok;
	if ( pthread_rwlock_init(&lock_seqloci_nok,NULL) != 0) Fatal("can't create rwlock");
	
	arg_threads_t argt;
	argt.complex_arg[0]=read_booster;
	argt.complex_arg[1]=a_tasks;
	argt.complex_arg[2]=&maxrdlen;
	argt.complex_arg[3]=&seqloci_nok;
	argt.complex_arg[4]=&lock_seqloci_nok;
	char status='0'; // dans un premier temps on essaye de construire les sequences ok
	argt.complex_arg[5]=&status;
	argt.complex_arg[6]=&no_high_freq;
	
	InitSMPStatusBar(read_booster->nb_cpus,&argt);
	Multithreading(BuildLongReadsSliceBlock,&argt,max,read_booster->nb_cpus);
	CloseSMPStatusBar(read_booster->nb_cpus,&argt);
	
	free(a_tasks);
	
	if ( read_booster->debug )
		InitMaxMemUsage(max_mem);
	
//	CleanSequences(read_booster,&seqloci_ok);
	if ( read_booster->nrdb )
	{
		if ( read_booster->debug )
			fprintf(stderr,"LRNR =====================\n");
		ResetKmerize(read_booster,LR);
		RemoveIncludedSequencesLR(read_booster,0);
	}
	DumpFasta(read_booster,longreadsfile,FALSE,TRUE,LR,0);
	
#ifndef NDEBUG
	if ( read_booster->debug )
	{
		char seqfile_nok[MAXFILENAME];
		sprintf(seqfile_nok,"%s.nok.fa",longreadsfile);
		DumpFastaHash(read_booster,seqfile_nok,&seqloci_nok,TRUE,FALSE);
	}
#endif
	
	if ( read_booster->debug )
	{
		InitMaxMemUsage(max_mem);
		fprintf(stderr,"Max Memory Usage: %d Mb\n",(int)*max_mem/1024);
	}
	DeleteSequenceSetHash(&seqloci_nok);
}

void *BuildLongReadsSliceBlock(void *argt)
{
	// recuperation des parametres
	arg_threads_t *thr_data=(arg_threads_t *)argt;
	unsigned long long imin=thr_data->imin;
	unsigned long long imax=thr_data->imax;
	read_booster_t *read_booster=(read_booster_t *)thr_data->complex_arg[0];
	seqid_t *ra_tasks=(seqid_t *)thr_data->complex_arg[1];
	int maxrdlen=*(int *)(thr_data->complex_arg[2]); // plantage iteration 2
	sequencesethash_t **seqloci_nok=(sequencesethash_t **)thr_data->complex_arg[3];
	pthread_rwlock_t *ptr_lock_seqloci_nok=(pthread_rwlock_t *)thr_data->complex_arg[4];
	char status=*(char *)(thr_data->complex_arg[5]);
	int no_high_freq=*(int *)(thr_data->complex_arg[6]);
	
	int with_optim = (! read_booster->disable_optimization);
	
	sequenceset_t **seqsrc = read_booster->h_seqcontigs;
	int nb_cpus = read_booster->nb_cpus;
	
// pour chaque insert, je compte le nombre de ponts entre extremites qu'il est capable de faire
	unsigned long long i;
	char pair[MAXREADNT];
	char mapped_pair[MAXREADNT];
	char* a_reads[2]={NULL,NULL};
	char* a_reads_extensible[2]={NULL,NULL};
	char* a_mapped_reads[2]={NULL,NULL};
	seqid_t seqid;
	hsp_t hsp;
	char  newseq1r[MAXREADNT]={0};
	char  newseq1l[MAXREADNT]={0};
	char  newseq2r[MAXREADNT]={0};
	char  newseq2l[MAXREADNT]={0};
	unsigned long long decime = (imax-imin)/10;
	if ( decime==0 )
		decime=1;
	hits_t *rh_hits=NULL;
	hits_t *onehit=NULL;
	hits_t *tmphit=NULL;
	
	short **p_mat=NULL;
	MallocMatrixOfShort(&p_mat,MAXREADNT_MATRIX);
	
//	int redundant_with_prev_lr=FALSE;
//	seqid_t computation_avoided=0;
	pthread_rwlock_t *seqlock=NULL;
	pthread_rwlock_t *mappedseqlock=NULL;
	int helper=-1;
	int helped=FALSE;
	for(i=imin;i<=imax;i++)
	{
		if ( (i % (read_booster->nb_cpus*3)) == 0 ) // je ne veux pas verifier trop souvent
		{
			if ( (helper=GetFinishedThreadId(thr_data->self_array)) != -1 )
			{
				if ( MultithreadingShareSliceWithHelper(BuildLongReadsSliceBlock,argt,i,imax,thr_data->thread_id,helper,thr_data->job_id) )
				{
					helped=TRUE;
					break;
				}
			}
		}
		
		if ( (i-imin) % decime == 0 )
			SMPStatusBar(thr_data->thread_id,(unsigned long long)(i-imin),(unsigned long long)(imax-imin+1),thr_data);
		
		//if ( i % nb_cpus != thr_data->thread_id )
		//	continue;
		
		seqid = *(ra_tasks+i);
		
		if ( seqsrc[seqid] == NULL )
			continue;
		
		if( seqsrc[seqid]->status != status ) // je passe aussi les redondants
			continue;
		
		time_t init_time;
		time_t curr_time;
		time(&init_time);
		
		seqlock = &read_booster->a_lock_a_kmer2contigs[seqid%MAX_THREAD_LOCKS];
		
		//fprintf(stderr,"seqid: %u\n",seqid);
		//assert(seqid);
		
		if (pthread_rwlock_rdlock(seqlock) != 0) Fatal("can't get rdlock");
		char *seqpair = GetSeq(seqsrc,seqid);
		pthread_rwlock_unlock(seqlock);
		
		if ( seqpair == NULL && with_optim )
		{
			if (pthread_rwlock_wrlock(seqlock) != 0) Fatal("can't get wrlock");
			seqsrc[seqid]->status='4';
			pthread_rwlock_unlock(seqlock);
			continue;
		}
		strcpy(pair,seqpair);
		Split(pair,'N',a_reads);
		int again = TRUE;
		int assembled = FALSE;
		hits_t *rh_cand1=NULL;
		hits_t *rh_cand2=NULL;
		hits_t *onecand1=NULL;
		hits_t *onecand2=NULL;
		hits_t *tmpcand1=NULL;
		hits_t *rh_notusable=NULL; //conserve les id des sequences qui n'allongent pas
		hits_t *onenotusable=NULL;
		int not_extensible_due_to_kmer_composition = FALSE;
		while(again)
		{
			time(&curr_time);
			unsigned long diff_time = difftime(curr_time,init_time);
			if ( diff_time > LR_MAX_TIME )
			{
				fprintf(stderr,"Warning: too much time for solving %u (%lus) len=%zu\n",seqid,diff_time,strlen(a_reads[0])+strlen(a_reads[1]));
				again = FALSE;
				break;
			}
			hits_t *rh_hits1=NULL;
			hits_t *rh_hits2=NULL;
			
			// pas besoin de locker, rh_hits1 est local et on lit un utarray
			if ( Map(read_booster,a_reads[0],&rh_hits1,read_booster->minkmers_for_lr_extension,ALL_SCORE,seqid,no_high_freq,TRUE) < read_booster->minkmers_for_lr_extension ||
				 Map(read_booster,a_reads[1],&rh_hits2,read_booster->minkmers_for_lr_extension,ALL_SCORE,seqid,no_high_freq,TRUE) < read_booster->minkmers_for_lr_extension )
			{
				again=FALSE;
				not_extensible_due_to_kmer_composition=TRUE;
				continue;
			}	
			
			//PrintHits(&rh_hits1);

			// r: extension a droite de la lecture
			// l: extension a gauche
			short continue1r = TRUE;
			short continue1l = TRUE;
			short continue2l = TRUE;
			short continue2r = TRUE;
			short maxpos1r=0;
			float maxpci1l=0;
			float maxpci1r=0;
			short maxpos1l=SHRT_MAX;
			short maxpos2r=0;
			float maxpci2r=0;
			float maxpci2l=0;
			short maxpos2l = SHRT_MAX;
			newseq1r[0]='\0';
			newseq1l[0]='\0';
			newseq2r[0]='\0';
			newseq2l[0]='\0';
			short lenread1 = strlen(a_reads[0]);
			short lenread2 = strlen(a_reads[1]);
			hits_t *onehit1=NULL;
			hits_t *tmphit1=NULL;
			hits_t *onehit2=NULL;
			HASH_SORT(rh_hits1,HitSortByCount);
			HASH_ITER(hh, rh_hits1, onehit1, tmphit1) 
			{
				// TRES IMPORTANT on n'utilise que des pe mappes des deux cotes
				HASH_FIND(hh,rh_hits2,&onehit1->id,sizeof(seqid_t),onehit2);
				if( onehit2 == NULL )
					continue;
				if (onehit1->strand != onehit2->strand )
					continue;
			
				mappedseqlock = &read_booster->a_lock_a_kmer2contigs[onehit1->id%MAX_THREAD_LOCKS];
				
				assembled=FALSE;
				if ( with_optim )
				{
					if (pthread_rwlock_rdlock(mappedseqlock) != 0) Fatal("can't get rdlock");
					char status = seqsrc[onehit1->id]->status;
					pthread_rwlock_unlock(mappedseqlock);
					if ( status != '0' && status != '2' )   // je vire les obsolete et les redondants
						continue;
					
					if ( seqsrc[onehit1->id]->lr != NULL )
					{ 
						StrCat(&a_reads_extensible[0],a_reads[0],FALSE);
						StrCat(&a_reads_extensible[1],a_reads[1],FALSE);
						assembled = FillInGapWithLR(seqid,a_reads_extensible,onehit1->id,onehit1->strand,read_booster,TRUE,&p_mat[0]);
						FreeSplit(a_reads_extensible,2);
						if ( assembled )
						{
							again=FALSE;
//							fprintf(stderr,"############################### Assemblage par Fill pour %u\n",seqid);
							break;
						}
					}
				}

				seqid_t seqid_nostrand = onehit1->id;
				HASH_FIND(hh,rh_notusable,&seqid_nostrand,sizeof(seqid_t),onenotusable);
				if ( onenotusable ) // local, mappes qui ne servent pas a etendre
					continue;
				
					
				if (pthread_rwlock_rdlock(mappedseqlock) != 0) Fatal("can't get rdlock");
				char *seqmapped = GetSeq(seqsrc,onehit1->id);
				pthread_rwlock_unlock(mappedseqlock);
				
				if ( seqmapped == NULL )
				{
					// cela arrive car meme pour une sequence deplacee il y reste suffisament de kmer pour mapper (je mets pas a jour la liste des kmer > seuil)
					if (pthread_rwlock_wrlock(mappedseqlock) != 0) Fatal("can't get wrlock");
					seqsrc[onehit1->id]->status='4';
					pthread_rwlock_unlock(mappedseqlock);
					continue;
				}
				strcpy(mapped_pair,seqmapped);
				
				if ( onehit1->strand == '-' )
				{
					char *rcseq = RevCompSeq(mapped_pair);
					strcpy(mapped_pair,rcseq);
					free(rcseq);
				}
				
				int a_notusable[3]={0,0,0};
				Split(mapped_pair,'N',a_mapped_reads);
				int t;
				for(t=0;t<2;t++)
				{
					char *tgt  = a_mapped_reads[t];
					short ltgt = strlen(tgt);
					if ( PairwiseMatchUngapped(a_reads[0],tgt,&hsp,p_mat) )
						if ( ! ExtendSequence(read_booster,seqid_nostrand,&hsp,a_reads[0],tgt,lenread1,ltgt,&rh_cand1,&maxpos1r,&maxpci1r,newseq1r,&maxpos1l,&maxpci1l,newseq1l) )
							a_notusable[t]++;
					if ( PairwiseMatchUngapped(a_reads[1],tgt,&hsp,p_mat) )
						if ( ! ExtendSequence(read_booster,seqid_nostrand,&hsp,a_reads[1],tgt,lenread2,ltgt,&rh_cand2,&maxpos2r,&maxpci2r,newseq2r,&maxpos2l,&maxpci2l,newseq2l) )
							a_notusable[t]++;
				}
				if ( a_notusable[0] == 2 || a_notusable[1] == 2 )
					HitPlusPlus(&rh_notusable,seqid_nostrand,'-');
			}
			
			if ( ! assembled ) // desormais cela peut etre assemble en comblant
			{
				// Read1
				if ( strlen(newseq1r) && ! StrEndWith(a_reads[0],newseq1r) )
					StrCat(&a_reads[0],newseq1r,FALSE);
				else
					continue1r=FALSE;
				
				if ( strlen(newseq1l)  && ! StrStartWith(a_reads[0],newseq1l) )
					StrCat(&a_reads[0],newseq1l,TRUE);
				else
					continue1l=FALSE;
				
				// Read2
				if ( strlen(newseq2r) && ! StrEndWith(a_reads[1],newseq2r) )
					StrCat(&a_reads[1],newseq2r,FALSE);
				else
					continue2r=FALSE;
				
				if ( strlen(newseq2l) && ! StrStartWith(a_reads[1],newseq2l) )
					StrCat(&a_reads[1],newseq2l,TRUE);
				else
					continue2l=FALSE;
				
				if ( (continue1r+continue1l+continue2l+continue2r) == 0  )
					again = FALSE;
				else if ( PairwiseMatchUngapped(a_reads[0],a_reads[1],&hsp,p_mat) )
				{
					//fprintf(stderr,"R1:%s\nR2:%s\n",a_reads[0],a_reads[1]);
					// verification du chev des deux extremites
					if ( (hsp.mm <= read_booster->maxmm || hsp.pci >= read_booster->minpci) &&  hsp.len >= read_booster->minovl && (hsp.qend >= (strlen(a_reads[0])-read_booster->maxovh)) && (hsp.tstart <= (1+read_booster->maxovh))) 
					{
						if ( IsSequenceConsistantWithKmers(read_booster,pair,read_booster->minkdepth) ) // je verif que je ne cree pas de kmer rare
						{
							again = FALSE;
							// je tolere l'overhang
							strncpy(pair,a_reads[0],hsp.qend);
							pair[hsp.qend]='\0';
							strcat(pair,a_reads[1]+(hsp.tend+1-1));
							if (pthread_rwlock_wrlock(seqlock) != 0) Fatal("can't get wrlock");
							SetSeqLR(seqsrc,seqid,pair);
							pthread_rwlock_unlock(seqlock);
							
							assembled = TRUE;
							// s'il a servi a etendre des deux cotes et que je trou a ete comble je ne l'utilise plus
						}
						else
						{
							//fprintf(stderr,"NOK:%u %s\nR1:%s l=%d\nR2:%s l=%d\n",seqid,pair,a_reads[0],strlen(a_reads[0]),a_reads[1],strlen(a_reads[1]));
							sprintf(pair,"%sN%s",a_reads[0],a_reads[1]);// je tente d'etendre encore
							//PrintHsp(&hsp);
							// sequence consensus potentiellement problematique, pas de N => read2 vide dans nok
						}
					}
					else if ( hsp.len >= STOP_EXTENTION_MINLEN && hsp.pci >=  STOP_EXTENTION_MINPCI ) //si cela chevauche beaucoup et que cela se ressemble il y a un souci sur overhang ou une divergence
					{
	//					PrintHsp(&hsp);
						sprintf(pair,"%sN%s",a_reads[0],a_reads[1]);
						again=FALSE;
					}
					else
						sprintf(pair,"%sN%s",a_reads[0],a_reads[1]);
					
					if ( (strlen(a_reads[0])+strlen(a_reads[1])+1) > maxrdlen )
					{
						again=FALSE;
					}
				}
			}
			if ( assembled && with_optim )
			{
				onecand1=NULL;
				tmpcand1=NULL;
				HASH_ITER(hh,rh_cand1,onecand1,tmpcand1)
				{
					onecand2=NULL;
					HASH_FIND(hh,rh_cand2,&onecand1->id,sizeof(seqid_t),onecand2);
					if ( onecand2 == NULL )
						continue;
					mappedseqlock = &read_booster->a_lock_a_kmer2contigs[onecand1->id%MAX_THREAD_LOCKS];
					if (pthread_rwlock_wrlock(mappedseqlock) != 0) Fatal("can't get wrlock");
					seqsrc[onecand1->id]->status='4';
					pthread_rwlock_unlock(mappedseqlock);
				}
				// je vais marquer redondant toutes les sequences completement mappees sur la LR
				rh_hits=NULL;
				onehit=NULL;
				tmphit=NULL;
#ifndef REVMAP
				Map(read_booster,pair,&rh_hits,read_booster->maxkmers*2,ALL_SCORE,seqid,no_high_freq,TRUE);
#else
				// je ne peux pas garder maxkmers mais il y a un risque de supprimer trop
				Map(read_booster,pair,&rh_hits,(read_booster->maxkmers-1)*2,ALL_SCORE,seqid,no_high_freq,TRUE);
#endif
				if ( rh_hits != NULL )
				{
					HASH_ITER(hh,rh_hits,onehit,tmphit)
					{
						mappedseqlock = &read_booster->a_lock_a_kmer2contigs[onehit->id%MAX_THREAD_LOCKS];
						if (pthread_rwlock_wrlock(mappedseqlock) != 0) Fatal("can't get wrlock");
						seqsrc[onehit->id]->status='4'; // redondant
						pthread_rwlock_unlock(mappedseqlock);
					}
					FreeHits(&rh_hits);
				}
			}
			if ( ! again )
				if ( ! assembled )
				{
					if (pthread_rwlock_wrlock(ptr_lock_seqloci_nok) != 0) Fatal("can't get wrlock");
					if ( 0 && read_booster->debug ) // je desactive tant que je ne veux pas creuser
						SetSeqHash(seqloci_nok,seqid,pair);
					seqsrc[seqid]->status='2';
					pthread_rwlock_unlock(ptr_lock_seqloci_nok);
				}
			FreeHits(&rh_hits1);
			FreeHits(&rh_hits2);
		}
		if ( not_extensible_due_to_kmer_composition )
		{
			if (pthread_rwlock_wrlock(ptr_lock_seqloci_nok) != 0) Fatal("can't get wrlock");
			seqsrc[seqid]->status='3';
			pthread_rwlock_unlock(ptr_lock_seqloci_nok);
		}
		FreeHits(&rh_cand1);
		FreeHits(&rh_cand2);
		FreeHits(&rh_notusable);
	}
	if ( ! helped )
		SetThreadIdAsFinished(thr_data->self_array,thr_data->thread_id);
	FreeSplit(a_reads_extensible,2);
	FreeSplit(a_reads,2);
	FreeSplit(a_mapped_reads,2);
	FreeMatrixOfShort(&p_mat[0],MAXREADNT_MATRIX);
//	if ( read_booster->debug )
	//	fprintf(stderr,"Thread: %i, computation avoided: %u/%u (%.2f)\n",thr_data->thread_id,computation_avoided,(imax-imin+1),(double)computation_avoided*100/(double)(imax-imin+1));
	return(NULL);
}

// prerequis ResetKmerize(read_booster,LR);
void RemoveIncludedSequencesLR(read_booster_t *read_booster,seqid_t limit)
{
	seqid_t *a_tasks=NULL;
	seqid_t maxid=0;
	seqid_t max=InitArrayOfSeqOk(read_booster,&a_tasks,&maxid,LR);
	
	int what = LR;
	arg_threads_t argt;
	argt.complex_arg[0]=read_booster;
	argt.complex_arg[1]=&limit;
	argt.complex_arg[2]=&what;
	argt.complex_arg[3]=a_tasks;
	
	InitSMPStatusBar(read_booster->nb_cpus,&argt);
	MultithreadingNoSlice(RemoveIncludedSequencesMappedSlice,&argt,max,read_booster->nb_cpus);
	CloseSMPStatusBar(read_booster->nb_cpus,&argt);
	
	free(a_tasks);
}

	
void *RemoveIncludedSequencesMappedSlice(void *argt)
{
	arg_threads_t *thr_data=(arg_threads_t *)argt;
	unsigned long long imin=thr_data->imin;
	unsigned long long imax=thr_data->imax;
	read_booster_t *local_read_booster=(read_booster_t *)thr_data->complex_arg[0];
	seqid_t limit=*(seqid_t *)thr_data->complex_arg[1];
	int what=*(int *)thr_data->complex_arg[2];
	seqid_t *a_tasks=(seqid_t *)thr_data->complex_arg[3];
	
	sequenceset_t **seqsrc = local_read_booster->h_seqcontigs;
	int nb_cpus = local_read_booster->nb_cpus;
	
	// en // je fais voir si chaque sequence est inclus dans une autre
	// si oui je la supprime
	// a la fin il ne doit rester que des sequences non inclues dans d'autres
	
	unsigned long long decime = (imax-imin)/10;
	if ( decime==0 )
		decime=1;
	hsp_t hsp;
	hits_t *onehit=NULL;
	hits_t *tmphit=NULL;
	char *seqprev;
	char *rcseqprev=NULL;
	seqid_t seqid;
	unsigned long long i;
	char *seq;
	char seqtmp[MAXREADNT];
	char seqprevtmp[MAXREADNT];
	int cutoff=0;
	rcseqprev=NULL; 
	
	short **p_mat=NULL;
	MallocMatrixOfShort(&p_mat,MAXREADNT_MATRIX);
	
	for(i=imin;i<=imax;i++)
	{
		if ( (i-imin) % decime == 0 )
			SMPStatusBar(thr_data->thread_id,(unsigned long long)(i-imin),(unsigned long long)(imax-imin+1),thr_data);
		
		if ( i % nb_cpus != thr_data->thread_id )
			continue;
		
		
		seqid = a_tasks[i];
		
		if (pthread_rwlock_rdlock(&local_read_booster->a_lock_a_kmer2contigs[seqid%MAX_THREAD_LOCKS]) != 0) Fatal("can't get rdlock");
		seq = GetSeqHub(seqsrc,seqid,what);
		if ( seq != NULL )
			strcpy(seqtmp,seq); // je travaille sur un tmp car Paiwise a besoin de la sequence qui peut etre supprimer par un autre
		pthread_rwlock_unlock(&local_read_booster->a_lock_a_kmer2contigs[seqid%MAX_THREAD_LOCKS]);
		
		if ( seq == NULL )
			continue;
		
//		fprintf(stderr,"%s\n",seqtmp);
		
		int len=strlen(seqtmp);
		// j'identifie les sequences precedement selectionnees sur lesquelles cela vaut le coup de tenter un alignement
		hits_t *rh_hits=NULL;
		/* je reviens en arriere, c'est trop stringent
		switch(local_read_booster->maxmm)
		{
			case 0:
				cutoff = ceil(len/local_read_booster->klen);
				break;
			case 1:
			case 2:
				cutoff = ceil(len/local_read_booster->klen)-local_read_booster->maxmm;
				break;
			default:
				cutoff = ceil((len/local_read_booster->klen)*0.66); // 2/3 des kmers possibles
				break;
		}
		*/
		cutoff = ceil((len/local_read_booster->klen)*0.66); // 2/3 des kmers possibles
		Map(local_read_booster,seqtmp,&rh_hits,cutoff,ALL_SCORE,seqid,FALSE,FALSE);
		HASH_SORT(rh_hits,HitSortByCount); // je compare en premier ce qui se ressemble le plus
		//PrintHits(&rh_hits);
		HASH_ITER(hh,rh_hits,onehit,tmphit)
		{
			if ( onehit->id < limit && seqid < limit ) // si les deux sont des vraies LR, j'ai deja fais la suppression si limit > 0 && je ne refais pas le test
				continue;
			if (pthread_rwlock_rdlock(&local_read_booster->a_lock_a_kmer2contigs[onehit->id%MAX_THREAD_LOCKS]) != 0) Fatal("can't get rdlock");
			seqprev = GetSeqHub(seqsrc,onehit->id,what);
			if ( seqprev != NULL )
				strcpy(seqprevtmp,seqprev); // je travaille sur un tmp car Paiwise a besoin de la sequence qui peut etre supprimer par un autre
			pthread_rwlock_unlock(&local_read_booster->a_lock_a_kmer2contigs[onehit->id%MAX_THREAD_LOCKS]);
			if ( seqprev == NULL )
				continue;
			if ( (strlen(seqprevtmp)+(2*local_read_booster->maxovh)) < len) // ne peut etre inclus
				continue;
			// avant de calculer je fais verifier que ce n'est pas exactement la meme sequence
			if ( onehit->strand == '-' )
				rcseqprev = RevCompSeq(seqprevtmp);
			int diff=TRUE;
			if ( len == strlen(seqprevtmp) ) // peut etre faire un HASH_ITER rien que pour cela ? => penible a cause du lock au dessus ca va doubler
			{	
				if ( onehit->strand == '-' )
					diff = strcmp(seqtmp,rcseqprev);
				else
					diff = strcmp(seqtmp,seqprevtmp);
			}
			if ( diff )
			{
				if ( onehit->strand == '-' )
					PairwiseMatchGapped(rcseqprev,seqtmp,&hsp,GAP_PENALTY,p_mat);
				else
					PairwiseMatchGapped(seqprevtmp,seqtmp,&hsp,GAP_PENALTY,p_mat);
			}
			if ( rcseqprev != NULL )
			{
				free(rcseqprev);
				rcseqprev=NULL;
			}
			// c'est fait expres de ne pas mettre maxmm comme seuil, on est sur des grandes sequences
			if ( ! diff || (hsp.pci >= local_read_booster->minpci && hsp.tstart <= (1+local_read_booster->maxovh) && hsp.tend >= (len-local_read_booster->maxovh)) )
			{
				if (pthread_rwlock_wrlock(&local_read_booster->a_lock_a_kmer2contigs[seqid%MAX_THREAD_LOCKS]) != 0) Fatal("can't get wrlock");
				DelSeqHub(seqsrc,seqid,what);
				pthread_rwlock_unlock(&local_read_booster->a_lock_a_kmer2contigs[seqid%MAX_THREAD_LOCKS]);
				break;
			}
		}
		FreeHits(&rh_hits);
	}
	return(NULL);
}

// je vais mapper les hq sur les long reads et je vais assembler les long reads s'il y a un lien
int BuildVeryLongReads (read_booster_t *read_booster,char *verylongreadsfile,unsigned long long *max_mem)
{
//	fprintf(stderr,"BuildVeryLongReads: minpci=%.2f minovl=%d maxovh=%d, maxmm=%d maxreadlen=%d (6*insert_size)\n",read_booster->minpci,read_booster->minovl,read_booster->maxovh,read_booster->maxmm,maxrdlen);
	seqid_t lastid_before_vlr=read_booster->nextid; // < dans la comparaison
	seqid_t init_vlr=GetNextContigSeqID(read_booster,1);
	seqid_t *a_tasks=NULL;
	seqid_t maxid=0;
	seqid_t max=InitArrayOfSeqOk(read_booster,&a_tasks,&maxid,LR); // liste des LR qui font potentiellement un pont avec d'autres LR
	
	dejavu_t *h_deja=NULL;
	
	fprintf(stderr,"max=%u, maxid=%u\n",max,maxid);
	
	int j;
	for(j=0;j<read_booster->nb_cpus;j++)
	{
		read_booster->nextid_per_threads_curr[j]=0;
		read_booster->nextid_per_threads_max[j]=0;
		read_booster->current_nt_size_per_threads[j]=init_vlr;
	}
	
	DeleteDejaArray(read_booster);
	
	arg_threads_t argt;
	argt.complex_arg[0]=read_booster;
	argt.complex_arg[1]=a_tasks;
	
//	fprintf(stderr,"NEXTID = %u\n",lastid_before_vlr);
	
	InitSMPStatusBar(read_booster->nb_cpus,&argt);
//	MultithreadingNoSlice(BuildVeryLongReadsSliceBlock,&argt,max,Min(24,read_booster->nb_cpus)); // (read_booster,0,max,a_tasks,tag,maxrdlen,&rh_pseudo_pe,ap_lr_links,&read_booster_verylong.h_seqcontigs);
	Multithreading(BuildVeryLongReadsSliceBlock,&argt,max,read_booster->nb_cpus); // (read_booster,0,max,a_tasks,tag,maxrdlen,&rh_pseudo_pe,ap_lr_links,&read_booster_verylong.h_seqcontigs);
	CloseSMPStatusBar(read_booster->nb_cpus,&argt);
	
	if ( read_booster->debug )
		InitMaxMemUsage(max_mem);
	
	ResetKmerize(read_booster,LR);
	if ( 0 && read_booster->debug )
	{
		RescueUnmappedReads(read_booster);
		char seqfile_nok[MAXFILENAME];
		sprintf(seqfile_nok,"%s.missing_reads.fa",verylongreadsfile);
		DumpFasta(read_booster,seqfile_nok,TRUE,FALSE,PAIRS,0);
	}
	
	if ( read_booster->debug )
		InitMaxMemUsage(max_mem);
	
//	DeleteSequenceSetHash(&rh_pseudo_pe);
//	for(i=1;i<cpt;i++)
//		FreeHits(&ap_lr_links[i]);
		
//	free(ap_lr_links);
	
	DeleteDeja(&h_deja);
	
	free(a_tasks);
	
	if ( read_booster->nrdb )
//		RemoveIncludedSequencesLR(read_booster,lastid_before_vlr);
		RemoveIncludedSequencesLR(read_booster,maxid);
	DumpFasta(read_booster,verylongreadsfile,FALSE,TRUE,LR,0);
	
//	fprintf(stderr,"BuildVeryLongReads: done >%s<\n",verylongreadsfile);
	return(TRUE);
}


void *BuildVeryLongReadsSliceBlock (void *argt)
{
	// recuperation des parametres
	arg_threads_t *thr_data=(arg_threads_t *)argt;
	unsigned long long imin=thr_data->imin;
	unsigned long long imax=thr_data->imax;
	read_booster_t *read_booster=(read_booster_t *)thr_data->complex_arg[0];
	seqid_t *ra_tasks=(seqid_t *)thr_data->complex_arg[1];

	hits_t *rh_hits1=NULL;
	hits_t *rh_hits2=NULL;
	hits_t *onehit1=NULL;
	hits_t *onehit2=NULL;
	hits_t *tmphit1=NULL;
	hits_t *tmphit2=NULL;
	
	//if ( read_booster->debug )
//		fprintf(stderr,"maxseqlen=%d\n",read_booster->max_seqlen); 
	
	short **p_mat=NULL;
	MallocMatrixOfShort(&p_mat,read_booster->max_seqlen*2);
	
#ifndef REVMAP
	short lmaxkmers = read_booster->maxkmers;
#else
	// lorsque j'indexe les LR tous les klen, je ne peux pas trouver les HQ incluses, il me manque un k
	short lmaxkmers = read_booster->maxkmers-1;
#endif
	
	sequenceset_t **seqsrc=read_booster->h_seqcontigs;
	int nb_cpus = read_booster->nb_cpus;
	unsigned long long i;
	
	char pair_of_lr[MAXREADNT]={0};
	char mergedpair[MAXREADNT];
	hsp_t hsp;
	char *a_reads_extensible[2]={NULL,NULL};
	char *a_readsori[2]={NULL,NULL};
	seqid_t gapid;
	unsigned long long decime=(imax-imin)/10;
	if ( decime == 0 )
		decime = 1;
	
	seqid_t cpt=GetNextContigSeqIDOneThread(read_booster,thr_data->thread_id);
	
//	fprintf(stderr,"cpt=%u, t(%d)\n",cpt,thr_data->thread_id);
	
	int helper=-1;
	int helped=FALSE;
	for(i=imin;i<=imax;i++)
	{
		if ( (i % (read_booster->nb_cpus*3)) == 0 ) // je ne veux pas verifier trop souvent
		{
			if ( (helper=GetFinishedThreadId(thr_data->self_array)) != -1 )
			{
				if ( MultithreadingShareSliceWithHelper(BuildVeryLongReadsSliceBlock,argt,i,imax,thr_data->thread_id,helper,thr_data->job_id) )
				{
					helped=TRUE;
					break;
				}
			}
		}
		if ( (i-imin) % decime == 0 )
			SMPStatusBar(thr_data->thread_id,(unsigned long long)(i-imin),(unsigned long long)(imax-imin+1),thr_data);
		
		gapid = *(ra_tasks+i);// paire resolue en LR potentiellement cabable de combler un trou
		
		char *pair = GetSeq(seqsrc,gapid);
		// est ce que c'est une seqid encore active
		if ( pair == NULL )
			continue;
		
		// je recupere la LR correspondant au trou
		char *seqlr = GetSeqLR(seqsrc,gapid); 
		if ( seqlr == NULL )
			continue;

		char *a_reads[2]={NULL,NULL};
		Split(pair,'N',a_reads);
		
		//DelSeq(seqsrc,gapid,TRUE); // deletion partielle, a ce stade je n'ai plus besoin de la paires
		
		onehit1=NULL;
		onehit2=NULL;
		tmphit1=NULL;
		tmphit2=NULL;
		rh_hits2=NULL;
		rh_hits2=NULL;
		
		// j'identifie les LR (indexes dans read_booster) qui mappent les deux extremites
		// si je mets BEST je n'arrive pas a voir que les deux extremites mappent et que ce n'est pas un pont
		// suite a une perte de perf, peut etre du a la suppression de la redondance, je mets minkmers au lieu de maxkmers qui est tres stringent sachant que je fais l'alignement
		// je remets maxkmers pour essayer de trouver le pb, il faut laisser
		Map(read_booster,a_reads[0],&rh_hits1,lmaxkmers,ALL_SCORE,gapid,FALSE,TRUE);  					
		if ( rh_hits1 == NULL ) // ca ne peut pas etre le cas si je n'occulte pas la sequence courante (la PE vont matcher avec le LR)
		{
			FreeSplit(a_reads,2);
			continue; 
		}
		Map(read_booster,a_reads[1],&rh_hits2,lmaxkmers,ALL_SCORE,gapid,FALSE,TRUE);
		if ( rh_hits2 == NULL )
		{
			FreeSplit(a_reads,2);
			FreeHits(&rh_hits1);
			continue; 
		}
		
// si un long read (dans rh_hits1/2) qui matche avec les deux extremites de la pe hq alors la pe n'est pas un pont entre LR disjoints je ne le prendrais pas en compte

		HASH_ITER(hh,rh_hits1,onehit1,tmphit1)
		{
			HASH_FIND(hh,rh_hits2,&onehit1->id,sizeof(seqid_t),onehit2);
			if ( onehit2 && onehit1->strand == onehit2->strand )
			{
				HASH_DEL(rh_hits1,onehit1);
				HASH_DEL(rh_hits2,onehit2);
				free(onehit2);
				free(onehit1);
			}
		}

//		# double sort pour reproduire en cpu1
		HASH_SORT(rh_hits1,HitSortByCount);
		HASH_SORT(rh_hits2,HitSortByCount);
		char *seq1,*rcseq1;
		char *seq2,*rcseq2;
		HASH_ITER(hh,rh_hits1,onehit1,tmphit1)
		{
			seq1 = GetSeqLR(seqsrc,onehit1->id);
			assert(seq1!=NULL);
			rcseq1 = NULL;
			HASH_ITER(hh,rh_hits2,onehit2,tmphit2)
			{
				seq2 = GetSeqLR(seqsrc,onehit2->id);
				assert(seq2!=NULL);
	
				if ( onehit1->strand == '+' )
				{
					if ( onehit2->strand == '+' )	 // + + (sachant que les deux ext sont dans le meme sens
						sprintf(pair_of_lr,"%sN%s",seq1,seq2);
					else
					{
						rcseq2 = RevCompSeq(seq2);
						sprintf(pair_of_lr,"%sN%s",seq1,rcseq2);
						free(rcseq2);
					}
				}
				else
				{
					if ( rcseq1 == NULL )
						rcseq1 = RevCompSeq(seq1);
					assert(rcseq1 != NULL);
					if ( onehit2->strand == '+' )
						sprintf(pair_of_lr,"%sN%s",rcseq1,seq2);
					else
					{
						rcseq2 = RevCompSeq(seq2);
						assert(rcseq2!=NULL);
						sprintf(pair_of_lr,"%sN%s",rcseq1,rcseq2);
						free(rcseq2);
					}
				}
				
				// TODO: voir si la paire a deja ete resolue (dans le thread?/shared?) et ne pas tenter si oui
				
				// au cas ou j'ai deja fais un test avorte je supprime la sequence
				DelSeq(seqsrc,cpt,FALSE);
				SetSeq(seqsrc,cpt,pair_of_lr);
				SetSeqStatus(seqsrc,cpt,'5');
				
				char *pair=pair_of_lr;
				
				Split(pair,'N',a_readsori);
				
				// je peux la supprimer une fois que je l'ai traite (apres le Split vu que je n'utilise plus pair)
				
				int assembled = FALSE;
				
				// avant d'etendre je regarde si je peux merger
				if ( PairwiseMatchUngapped(a_readsori[0],a_readsori[1],&hsp,p_mat) && ( hsp.mm <= read_booster->maxmm || hsp.pci >= read_booster->minpci ) &&  hsp.len >= read_booster->minovl )
				{
					mergedpair[0]='\0';
					if ( hsp.qend >= (strlen(a_readsori[0])-read_booster->maxovh) && (hsp.tstart <= (1+read_booster->maxovh) ) )
					{
						// je tolere l'overhang
						strncpy(mergedpair,a_readsori[0],hsp.qend);
						mergedpair[hsp.qend]='\0';
						strcat(mergedpair,a_readsori[1]+(hsp.tend+1-1));
					}
					else if ( hsp.tend >= (strlen(a_readsori[1])-read_booster->maxovh) && ( hsp.qstart <= (1+read_booster->maxovh) ) ) //je regarde en intervertissant l'ordre
					{
						strncpy(mergedpair,a_readsori[1],hsp.tend);
						mergedpair[hsp.tend]='\0';
						strcat(mergedpair,a_readsori[0]+(hsp.qend+1-1));
					}
					if ( strlen(mergedpair) && IsSequenceConsistantWithKmers(read_booster,mergedpair,read_booster->minkdepth) )
					{
						// je ne locke pas, il est unique
						SetSeqLR(seqsrc,cpt,mergedpair);
						assembled = TRUE;
					}
				}
				
				if ( ! assembled )
				{
					StrCat(&a_reads_extensible[0],a_readsori[0],FALSE);
					StrCat(&a_reads_extensible[1],a_readsori[1],FALSE);
					assembled = FillInGapWithLR(cpt,a_reads_extensible,gapid,'+',read_booster,FALSE,&p_mat[0]);
				}
				// je fais le choix de faire le calcul et de supprimer si redondant
				// avant je ne faisais pas le calcul si les paires identiques etaient resolues
				// ==> c'est plus couteux en CPU mais j'aurais moins de concurrence en SMP
				// ==> je pourrais voir des solutions alternatives
				if ( assembled )
				{
					char *seq=GetSeqLR(seqsrc,cpt);
					if ( ! AddNewVeryLongRead(read_booster,seq,cpt) )
					{
						DelSeq(seqsrc,cpt,FALSE); // je supprime la paire et la LR car je l'ai deja
					}
					else
					{
						//fprintf(stderr,"Assembled: %u (t%d)\n",cpt,thr_data->thread_id);
						DelSeq(seqsrc,cpt,TRUE); // del partiel = que la paire
						cpt=GetNextContigSeqIDOneThread(read_booster,thr_data->thread_id);
					}
				}
				
				if ( a_reads_extensible[0] != NULL )
					*a_reads_extensible[0]='\0';
				if ( a_reads_extensible[1] != NULL )
					*a_reads_extensible[1]='\0';
				
			}
			if ( rcseq1 != NULL )
				free(rcseq1);
		}
		FreeSplit(a_reads,2);
		FreeHits(&rh_hits1);
		FreeHits(&rh_hits2);
	}
	if ( ! helped )
		SetThreadIdAsFinished(thr_data->self_array,thr_data->thread_id);
	FreeSplit(a_readsori,2);
	FreeSplit(a_reads_extensible,2);
	FreeMatrixOfShort(&p_mat[0],read_booster->max_seqlen*2);

	return(NULL);
}


int FillInGapWithLR(seqid_t vlrid,char *a_reads_extensible[],seqid_t lrid,short strand,read_booster_t *read_booster,int lock,short **p_mat)
{
hits_t *rh_cand1=NULL;
hits_t *rh_cand2=NULL;
char  newseq1r[MAXREADNT]={0};
char  newseq1l[MAXREADNT]={0};
char  newseq2r[MAXREADNT]={0};
char  newseq2l[MAXREADNT]={0};
char seqlrtmp[MAXREADNT];
hsp_t hsp;
char mergedpair[MAXREADNT];
int assembled = FALSE;

		sequenceset_t **seqsrc=read_booster->h_seqcontigs;
		pthread_rwlock_t *seqlock = &read_booster->a_lock_a_kmer2contigs[lrid%MAX_THREAD_LOCKS];
		
		int lenread1 = strlen(a_reads_extensible[0]);
		int lenread2 = strlen(a_reads_extensible[1]);
	
		// r: extension a droite de la lecture
		// l: extension a gauche
		short maxpos1r=0;
		float maxpci1l=0;
		float maxpci1r=0;
		short maxpos1l=SHRT_MAX;
		short maxpos2r=0;
		float maxpci2r=0;
		float maxpci2l=0;
		short maxpos2l = SHRT_MAX;
		newseq1r[0]='\0';
		newseq1l[0]='\0';
		newseq2r[0]='\0';
		newseq2l[0]='\0';
											
//			lrid = onelr->id;
//			char strand = onelr->strand; // c'est toujours + 
		
		char *seqlr;
		if ( lock )
		{
			if (pthread_rwlock_rdlock(seqlock) != 0) Fatal("can't get rdlock");
			seqlr=GetSeqLR(seqsrc,lrid); // recuperation de la LR qui pourrait faire le pont (elle matche sur les 2 ends (elles meme des LR))
			pthread_rwlock_unlock(seqlock);
		}
		else
			seqlr=GetSeqLR(seqsrc,lrid); // recuperation de la LR qui pourrait faire le pont (elle matche sur les 2 ends (elles meme des LR))
		
		if ( seqlr == NULL )
			return(FALSE);
		strcpy(seqlrtmp,seqlr);
		
		char *rctgt = NULL;
		if ( strand == '-' ) // je ne pense pas que cela serve
		{
			rctgt= RevCompSeq(seqlrtmp);
			strcpy(seqlrtmp,rctgt);
			free(rctgt);
		}
		int ltgt = strlen(seqlrtmp);

		// ici je regarde seulement si la LR comble le trou
		if ( PairwiseMatchUngapped(a_reads_extensible[0],seqlrtmp,&hsp,p_mat) )
			ExtendSequence(read_booster,lrid,&hsp,a_reads_extensible[0],seqlrtmp,lenread1,ltgt,&rh_cand1,&maxpos1r,&maxpci1r,newseq1r,&maxpos1l,&maxpci1l,newseq1l);
		//PrintHsp(&hsp);
		
		if ( PairwiseMatchUngapped(a_reads_extensible[1],seqlrtmp,&hsp,p_mat) )
			ExtendSequence(read_booster,lrid,&hsp,a_reads_extensible[1],seqlrtmp,lenread2,ltgt,&rh_cand2,&maxpos2r,&maxpci2r,newseq2r,&maxpos2l,&maxpci2l,newseq2l);
		//PrintHsp(&hsp);

		if ( strlen(newseq1r)+strlen(newseq2l)  == 0 )
		{
			FreeHits(&rh_cand1);
			FreeHits(&rh_cand2);
			return(FALSE);
		}
		
		if ( strlen(newseq1r) && ! StrEndWith(a_reads_extensible[0],newseq1r) )
		{	
			StrCat(&a_reads_extensible[0],newseq1r,FALSE);
		}
		if ( strlen(newseq2l) && ! StrStartWith(a_reads_extensible[1],newseq2l) )
		{	
			StrCat(&a_reads_extensible[1],newseq2l,TRUE);
		}
		// factorisable avec plus haut et avec LR
		if ( PairwiseMatchUngapped(a_reads_extensible[0],a_reads_extensible[1],&hsp,p_mat) )
		{
			if ( (hsp.mm <= read_booster->maxmm || hsp.pci >= read_booster->minpci) &&  hsp.len >= read_booster->minovl && ( hsp.qend >= (strlen(a_reads_extensible[0])-read_booster->maxovh)) && (hsp.tstart <= (1+read_booster->maxovh)) )
			{
				// je tolere l'overhang
				strncpy(mergedpair,a_reads_extensible[0],hsp.qend);
				mergedpair[hsp.qend]='\0';
				strcat(mergedpair,a_reads_extensible[1]+(hsp.tend+1-1));
				if ( strlen(mergedpair) && IsSequenceConsistantWithKmers(read_booster,mergedpair,read_booster->minkdepth) )
				{
					// je stocke les vlr comme lr a la fin du tableau de sequences
					//if (pthread_rwlock_wrlock(ptr_lock_seqloci) != 0) Fatal("can't get wrlock");
					// je ne locke pas car elle n'est traite que par une slice
					SetSeqLR(seqsrc,vlrid,mergedpair);
					//pthread_rwlock_unlock(ptr_lock_seqloci);
					assembled = TRUE;
//					fprintf(stderr,"Assemble: %s\n",mergedpair);
				}
			}
		}
	//	PrintHsp(&hsp);
		
	FreeHits(&rh_cand1);
	FreeHits(&rh_cand2);
	return(assembled);
}

// je fais mapped toutes les lectures notees '0'
void RescueUnmappedReads(read_booster_t *read_booster)
{
	fprintf(stderr,"RUR\n");
	sequenceset_t **seqsrc=read_booster->h_seqcontigs;
	seqid_t seqid;
	seqid_t maxid = read_booster->nextid;
	short klen=read_booster->klen;
	int i,r;
	for(seqid=1;seqid<maxid;seqid++) // pour toutes les paires hq
	{
		if ( seqsrc[seqid] == NULL )
			continue;
		
		if ( seqsrc[seqid]->lr != NULL )
		{
			seqsrc[seqid]->status='4';
			continue;
		}
		
		char *pair = GetSeq(read_booster->h_seqcontigs,seqid);
		if ( pair == NULL )
			continue;
		
		char *a_reads[2]={NULL,NULL};
		Split(pair,'N',a_reads);
		int count=0; // depuis que j'ai des klen > 16, il me faut 2 kmers rares pour declencher
		for(r=0;r<2 && count < 1;r++)
		{
			char *ptr_str=a_reads[r];
			int lmax = strlen(a_reads[r])-klen;
			kmercode_t kmercode;
			short init = FALSE;
			int countone=0;
			for(i=0;i<=lmax;i++,ptr_str++)
			{	
				init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
				if ( ! init )
				{
					i+=(klen-1);
					ptr_str+=(klen-1);
					continue;
				}
				// je trouve une lecture avec un kmer oublie=presente dans le jeu de depart et absente ou pas suffisament presente 
#ifndef REVMAP
				if ( read_booster->a_kmercode[kmercode] > read_booster->minkdepth && (read_booster->a_kmer2contigs[kmercode] == NULL || utarray_len(read_booster->a_kmer2contigs[kmercode]) < read_booster->minkdepth) ) 
#else
				if ( read_booster->a_kmercode[kmercode] > read_booster->minkdepth && read_booster->a_kmercode_selection[kmercode] < read_booster->minkdepth ) 
#endif
					countone++;
				
				if (countone > 1 )
				{
					count++;
					break;
				}
					/*
					seqid_t newid = GetNextContigSeqID(read_booster);
					SetSeqLR(seqsrc,newid,a_reads[r]);
					break;
					*/
			}
		}
		if ( count == 0 )
			seqsrc[seqid]->status='4';
		FreeSplit(a_reads,2);
	}
}


kmercode_t IdentifyFrequentsButUncapturedKmers(read_booster_t *read_booster)
{
	// Identification des kmers manquants
	short maxk_for_alloc = Min(read_booster->klen,16);
	unsigned long long max = pow(4,maxk_for_alloc);
	unsigned long long i;
//	char kmerdec[100];
	kmercode_t problematic_kmers=0;
	for(i=0;i<max;i++)
	{
		unsigned short nb = 0;
#ifndef REVMAP
		if ( read_booster->a_kmer2contigs[i] != NULL )
			nb = utarray_len(read_booster->a_kmer2contigs[i]);
#else
		nb = read_booster->a_kmercode_selection[i];
#endif
		if ( (read_booster->a_kmercode[i] > read_booster->depth ) && (nb < read_booster->minkdepth) )
		{
			problematic_kmers++;
		//	KmerDecode(i,read_booster->klen,kmerdec);
//			fprintf(stderr,"%u,%s IN=%d, SEL=%d\n",i,kmerdec,read_booster->a_kmercode[i],nb);
		}
		else
			read_booster->a_kmercode[i] = read_booster->minkdepth; // je mets a  minkdepth pour que cela passe IsKmerCompositionOK
	}
	if ( read_booster->debug )
		fprintf(stderr,"%u frequent kmers (>depth in data) are missing\n",problematic_kmers);
	return(problematic_kmers);
//	exit(EXIT_FAILURE);
}


void StrCat(char **read,char *newseq,int before)
{
//	char *check=(char *)malloc(sizeof(char)*(strlen(*read)+1));
	int len=0;
	char *tmpread=NULL;
	if ( *read != NULL )
	{
//		strcpy(check,*read);
		tmpread=(char *)malloc((strlen(*read)+1));
		assert(tmpread != NULL);
		strcpy(tmpread,*read);
		free(*read);
		*read=NULL;
		len=strlen(tmpread);
	}
	else
	{
		tmpread=(char *)malloc(sizeof(char));
		assert(tmpread != NULL);
		tmpread[0]='\0';
	}
	(*read)=(char *)malloc((len+strlen(newseq)+1));
	assert(*read!=NULL);
	if ( before )
		sprintf(*read,"%s%s",newseq,tmpread);
	else
		sprintf(*read,"%s%s",tmpread,newseq);
	
	free(tmpread);
	
	/*
	if ( strstr(*read,check) == NULL )
	{
		fprintf(stderr,"StrCat: %s %p check:%s\n",*read,&read,check);
		assert(0);
	}
	free(check);
	*/
}


int ExtendSequence(read_booster_t *read_booster,seqid_t id,hsp_t *hsp,char *read,char *r_tgt,short lenread,short ltgt,hits_t **rh_cand,short *r_maxposr,float *r_maxpcir,char* r_newseqr,short *r_maxposl,float *r_maxpcil,char* r_newseql)
{
	if ( ( hsp->mm > read_booster->maxmm && hsp->pci < read_booster->minpci ) ||  hsp->len < read_booster->minovl )
		return(FALSE);
	/* sur ralsto cela semble moins bien marcher avec le truc ci dessous
	if (  hsp->mm > read_booster->maxmm || hsp->pci < read_booster->minpci ||  hsp->len < read_booster->minovl )
		return(FALSE);
	*/
	
	// Q --------
	// S     ---------
	// je ne tolere l'overhang que sur la target sinon c'est mal gere car je ne rabote pas la query (=0)
	if ( (hsp->qend >= lenread-0) && (hsp->tstart <= (1+read_booster->maxovh) && hsp->tend <= ltgt-read_booster->minovl ))
	{
		if ( IsMergingConsistantWithKlib(read_booster,read,r_tgt+hsp->tend+1-1) ) // je verif que je ne cree pas de kmer rare, si ce n'est pas le cas c'est peut etre un variant
		{
			HitPlusPlus(rh_cand,id,'-'); //meme si je n'utilise pas il est compatible sur au moins une extremite
			if ( *r_maxpcir < hsp->pci || *r_maxposr < hsp->qstart) // meilleur pci ou plus long
			{
					*r_maxposr = hsp->qstart;
					*r_maxpcir = hsp->pci;
					strcpy(r_newseqr,r_tgt+hsp->tend+1-1);
			}
		}
	}
	// Q      --------
	// S ---------
	// je ne tolere l'overhang que sur la target sinon c'est mal gere car je ne rabote pas la query (=0)
//	else if ( (hsp.tend <= ltgt-read_booster->maxovh) && ( hsp.qstart <= (1+0) && hsp.tstart >= read_booster->minovl))
	else if ( (hsp->tend >= ltgt-read_booster->maxovh) && ( hsp->qstart <= (1+0) && hsp->tstart >= read_booster->minovl))
	{
		char lext[MAXREADNT]="";
		strncpy(lext,r_tgt,hsp->tstart-1);
		lext[hsp->tstart-1]='\0';
		if ( IsMergingConsistantWithKlib(read_booster,lext,read) )
		{
			HitPlusPlus(rh_cand,id,'-'); //meme si je n'utilise pas il est compatible sur au moins une extremite
			if ( *r_maxpcil < hsp->pci || hsp->qend < *r_maxposl )
			{
				*r_maxposl = hsp->qend;
				*r_maxpcil = hsp->pci;
				strcpy(r_newseql,lext);
			}
		}
	}
	else 
		return(FALSE);
	
	//fprintf(stderr,"%s\n>%s<//>%s< %p %p\n",r_tgt,r_newseql,r_newseqr,r_newseql,r_newseqr);
	return(TRUE);
}

int IsMergingConsistantWithKlib(read_booster_t *read_booster,char *seqleft,char *seqright)
{
	char newjonction[MAXREADNT]="";
	char tmp[MAXREADNT]="";
	if ( seqleft!=NULL)
	{
		if ( strlen(seqleft) > read_booster->klen )
			strcpy(newjonction,seqleft+strlen(seqleft)-read_booster->klen); // je prend un kmer a droite
		else
			strcpy(newjonction,seqleft);
	}
	else
		return(FALSE);
	if ( seqright != NULL )
	{
		if (  strlen(seqright) > read_booster->klen )
		{
			strncpy(tmp,seqright,read_booster->klen);
			tmp[read_booster->klen]='\0';
			strcat(newjonction,tmp);
		}
		else
			strcat(newjonction,seqright);
	}
	else
		return(FALSE);
	if ( ! IsSequenceConsistantWithKmers(read_booster,newjonction,read_booster->minkdepth) )
	{
//		fprintf(stderr,"IsMergingConsistantWithKlib: %s   %s||%s\n",newjonction,seqleft,seqright);
		return(FALSE);
	}
	return(TRUE);
}

int HitPlusPlus(hits_t **rh_hits,seqid_t id,char c)
{
	hits_t *onehit=NULL;
	HASH_FIND(hh,*rh_hits,&id,sizeof(seqid_t),onehit);
	if (onehit)
	{	
		onehit->count++;
	}
	else
	{
		onehit=(hits_t *)malloc(sizeof(hits_t));
		assert(onehit!=NULL);
		onehit->count=1;
		onehit->id=id;
		onehit->strand=c;
		HASH_ADD(hh,*rh_hits,id,sizeof(seqid_t),onehit);
	}
	return onehit->count;
}

int HitSetCount(hits_t **rh_hits,seqid_t id,char c,short count)
{
	hits_t *onehit=NULL;
	HASH_FIND(hh,*rh_hits,&id,sizeof(seqid_t),onehit);
	if ( onehit )
		onehit->count=count;
	else
	{
		onehit=(hits_t *)malloc(sizeof(hits_t));
		assert(onehit!=NULL);
		onehit->id=id;
		onehit->strand=c;
		onehit->count=count;
		HASH_ADD(hh,*rh_hits,id,sizeof(seqid_t),onehit);
	}
	return onehit->count;
}

char *MallocSeqPair(char *read1,char *read2)
{
	int len = strlen(read1);
	if (read2!=NULL)
		len+=strlen(read2);
	char *seq = (char *)malloc(len+2);
	assert(seq != NULL);
	strcpy(seq,read1);
	if ( read2 != NULL )
	{
		char *revcompread2 = RevCompSeq(read2);
		assert(revcompread2!=NULL);
		strcat(seq,"N");
		strcat(seq,revcompread2);
		free(revcompread2);
	}
	return(seq);
}
