/*--------------------------------
 * 
 * Jerome.Gouzy@inra.fr
 * Created: 20140120
 * Copyright INRA, All rights reserved
 * 
 * $Id: read_booster.h 520 2016-12-02 12:35:31Z gouzy $
 * 
 ---------------------------------*/

#ifndef READ_BOOSTER_H

#define READ_BOOSTER_H

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "uthash.h"
#include "utarray.h"
#include "digest_md5.h"
#include "paramparser.h"
#include "general.h"
#include "generalbioinfo.h"
#include "sequenceset.h"
#include "sequencesethash.h"
#include "sys/sysinfo.h" 
//#include "rd_utils.h"
#include <sys/types.h>
#include <sched.h>
#include <unistd.h>

             
#define MAXREADNT                 		200000
#define MAXREADNT_MATRIX                10000

#define CLEANUP_DATASET					0
#define ONLY_UNMAPPED 					1
#define INCREASE_DEPTH					2
#define REPLACE_PROBLEMATIC_PAIRS		4
#define RESCUE_MISSING_KMERS	        3
#define TAG_SEQUENCES_WITH_RARE_KMERS	5
#define ONLY_UNMAPPED_PAIRS				8
#define INCREASE_DEPTH_PAIRS			9

#define BEST_SCORE 						0
#define ALL_SCORE  						1
#define GAP_PENALTY -4

#define MAXKMER                         32

#define CONTIGER_CLEAN_ALL              1
#define CONTIGER_CLEAN_TMP              0
#define CONTIGER_CLEAN_SEQOBS           2

//#define MAP_HIGH_FREQ                   10000 // seuil a partir je zappe un kmer au moment du mapping (pour lr)
//#define MINKMERS_FOR_LR_EXTENSION       2

#define DEFAULT_NT_PER_LINE 200
#define DEFAULT_QUAL_VALUE  21

// seuils pour arreter l'extension dans long read (chevauchement significatif avec divergence nucleotidique)
#define STOP_EXTENTION_MINLEN 100
#define STOP_EXTENTION_MINPCI 90

#define MAXK_FOR_DEJAVU 			13

#define DEBUG 0 

#ifndef HASH_OVERHEAD
#define HASH_OVERHEAD(hh,head) (size_t)((((head)->hh.tbl->num_items * sizeof(UT_hash_handle)) + ((head)->hh.tbl->num_buckets * sizeof(UT_hash_bucket)) + (sizeof(UT_hash_table)) ))
#endif

#ifdef CARACAL_H
#define MAX_THREAD_LOCKS 1000
#else
#define MAX_THREAD_LOCKS 20000
#endif

#define SYNC_STEP	10

#define MAXDEPTH_ATTEMPTS 		5

#define LR_MAX_TIME 3600 // temps max pour resoudre une lecture

typedef struct depth
{
	char  *id;
	unsigned short count;
	UT_hash_handle hh;         /* makes this structure hashable */
} depth_t;


typedef struct dejavu 
{
	char *str;			/* ATTENTION est ce que l'on ne peut pas etre limite ou est ce que ce n'est pas trop couteux ? */
	UT_hash_handle hh;         /* makes this structure hashable */
} dejavu_t;

typedef struct dejavucode
{
	kmercode_t     code;		
	UT_hash_handle hh;         /* makes this structure hashable */
} dejavucode_t;

typedef struct hits
{
	seqid_t id;
	char    strand;
	int 	count;
	UT_hash_handle hh;         /* makes this structure hashable */
} hits_t;

typedef struct kmer2contigs  
{
	kmercode_t kmer;
	UT_array *list_of_contigs;
	UT_hash_handle hh;         /* makes this structure hashable */
} kmer2contigs_t;

typedef struct seqcount_t  {
     char *filename;
     unsigned long long count;
     UT_hash_handle hh;         /* makes this structure hashable */
} seqcount_t;

typedef struct read_booster 
{
	paramparser_t    *o_param;
	UT_array**       a_deja;
	pthread_rwlock_t a_lock_a_deja[MAX_THREAD_LOCKS]; // pour faire des lock sur des tranches calculees plutot que sur tout le tableau
	dejavu_t         *h_deja;	
	pthread_rwlock_t lock_h_deja;
	sequenceset_t**  h_seqcontigs;
	seqid_t          nb_max_seq;
	pthread_rwlock_t lock_seqcontigs; 
	depth_t          *h_depth_of_contigpairs;
	char             *pairdepth;
	seqid_t          mask_pairdepth;
	pthread_rwlock_t lock_h_depth_of_contigpairs;
	UT_array**       a_kmer2contigs;
	pthread_rwlock_t a_lock_a_kmer2contigs[MAX_THREAD_LOCKS]; // pour faire des lock sur des tranches calculees plutot que sur tout le tableau
//	UT_array**     a_kmer2obsoletecontigs;
	short          nb_cpus;
	char           *outdir;
	short          stranded_rnaseq;
	short          debug;
	seqid_t   		nextid;
	seqid_t   		nextid_per_threads_curr[MAX_CPUS];
	seqid_t   		nextid_per_threads_max[MAX_CPUS];
	short          *a_kmercode; // remplit a partir de fichier
	short          *a_kmercode_selection; // remplit a partir des donnees en RAM
	kmercode_t     kmercode_mask;
	unsigned long long   current_nt_size;
	unsigned long long   current_nt_size_per_threads[MAX_CPUS];
	pthread_rwlock_t a_lock_current_nt_size_per_threads[MAX_CPUS];
	short          klen;
	short          insert_size;
	short 		   min_readlen_foruse;
	unsigned long long   expected_nt_size;
	short          minkmers_for_lr_extension;
	short          map_high_freq_cutoff;
	short          minkmers;
	short          lminkmers;
	short          maxkmers;
	short          minkdepth;
	short		   maxpaircov;
	short		   minovl;
	float		   minpci;
	short		   maxovh;
	short		   maxmm;
	short 		   compress;
	short          paired_files;
	short          single_end;
	short		   depth;
	short		   nrdb;
	short		   disable_optimization;
	short		   process_done;
	pthread_rwlock_t lock_process_done;
	short          offsetk;
	int            step_thread;
	short		   maxdepth_attempts;
	int            max_seqlen;
} read_booster_t;


void 	InitBoostr(read_booster_t *read_booster,paramparser_t *o_param);
void 	InitBoostrCopy(read_booster_t *read_booster,read_booster_t *read_booster_ori);

void 	CleanUpBoostr(read_booster_t *read_booster,int mode);
void    ResetKmers(UT_array **kmer2contigs_param,int short);
static unsigned long long GetCurrentCoverage(read_booster_t *read_booster);
// j'en ai marre des pb de conversions, j'arrondi
inline unsigned long long GetCurrentCoverage(read_booster_t *read_booster)
{
	int i;
	unsigned long long total = read_booster->current_nt_size;
	for(i=0;i<read_booster->nb_cpus;i++)
	{
		if (pthread_rwlock_rdlock(&read_booster->a_lock_current_nt_size_per_threads[i]) != 0) Fatal("can't get rdlock");
		total+=read_booster->current_nt_size_per_threads[i];
		pthread_rwlock_unlock(&read_booster->a_lock_current_nt_size_per_threads[i]);
	}
	return(total/read_booster->expected_nt_size);
}

unsigned long long SetCurrentNtSize(read_booster_t *read_booster);


void 	ResetKmerize(read_booster_t *read_booster,int mode);
void 	*ResetKmerizeSlice(void *argt);
int     KmerizeContigSeq(read_booster_t *read_booster,seqid_t id,UT_array **kmer2ctg_param,sequenceset_t **seqcontigs_param,int what);
int 	Map(read_booster_t *read_booster,char *read,hits_t **h_hits,short minkmers,int type,seqid_t avoid,int no_high_freq,int for_assembly);
int  	AddIfNotDejaVu(dejavu_t **h_deja,char *str,int do_digest,int revcomp,int noadd,pthread_rwlock_t *lock);
int  	AddIfNotCodeDejaVu(dejavucode_t **h_deja,kmercode_t kmercode,short klen,int revcomp);
seqid_t AddNewContigSeq(read_booster_t *read_booster, char *read1, char *read2,int kmerize,seqid_t id);
seqid_t AddNewContigSeqDejaArray(read_booster_t *read_booster, char *read1, char *read2,int kmerize,seqid_t id,int thr_id);
int     RescuePairsWithFrequentsKmers(read_booster_t *read_booster,char *read1,char *read2,short depth,short *used);
//void 	EvalKmerDepth(read_booster_t *read_booster,int mode);
void 	IdentifyErroneousSequences(read_booster_t *read_booster);
// version revmap (seule //e pour le momment)
void IdentifyErroneousSequencesSeqBySeq(read_booster_t *read_booster);
void *IdentifyErroneousSequencesSeqBySeqSlice(void *argt);
seqid_t AddIfDepthOk(read_booster_t *read_booster,char *read1,char *read2,seqid_t ctg1ns,seqid_t ctg2ns,int max_depth,int kmerise,short *used, int thr_id);
int     UpdatePairDepth(read_booster_t *read_booster,seqid_t ctg1ns,seqid_t ctg2ns);
int 	UpdatePairDepthHash(read_booster_t *read_booster,char *ctgpair);
int     GetKmerCount(kmer2contigs_t **kmer2contigs,kmercode_t kmer);
void 	PushContigIntoKmerHash(read_booster_t *read_booster,UT_array **kmer2contigs,kmercode_t kmercode,seqid_t id);
int 	SortArrayBySeqid(const void *_a, const void *_b);

int  	AddPairOfReads(read_booster_t *read_booster,char *read1,char *read2,int mode,short depth,short *used,int thr_id);
int  	ReplaceProblematicPairs(read_booster_t *read_booster,char *read1,char *read2,short depth,short *used);

void 	DeleteDepth(depth_t **h_depth_of_contigpairs,int debug);
void DeleteSeqCount(seqcount_t **h_depth_of_contigpairs,int debug);
void 	DeleteDeja(dejavu_t **h_deja);
void 	DeleteDejaCode(dejavucode_t **h_deja);

void 	RescueUnmappedReads(read_booster_t *read_booster);

int 	TestMaxCoverage (read_booster_t *read_booster,short depth);
seqid_t GetNextContigSeqID(read_booster_t *read_booster,int step);
seqid_t GetNextContigSeqIDOneThread(read_booster_t *read_booster,int thr_id);
void 	UpdateCurrentNtSize(read_booster_t *read_booster,unsigned long long nt);
void 	UpdateCurrentNtSizeOneThread(read_booster_t *read_booster,unsigned long long nt,int thr_id);

seqid_t Load(read_booster_t *read_booster,char *filename,int kmerize,int preserve_id,int paired,unsigned int *maxrdlen,int what);
void 	DumpKmers(kmer2contigs_t *kmer2contigs,FILE *out);
int  	DumpFastaHash(read_booster_t *read_booster,char *fastafile,sequencesethash_t **seqset,int pair_files,int nrdb);
int  	DumpFasta(read_booster_t *read_booster,char *fastafile,int pair_files,int nrdb,int what,seqid_t init);
void 	DumpQual(FILE *f_qual,char *seq,int qual);
seqid_t GetSeqId(char *line);
int 	DumpOneSeq(FILE *f_out,seqid_t seqid,char *seq,float qualscore,int pair_files);

int 	IsMergingConsistantWithKlib(read_booster_t *read_booster,char *seqleft,char *seqright);
int 	IsSequenceDejaVu(read_booster_t *read_booster,char *seq,int what,kmercode_t *kmercode);
int		AddNewVeryLongRead(read_booster_t *read_booster,char *seq,seqid_t seqid);

void 	BuildLongReads(read_booster_t *read_booster,char *longreadsfile,unsigned long long *max_mem);
void 	*BuildLongReadsSliceBlock(void *argt);
int 	BuildVeryLongReads (read_booster_t *read_booster,char *verylongreadsfile,unsigned long long *max_mem);
void 	*BuildVeryLongReadsSliceBlock(void *argt);
int 	FillInGapWithLR(seqid_t vlrid,char *a_reads_extensible[],seqid_t lrid,short strand,read_booster_t *read_booster,int lock,short **p_mat);
void 	*RemoveIncludedSequencesMappedSlice(void *argt);
void 	RemoveIncludedSequencesLR(read_booster_t *read_booster,seqid_t limit);

seqid_t InitArrayOfSeqOk(read_booster_t *read_booster,seqid_t **a_tasks,seqid_t *maxid,int what);

kmercode_t 	IdentifyFrequentsButUncapturedKmers(read_booster_t *read_booster);

int     ExtendSequence(read_booster_t *read_booster,seqid_t id,hsp_t *hsp,char *read,char *r_tgt,short lenread,short ltgt,hits_t **rh_cand,short *r_maxposr,float *r_maxidr,char* r_newseqr,short *r_maxposl,float *r_maxidl,char* r_newseql);
void 	StrCat(char **read,char *newseq,int before);
int 	IsSequenceConsistantWithKmers(read_booster_t *read_booster,char *seq, short minkdepth);
float 	GetQualScore(read_booster_t *read_booster,char *seq);

void 	FreeHits(hits_t **h_hits);
void 	PrintHits(hits_t **h_hits);
void 	FreeSplit(char **a_strs,int nb);
int 	HitPlusPlus(hits_t **rh_hits,seqid_t id,char c);
int     HitSetCount(hits_t **rh_hits,seqid_t id,char c,short count);
char *  MallocSeqPair(char *read1,char *read2);
int     HitSortByCount(hits_t *a,hits_t *b);
int 	UpdateHits(UT_array *a_lst,hits_t **rh_hits,char strand,short minkmers,short *max, int type,seqid_t avoid);

double GetMeanLengthLR(read_booster_t *read_booster);


#endif
