/*--------------------------------
 * 
 * Jerome.Gouzy@inra.fr
 * Created: 20140120
 * Copyright INRA, All rights reserved
 * 
 * $Id: paramparser.c 475 2015-12-28 15:51:57Z gouzy $
 * 
 ---------------------------------*/
#include "paramparser.h"

char *itoa(long val,char buf[])
{
	sprintf(buf,"%ld",val);
	return buf;
}

char *ftoa(double val,char buf[])
{
	sprintf(buf,"%lf",val);
	return buf;
}

void ParamParserDelete(paramparser_t **pp)
{
	paramparser_t *onepp = NULL;
	paramparser_t *tmppp = NULL;
	HASH_ITER(hh,*pp,onepp,tmppp)
	{
		HASH_DEL(*pp,onepp);
		free(onepp->key);
		free(onepp->value);
		free(onepp);
	}
}

void ParamParserPrintCli(int argc,char *argv[],FILE *out,char *revision)
{
	int i;
	
	for(i=0;i<argc;i++)
		fprintf(out,"%s ",argv[i]);
	fprintf(out,"\n");
	if ( revision != NULL )
		fprintf(out,"Revision: %s\n",revision);
}

int  ParamParserIsDefined(paramparser_t **pp,const char *key)
{
	paramparser_t *onepp = NULL;
	
	HASH_FIND_STR(*pp,key,onepp);
	if ( ! onepp || onepp->value == NULL )
		return 0;
	return 1;
}

void  ParamParserSet(paramparser_t **pp,const char *key,const char *value)
{
paramparser_t *onepp = NULL;
	
	HASH_FIND_STR(*pp,key,onepp);
	
	if ( ! onepp )
	{
		onepp = malloc (sizeof(paramparser_t));
		assert(onepp);
		onepp->key = (char *) malloc(sizeof(char)*(strlen(key)+1));
		assert(onepp->key);
		strcpy(onepp->key,key);
		if ( value != NULL )
		{
			onepp->value = (char *) malloc(sizeof(char)*(strlen(value)+1));
			assert(onepp->value);
			strcpy(onepp->value,value);
		}
		else
		{
			onepp->value=NULL;
		}
		onepp->type = '_';
		HASH_ADD_KEYPTR(hh,*pp,onepp->key,strlen(onepp->key),onepp);
	}
	else
	{
		if(onepp->value!= NULL)
			free(onepp->value);
		onepp->value = (char *) malloc(sizeof(char)*(strlen(value)+1));
		assert(onepp->value);
		strcpy(onepp->value,value);
	}
}

void  ParamParserSetType(paramparser_t **pp,const char *key,char type)
{
paramparser_t *onepp = NULL;
	
	HASH_FIND_STR(*pp,key,onepp);
	if ( onepp )
		onepp->type=type;
	else
		fprintf(stderr,"ParamParser SetType invalid on unset parameter : %s\n",key);
}

void ParamParserAssertDefined(paramparser_t **pp,const char *key)
{
	if ( ! ParamParserIsDefined(pp,key) )
	{
		fprintf(stderr,"Mandatory parameter undefined: %s\n",key);
		Usage(NULL);
		exit(0);
	}
}

void ParamParserAssertDirExists(paramparser_t **pp,const char *key)
{
	ParamParserAssertDefined(pp,key);
	char *value=ParamParserGet(pp,key);
	DIR *dir= opendir(value);
	if ( dir == NULL )
	{
		fprintf(stderr,"Directory must exist : %s = %s\n",key,value);
		exit(0);
	}
	free(dir);
}

void ParamParserAssertFileExists(paramparser_t **pp,const char *key)
{
	ParamParserAssertDefined(pp,key);
	char *value=ParamParserGet(pp,key);
	FILE *file= fopen(value,"r");
	if ( file == NULL )
	{
		fprintf(stderr,"File must exist : %s = %s\n",key,value);
		exit(0);
	}
	fclose(file);
}

void ParamParserInitLof(lof_t lof)
{
	int i;
	for(i=0;i<LOFMAXFILES;i++)
		lof[i]=NULL;
}

int ParamParserGetListOfFiles(paramparser_t **pp,const char *key,lof_t lof)
{
	lof[0]=NULL;
	FILE *f_in=NULL;
	char *value=ParamParserGet(pp,key);
	if ( value == NULL || strlen(value) == 0 )
		return(0);
	char cmd[MAXFILENAME+1];
	sprintf(cmd,"ls -1 %s",value);
	f_in=(FILE *)popen(cmd,"r");
	int cpt=0;
	int i;
	for(i=0;i<LOFMAXFILES;i++)
	{
		if ( lof[i] != NULL )
			free(lof[i]);
		lof[i]=NULL;
	}
	while(fgets(cmd,MAXFILENAME,f_in) != NULL) 
	{
		cmd[strlen(cmd)-1]='\0';
		assert(cpt < LOFMAXFILES);
		if ( lof[cpt] != NULL )
			free(lof[cpt]);
		lof[cpt]=(char *)malloc(sizeof(char)*(strlen(cmd)+1));
		assert(lof[cpt]);
		strcpy(lof[cpt],cmd);
		cpt++;
	}
	if ( cpt == 0 )
	{
		fprintf(stderr,"Fatal error: no file found\n");
		exit(EXIT_FAILURE);
	}
	lof[cpt]=NULL;
	pclose(f_in);
	return(cpt);
}

void ParamParserDeleteListOfFiles(paramparser_t **pp,lof_t lof)
{
	int i=0;
	for(i=0;i<LOFMAXFILES;i++)
		if ( lof[i] != NULL )
		{
			free(lof[i]);
			lof[i]=NULL;
		}
}


void ParamParserAssertInteger(paramparser_t **pp,const char *key)
{
	ParamParserAssertDefined(pp,key);
	int i;
	char *value=ParamParserGet(pp,key);
	char buf[1000];
	sscanf(value,"%d",&i);
	if ( strcmp(itoa(i,buf),value) != 0 )
	{
		fprintf(stderr,"Integer expected for parameter : %s = %s\n",key,value);
		Usage(NULL);
		exit(0);
	}
}

char *ParamParserGet(paramparser_t **pp,const char *key)
{
	paramparser_t *onepp = NULL;
	
	HASH_FIND_STR(*pp,key,onepp);
	
	if ( ! onepp )
	{
		return NULL;
	}
	return onepp->value;
}


void  ParamParserSetUnlessDefined(paramparser_t **pp,const char *key,const char *value)
{
	paramparser_t *onepp = NULL;
	
	HASH_FIND_STR(*pp,key,onepp);
	
	if ( ! onepp || onepp->value == NULL )
	{
		ParamParserSet(pp,key,value);
	}
}

void ParamParserDump(paramparser_t *pp,FILE *out)
{
	fprintf(out,"=========== Parameters ============\n");
	paramparser_t *ipp;
	for(ipp=pp; ipp != NULL; ipp=ipp->hh.next)
		fprintf(out,"%s=%s type=%c\n",ipp->key,ipp->value,ipp->type);
	fprintf(out,"===================================\n");
}

/*
"help readfiles=s outdir=s klen=i minkmers=i outprefix=s nb_cpus=i expected_nt_size=i depth=i build_long_reads replace_lowcov final_assembly preprocessed=s templ_depth=i minovl=i minpci=s maxovh=i build_very_long_reads long_reads=s");
*/
void ParamParserInit(paramparser_t **pp,int argc,char *argv[],char *str_getopt)
{
	char parameter[5000];
	char buf[5000];
	char type;
	int i;
	/* 1. Initialisation a partir des parametre geres par le programme en ligne de commande */
	char *tok = strtok(str_getopt," ");
	while( tok != NULL)
	{
		type='_';
		strcpy(buf,tok);
		for(i=0;i<strlen(buf);i++)
		{
			if ( buf[i] == '=' )
				buf[i]=' ';
		}
		sscanf(buf,"%s %c",parameter,&type);
		ParamParserSet(pp,parameter,NULL);
		ParamParserSetType(pp,parameter,type);
		tok = strtok(NULL," ");
	}
	/* 2. Lecture des parametres en ligne de commande */
	paramparser_t *ipp;
	char *param_read;
	int ambiguous=0;
	int skip=0;
	for(i=1;i<argc;i++)
	{
		if ( argv[i][0] != '-' )
		{
			fprintf(stderr,"Error parsing parameter >%s<. ParamParser manages parameters a la getopt only\n",argv[i]);
			exit(EXIT_FAILURE);
		}
			
		ambiguous=0;
		skip=0;
		param_read = argv[i];
		while(*param_read == '-')
			param_read++;
		char *found=NULL;
		for(ipp=*pp; ipp != NULL; ipp=ipp->hh.next)
		{
			if ( (found=strstr(ipp->key,param_read)) != NULL )
			{
				if ( strcmp(found,ipp->key) == 0 ) /* je verifie que je matche en debut de chaine */
				{
					if ( ipp->type != '_' )
					{
						if ( i+1 >= argc )
						{
							fprintf(stderr,"Error: can't get the expected value of parameter %s (%s)\n",param_read,ipp->key);
							exit(EXIT_FAILURE);
						}
						ParamParserSet(pp,ipp->key,argv[i+1]);
						skip=1;
					}
					else
						ParamParserSet(pp,ipp->key,"1");
					
					ambiguous++;
				}
			}
		}
		if ( ambiguous > 1 )
		{
			fprintf(stderr,"Error: ambiguous parameter %s\n",param_read);
			exit(EXIT_FAILURE);
		}
		if ( skip )
				i++;
	}
	if ( ParamParserIsDefined(pp,"help") )
	{
		Usage(argv);
		exit(0);
	}
}
