/*--------------------------------
 * 
 * Jerome.Gouzy@inra.fr
 * Created: 20140407
 * Copyright INRA, All rights reserved
 * 
 * $Id: sequencesethash.c 432 2014-07-22 17:28:45Z gouzy $
 * 
 ---------------------------------*/

// uthash Copyright (c) 2005-2014, Troy D. Hanson  http://troydhanson.github.com/uthash/ All rights reserved.

#include "sequencesethash.h"
#include "uthash.h"

char *GetSeqHash(sequencesethash_t **seqset,seqid_t id)
{
	sequencesethash_t *oneseq=NULL;
	HASH_FIND(hh,*seqset,&id,sizeof(seqid_t),oneseq);
	if ( ! oneseq )
		return NULL;
	return oneseq->sequence;
}

int MoveSeqHash(sequencesethash_t **seqset_src,sequencesethash_t **seqset_dst,seqid_t id)
{
	sequencesethash_t *oneseq=NULL;
	HASH_FIND(hh,*seqset_src,&id,sizeof(seqid_t),oneseq);
	if ( ! oneseq )
		return(FALSE);
	else
	{
		SetSeqHash(seqset_dst,id,oneseq->sequence);
		HASH_DEL(*seqset_src,oneseq);
		free(oneseq->sequence);
		free(oneseq);
	}
	return(TRUE);
}

/*
void SetQualHash(sequencesethash_t **seqset,seqid_t id,float qualscore)
{
	sequencesethash_t *oneseq=NULL;
	HASH_FIND(hh,*seqset,&id,sizeof(seqid_t),oneseq);
	if ( oneseq )
		oneseq->qualscore=qualscore;
}


void PrintQualHash(sequencesethash_t **seqset,seqid_t id)
{
	sequencesethash_t *oneseq=NULL;
	HASH_FIND(hh,*seqset,&id,sizeof(seqid_t),oneseq);
	if ( oneseq )
		fprintf(stderr,"%u Q=%f\n",oneseq->id,oneseq->qualscore);
}

*/

void SetSeqHash(sequencesethash_t **seqset,seqid_t id,char *seq)
{
	sequencesethash_t *oneseq=NULL;
	HASH_FIND(hh,*seqset,&id,sizeof(seqid_t),oneseq);
	if (  oneseq )
	{	
		//fprintf(stderr,"DEJA %u\n%s\n%s\n",id,oneseq->sequence,seq);
		HASH_DEL(*seqset,oneseq);
		free(oneseq->sequence);
		free(oneseq);
	}
	oneseq = (sequencesethash_t *)malloc(sizeof(sequencesethash_t));
	assert(oneseq);
	oneseq->id=id;
	oneseq->sequence=(char *)malloc(strlen(seq)+1);
	assert(oneseq->sequence);
	strcpy(oneseq->sequence,seq);
	HASH_ADD(hh,*seqset,id,sizeof(seqid_t),oneseq);
}

int SeqSortByLenHash(sequencesethash_t *a,sequencesethash_t *b)
{
	int diff = strlen(b->sequence) - strlen(a->sequence);
	if ( diff )
		return (diff);
	diff = a->id - b->id; // id croissant en cas d'egalite
	return (diff);
}

int SeqidLenSortByLen(seqidlenhash_t *a,seqidlenhash_t *b)
{
	int diff = (b->len) - (a->len);
	if ( diff )
		return (diff);
	diff = a->id - b->id; // id croissant en cas d'egalite
	return (diff);
}


/*
int SeqSortByQualHash(sequencesethash_t *a,sequencesethash_t *b)
{
	if ( b->qualscore > a->qualscore )
		return(1);
	else if ( b->qualscore < a->qualscore )
		return(-1);
	return (0);
}
*/

void DeleteSequenceSetHash(sequencesethash_t **h_seqset)
{
	sequencesethash_t *oneseq=NULL;
	sequencesethash_t *tmpseq=NULL;
	HASH_ITER(hh,*h_seqset,oneseq,tmpseq)
	{
		HASH_DEL(*h_seqset,oneseq);
		free(oneseq->sequence);
		free(oneseq);
	}
	*h_seqset=NULL;
}

seqid_t GetMaxSeqIdHash(sequencesethash_t **seqset)
{
	seqid_t max=0;
	sequencesethash_t *oneseq=NULL;
	sequencesethash_t *tmpseq=NULL;
	HASH_ITER(hh,*seqset,oneseq,tmpseq)
	{
		max=Max(max,oneseq->id);
	}
	return(max);
}


// je ne duplique pas, je supprime
void DuplicateSeqSetIntoSeqSetHash(sequenceset_t **src,sequencesethash_t **dst,seqid_t nb_seq_max)
{
	seqid_t i;
	for(i=0;i<=nb_seq_max;i++)
		if ( src[i] != NULL )
		{
			if (src[i]->lr != NULL )
			{
				SetSeqHash(dst,i,src[i]->lr);
	//			SetQualHash(dst,i,0);
				DelSeqLR(src,i); // ca consomme trop je supprime
			}
		}
}
